#require(MASS)
#require(fitdistrplus)
require(parallel)
useSmooth <- library(fields,logical.return=T)

DEFAULT_BASELINE_LENGTH = 90

BaselineHandlerEnv <- function(){
  pos <- match("baselineHandlerEnv", search())
  if (is.na(pos)){
    baselineHandlerEnv <- list()
    attach(baselineHandlerEnv, pos = length(search()) - 1)
    rm(baselineHandlerEnv)
    pos <- match("baselineHandlerEnv",search())
  }
  return(pos.to.env(pos))
}
putBaselineMeans <- function(key, value){
  BAM <- get("BaselineAlignmentMeans", envir = BaselineHandlerEnv(), mode = "any", inherits = FALSE)
  BAM[[key]] <- value
  assign("BaselineAlignmentMeans", BAM, envir = BaselineHandlerEnv())
}
getBaselineMeans <- function(key){
  BAM <- get("BaselineAlignmentMeans", envir = BaselineHandlerEnv(), mode = "any", inherits = FALSE)
  return(BAM[[key]])
}
baselineMeansContains <- function(key){
  BAM <- get("BaselineAlignmentMeans", envir = BaselineHandlerEnv(), mode = "any", inherits = FALSE)
  return(key %in% names(BAM))
}
putBaselineSDs <- function(key, value){
  BAS <- get("BaselineAlignmentSDs", envir = BaselineHandlerEnv(), mode = "any", inherits = FALSE)
  BAS[[key]] <- value
  assign("BaselineAlignmentSDs", BAS, envir = BaselineHandlerEnv())
}
getBaselineSDs <- function(key){
  BAS <- get("BaselineAlignmentSDs", envir = BaselineHandlerEnv(), mode = "any", inherits = FALSE)
  return(BAS[[key]])
}
baselineSDsContains <- function(key){
  BAS <- get("BaselineAlignmentSDs", envir = BaselineHandlerEnv(), mode = "any", inherits = FALSE)
  return(key %in% names(BAS))
}

assign("BaselineAlignmentMeans", list(), envir = BaselineHandlerEnv())
assign("BaselineAlignmentSDs", list(), envir = BaselineHandlerEnv())
assign("BaselineLength", DEFAULT_BASELINE_LENGTH, envir = BaselineHandlerEnv())

setBaselineLength <- function(newLength){
  assign("BaselineLength", newLength, envir = BaselineHandlerEnv())
}

getBaselineLength <- function(){
  baselineLength <- get("BaselineLength", envir = BaselineHandlerEnv(), mode = "any", inherits = FALSE)
  return(baselineLength)
}

loadBaselineAlignmentStats <- function(baselineKey,travOrder="STL",directory,gapExtendCost=1,gapOpenCost=3){
  # Patterns to find proper files and retrieve the lengths of their source sequences
  filePattern <- paste("BaselineScore",travOrder,"-Lengths[0-9]+_[0-9]+.tab",sep="")
  lengthsPattern <- paste("BaselineScore",travOrder,"-Lengths([0-9]+)_([0-9]+)\\.tab",sep="")
  
  # Get all files from the baseline directory
  dirFiles <- list.files(directory,pattern=filePattern)
  means <- list()
  sds <- list()
  lengths <- c()
  
  # Go through each file, retrieve data, and extract distribution parameters
  for (file in dirFiles){
    fullPath <- paste(directory,"/",file,sep="")
    m <- regexpr(lengthsPattern,file)
    lengthKey <- gsub(lengthsPattern,"\\1_\\2",file)
    currLengths <- as.numeric(strsplit(lengthKey,"_")[[1]])
    # Assumes symmetry with regards to sequence order, but requires the shorter number first in this code
    if (currLengths[2] < currLengths[1]){
      length1 <- currLengths[2]
      length2 <- currLengths[1]
      lengthKey <- paste(currLengths[2],currLengths[1],sep="_")
    }else{
      length1 <- currLengths[1]
      length2 <- currLengths[2]
    }
    # Only add information if it is new (duplicates should not occur)
    if (!lengthKey %in% names(means)){
      # Read scores from file
      baselineData <- read.csv(fullPath,sep="\t")
      normScores <- normalizeAlignmentScore(baselineData$Score,length1,length2,gapExtendCost,gapOpenCost)
      # Calculate distribution parameters
      means[[lengthKey]] <- mean(normScores)
      sds[[lengthKey]] <- sd(normScores)
      singleLengthKey1 <- as.character(length1)
      if (!length1 %in% lengths){
        lengths <- append(lengths,length1)
      }
    }
  }
  
  # Transform mean and sd data into matrices
  lengths <- sort(lengths)
  numLengths <- length(lengths)
  meanMat <- matrix(0,numLengths,numLengths)
  sdMat <- matrix(0,numLengths,numLengths)
  for (i in 1:numLengths){
    for (j in i:numLengths){
      len1 <- lengths[i]
      len2 <- lengths[j]
      lengthKey <- paste(len1,len2,sep="_")
      if (lengthKey %in% names(means)){
        meanMat[i,j] <- means[[lengthKey]]
        meanMat[j,i] <- means[[lengthKey]]
        sdMat[i,j] <- sds[[lengthKey]]
        sdMat[j,i] <- sds[[lengthKey]]
      }else{
        meanMat[i,j] <- NA
        meanMat[j,i] <- NA
        sdMat[i,j] <- NA
        sdMat[j,i] <- NA
        print(paste(lengthKey,"not in data"))
      }
    }
  }
  
  # Prep data for 2d interpolation
  meanVals <- meanMat[1,]
  sdVals <- sdMat[1,]
  xToSmoothI <- rep(log(lengths[1]),length(lengths))
  xToSmoothJ <- log(lengths)
  for (i in 2:numLengths){
    meanVals <- append(meanVals,meanMat[i,])
    sdVals <- append(sdVals,sdMat[i,])
    xToSmoothI <- append(xToSmoothI,rep(log(lengths[i]),length(lengths)))
    xToSmoothJ <- append(xToSmoothJ,log(lengths))
  }
  xToSmooth <- data.frame(length1=xToSmoothI,length2=xToSmoothJ)
  
  # Interpolate and store
  baselineAlignmentMeans <- list()
  baselineAlignmentSDs <- list()
  
  if (useSmooth){
    meanSmooth <- smooth.2d(meanVals,x=xToSmooth,theta=.1)
    baselineAlignmentMeans$Fit <- meanSmooth
    
    sdSmooth <- smooth.2d(sdVals,x=xToSmooth,theta=.1)
    baselineAlignmentSDs$Fit <- sdSmooth
  }else{
    baselineAlignmentMeans$Fit <- NA
    baselineAlignmentSDs$Fit <- NA
  }

  baselineAlignmentMeans$Raw <- list(Lengths=lengths,Stats=meanMat)
  putBaselineMeans(baselineKey,baselineAlignmentMeans)
  
  baselineAlignmentSDs$Raw <- list(Lengths=lengths,Stats=sdMat)
  putBaselineSDs(baselineKey,baselineAlignmentSDs)
}

# Assumes as symmetric score matrix (same sequences on both axes)
baselineNormalizedDistancesFromScores <- function(key,normScoreMat,lengths,numCores=detectCores(),batchSize=1200){
  # First get all unique lengths
  uniqueLengths <- sort(unique(lengths))
  numUniqueLengths <- length(uniqueLengths)
  
  # Generate reverse lookup: length->index
  lIndLookup <- list()
  for (i in 1:numUniqueLengths){
    lIndLookup[[uniqueLengths[i]]] <- i
  }

  print(paste(numCores,"cores"))
  
  # Generate lookup tables of interpolated means and SDs for length pairs
  meanMat <- matrix(0,nrow=numUniqueLengths,ncol=numUniqueLengths)
  sdMat <- matrix(0,nrow=numUniqueLengths,ncol=numUniqueLengths)
  if (!is.na(numCores) && numCores > 1){
    dataList <- 1:numUniqueLengths
    results <- mclapply(dataList,function(i,key,uniqueLengths){
        len1 <- uniqueLengths[i]
        means <- c()
        sds <- c()
        for (j in i:length(uniqueLengths)){
          len2 <- uniqueLengths[j]
          means[j-i+1] <- getBaselineMean(key,len1,len2)
          sds[j-i+1] <- getBaselineSD(key,len1,len2)
        }
        return(list(means,sds))
      },key,uniqueLengths)
    for (i in 1:numUniqueLengths){
      meanMat[i,i:numUniqueLengths] <- results[[i]][[1]]
      meanMat[i:numUniqueLengths,i] <- results[[i]][[1]]
      sdMat[i,i:numUniqueLengths] <- results[[i]][[2]]
      sdMat[i:numUniqueLengths,i] <- results[[i]][[2]]
    }
  }else{
    for (i in 1:numUniqueLengths){
      len1 <- uniqueLengths[i]
      for (j in i:numUniqueLengths){
        len2 <- uniqueLengths[j]
        meanMat[i,j] <- getBaselineMean(key,len1,len2)
        meanMat[j,i] <- meanMat[i,j]
        sdMat[i,j] <- getBaselineSD(key,len1,len2)
        sdMat[j,i] <- sdMat[i,j]
      }
    }
  }

  baselineLength <- getBaselineLength()
  standardMean <- getBaselineMean(key,baselineLength,baselineLength)
  standardSD <- getBaselineSD(key,baselineLength,baselineLength)
  print("Done prepping mean and sd matrices")
  
  # Initialize matrix and generate distances
  numCols <- ncol(normScoreMat) # Same as numRows
  distMat <- matrix(0,nrow=numCols,ncol=numCols)
  colnames(distMat) <- colnames(normScoreMat)
  rownames(distMat) <- rownames(normScoreMat)
  if (!is.na(numCores) && numCores > 1){
    # Run in parallel: Requires setup of each row data into a list
#     dataList <- list()
#     for (i in 1:(numCols-1)){
#       dataList <- append(dataList,list(as.numeric(normScoreMat[i,(i+1):numCols])))
#     }
#     results <- mclapply(dataList,baselineNormDistancesMCSubroutine,
#                         lengths,lIndLookup,meanMat,sdMat,mc.cores=numCores)
    
    matrixList <- lapply(1:(numCols-1), 
                         function(i,normScoreMat){return(as.matrix(normScoreMat)[i,])},
                         normScoreMat)
    print("Matrix list generated")
    
    # Deal with possibility of not having enough memory
    for (i in seq(1,nrow(normScoreMat)-1,batchSize)){
      rowRange <- i:(min(i+batchSize,nrow(normScoreMat))-1)
      print(paste("Working on rows",range(rowRange)))
      results <- mcmapply(baselineNormDistancesMCSubroutine,matrixList[rowRange],rowRange,
                          list(lengths),list(lIndLookup),list(meanMat),list(sdMat),standardMean,standardSD,mc.cores=numCores)
      
      #print("Calculated results, now putting them in output distance matrix")
      for (j in 1:length(results)){
        distMat[i+j-1,(i+j):numCols] <- results[[j]]
        distMat[(i+j):numCols,i+j-1] <- results[[j]]
      }
      print(paste("Calculated result section",i))
      gc()
    }
    
#    results <- mcmapply(baselineNormDistancesMCSubroutine,matrixList,1:(nrow(normScoreMat)-1),
#             list(lengths),list(lIndLookup),list(meanMat),list(sdMat),standardMean,standardSD,mc.cores=numCores)
    
#    print("Calculated results, now putting them in output distance matrix")
#    for (i in 1:length(results)){
#      distMat[i,(i+1):numCols] <- results[[i]]
#      distMat[(i+1):numCols,i] <- results[[i]]
#    }
  }else{
    for (i in 1:(numCols-1)){
      len1 <- lengths[i]
      print(paste(i,len1))
      # Just get lengths for columns that need to be calculated on current row
      currLengths <- lengths[(i+1):numCols]
      lengthIndicies <- c(do.call("cbind",lIndLookup[currLengths]))
      means <- meanMat[lIndLookup[[len1]],lengthIndicies]
      sds <- sdMat[lIndLookup[[len1]],lengthIndicies]
      
      # Calculate zscores
      zscores <- as.numeric((normScoreMat[i,(i+1):numCols] - means)/sds)
      # Calculate distance given a standard length
      distances <- 1 - (zscores*standardSD + standardMean)
      # Some distances may end up being negative. If so, set them to .1*exp(dist)
      if (sum(distances < 0.2) > 0){
        distances[distances < 0.2] <- exp(2*distances[distances < 0.2] - 2.00948)
      }
      
      distMat[i,(i+1):numCols] <- distances
      distMat[(i+1):numCols,i] <- distances
      
    }
  }
  return(distMat)
}

#baselineNormDistancesMCSubroutine <- function(currData,currInd,lengths,lIndLookup,meanMat,sdMat,standardMean,standardSD){
baselineNormDistancesMCSubroutine <- function(currData,currInd,lengths,lIndLookup,meanMat,sdMat,standardMean,standardSD){
  numCols <- length(lengths)
#  currInd <- numCols - length(currData)
  currLen <- lengths[currInd]
  
  #print(paste("Working on index",currInd))
  
  currData <- currData[(currInd+1):numCols]
  currLengths <- lengths[(currInd+1):numCols]
  lengthIndicies <- c(do.call("cbind",lIndLookup[currLengths]))
  means <- meanMat[lIndLookup[[currLen]],lengthIndicies]
  sds <- sdMat[lIndLookup[[currLen]],lengthIndicies]
    
  # Calculate zscores
  zscores <- as.numeric((currData - means)/sds)
  # Calculate distance given a standard length
  distances <- 1 - (zscores*standardSD + standardMean)
  # Some distances may end up being negative. If so, set them to .1*exp(dist)

  if (sum(distances < 0.2) > 0){
    #distances[distances < 0.2] <- .134*exp(distances[distances < 0.2])^2
    distances[distances < 0.2] <- exp(2*distances[distances < 0.2] - 2.00948)
  }
  return(distances)
}

getBaselineNormalizedDistance <- function(key,normScore,length1,length2){
  zscore <- getBaselineZScore(key,normScore,length1,length2)
  # Get the mean and SD of a reasonably representative length as a basis for distance
  baselineLength <- getBaselineLength()
  standardMean <- getBaselineMean(key,baselineLength,baselineLength)
  standardSD <- getBaselineSD(key,baselineLength,baselineLength)
  rescaledScore <- standardSD*zscore + standardMean
  distance <- 1 - rescaledScore
  if (distance < 0.2){
    distance <- exp(2*distance - 2.00948)
  }
  return(distance)
}

getBaselineZScore <- function(key,normScore,length1,length2){
  baselineMean <- getBaselineMean(key,length1,length2)
  baselineSD <- getBaselineSD(key,length1,length2)
  return((normScore-baselineMean)/baselineSD)
}
getBaselineMean <- function(key,length1,length2){
  if (baselineMeansContains(key)){
    return(getBaselineValue(getBaselineMeans(key),length1,length2))
  }
  return(NA)
}
getBaselineSD <- function(key,length1,length2){
  if (baselineSDsContains(key)){
    return(getBaselineValue(getBaselineSDs(key),length1,length2,ybound=0))
  }
  return(NA)
}
getBaselineValue <- function(fitData,length1,length2,type=c("Determine","Fit","Stats"),ybound=NA){
  type <- match.arg(type)
  log1 <- log(min(length1,length2))
  log2 <- log(max(length1,length2))
  position <- c(log1,log2)
  if (type == "Determine"){
    type <- "Fit"
    if (!useSmooth || length2/length1 < 1.2){
      type <- "Stats"
    }
  }
  if (useSmooth && type == "Fit"){
    interpParams <- getInterpolationParameters(position,fitData$Fit$x,fitData$Fit$z)
  }else{
    if (type=="Fit"){ warning("Cannot use Fit method (smooth.2d) as 'fields' library cannot be loaded") }
    interpParams <- getInterpolationParameters(position,log(fitData$Raw$Lengths),fitData$Raw$Stats)
  }
  return(interpolation2DSquare(position,interpParams$XCorners,interpParams$YValues,ybound=ybound))
}

interpolation <- function(x,x0,y0,x1,y1,FUNC=function(x){return(x)},phi=function(x){return(x)},psi=FUNC){
  y0 <- phi(y0)
  y1 <- phi(y1)
  return(((y1 - y0)*psi(x) + y0*psi(x1) - y1*psi(x0)) / (psi(x1) - psi(x0)))
}

# Expects xs to be a 3x2 matrix, each row being a point, each column an axis or dimension
interpolation2DSquare <- function(x,xs,ys,FUNC=function(x){return(x)},ybound=NA){
  if (x[1] == xs[1,1] && x[2] == xs[1,2]){
    return(ys[1])
  }else if (x[1] == xs[1,1]){
    # Interpolate just between 2 points
    matchInds <- which(xs[,1] == x[1])
    return(interpolation(x[2],xs[matchInds[1],2],ys[matchInds[1]],xs[matchInds[2],2],ys[matchInds[2]]))
  }else if (x[2] == xs[1,2]){
    # Interpolate just between 2 points
    matchInds <- which(xs[,2] == x[2])
    return(interpolation(x[1],xs[matchInds[1],1],ys[matchInds[1]],xs[matchInds[2],1],ys[matchInds[2]]))
  }
  else{
    if (!is.matrix(ys)){
      ys <- matrix(ys,length(ys),1)
    }
    # Create line between x1 and the point in question (x); s=slope, b=intercept
    s1 <- (xs[1,2]-x[2])/(xs[1,1]-x[1])
    b1 <- (x[1]*xs[1,2] - x[2]*xs[1,1]) / (x[1] - xs[1,1])
    # Create a line between x2 and x3
    s2 <- (xs[3,2]-xs[2,2])/(xs[3,1]-xs[2,1])
    b2 <- (xs[2,1]*xs[3,2] - xs[2,2]*xs[3,1]) / (xs[2,1] - xs[3,1])
    
    # Find the intersection point
    newX <- c()
    newX[1] <- (b2-b1)/(s1-s2)
    newX[2] <- s1*newX[1]+b1
    
    # Find the y-value at the intersection point
    # First get the ratio of the distances: point 2 to the new point and point 2 to point 3
    dist1Sq <- (newX[1]-xs[2,1])^2 + (newX[2]-xs[2,2])^2
    dist2Sq <- (xs[3,1]-xs[2,1])^2 + (xs[3,2]-xs[2,2])^2
    distanceRatio <- sqrt(dist1Sq/dist2Sq)
    newY <- ys[3,]*distanceRatio + ys[2,]*(1-distanceRatio)
    
    # Now find the y-value at x
    dist1Sq <- (x[1]-xs[1,1])^2 + (x[2]-xs[1,2])^2
    dist2Sq <- (newX[1]-xs[1,1])^2 + (newX[2]-xs[1,2])^2
    distanceRatio <- sqrt(dist1Sq/dist2Sq)
    calcY <- newY*distanceRatio + ys[1,]*(1-distanceRatio)
    
    # Correct in case where a boundary is specified and the value at specified point is 
    # beyond xs and ys (extrapolating) in the direction of the boundary (towards or beyond boundary)
    if (!is.na(ybound)){
      # If calcY is less than or greater than the min and max of ys, then we are extrapolating
      if (calcY < min(ys) && ybound < min(ys)){
        # offset values so boundary is at 0
        minY <- min(ys) - ybound
        calcY <- calcY - ybound
        const <- minY/exp(minY)
        return(const*exp(calcY) + ybound)
      }else if  (calcY > max(ys) && ybound > max(ys)){
        # Flip such that values are decreasing and boundary is 0
        minY <- ybound - max(ys)
        calcY <- ybound - calcY
        const <- minY/exp(minY)
        return(ybound - const*exp(calcY))
      }
    }
    return(calcY)
  }
}

getInterpolationParameters <- function(position,indexValues,matrixValues){
  x1Ind <- sum(position[1] >= indexValues)
  if (x1Ind == length(indexValues)){
    warning(paste("Position",position[1],"(position[1]) is outside of known values. Using these parameters will result in extrapolation."))
    x1Ind <- x1Ind-1
  }else if (x1Ind == 0){
    warning(paste("Position",position[1],"(position[1]) is outside of known values. Using these parameters will result in extrapolation."))
    x1Ind <- 1
  }
  x2Ind <- sum(position[2] >= indexValues)
  if (x2Ind == length(indexValues)){
    warning(paste("Position",position[2],"(position[2]) is outside of known values. Using these parameters will result in extrapolation."))
    x2Ind <- x2Ind-1
  } else if (x2Ind == 0){
    warning(paste("Position",position[2],"(position[2]) is outside of known values. Using these parameters will result in extrapolation."))
    x2Ind <- 1
  }
  x1Low <- indexValues[x1Ind]
  x1High <- indexValues[x1Ind+1]
  x2Low <- indexValues[x2Ind]
  x2High <- indexValues[x2Ind+1]
  # Determine whether the position is in the upper-right triangle of the rectangle
  slope <- (x2High-x2Low)/(x1Low-x1High)
  intercept <- (x1High*x2High-x2Low*x1Low)/(x1High-x1Low)
  if (position[2] > slope*position[1] + intercept){
    # Interpolation points form upper-right triangle
    xs <- matrix(c(x1High,x1High,x1Low,x2Low,x2High,x2High),3,2)
    xInds <- matrix(c(x1Ind+1,x1Ind+1,x1Ind,x2Ind,x2Ind+1,x2Ind+1),3,2)
  }else{
    # Interpolation points form lower-left triangle
    xs <- matrix(c(x1Low,x1High,x1Low,x2Low,x2Low,x2High),3,2)
    xInds <- matrix(c(x1Ind,x1Ind+1,x1Ind,x2Ind,x2Ind,x2Ind+1),3,2)
  }
  ys <- c()
  for (i in 1:3){
    ys <- append(ys,matrixValues[xInds[i,1],xInds[i,2]])
  }
  return(list(XCorners=xs,YValues=ys))
}

# getBaselineMean <- function(key,length1,length2){
#   if (key %in% names(BaselineAlingmentMeans)){
#     currBaselines <- BaselineAlingmentMeans[[key]]
#     if (length1 < length2){
#       lengthKey <- paste(length1,length2,sep="_")
#     }else{
#       lengthKey <- paste(length2,length1,sep="_")
#     }
#     # If this length pair is in the DB, just get it
#     if (lengthKey %in% names(currBaselines)){
#       return(currBaselines[[lengthKey]])
#     }else{ # Otherwise we need to interpolate or extrapolate
#       
#       if (length1 < min(currBaselines$Lengths)){
#         # Length1 is smaller than the smallest, much extrapolate down
#         
#       } else if (length1 > max(currBaselines$Lengths)){
#         # Length1 is larger than the largest, much extrapolate up
#         
#       } else {
#         # Interpolate
#         lower1Ind <- sum(currBaseline$Lengths < length1)
#         lower1 <- currBaseline$Lengths[lower1Ind]
#         upper1 <- currBaseline$Lengths[lower1Ind+1]
#         
#       }
#     }
#   }
#   return(NA)
# }

normalizeAlignmentScore <- function(rawScore,length1,length2,gapExtendCost=-1,gapOpenCost=0){
  if (length1 == length2){
    return(rawScore/length1)
  }else{
    # Expect negative costs, so the absolute values should be added back
    return((rawScore+abs(gapExtendCost)*abs(length2-length1)+abs(gapOpenCost))/min(length1,length2))
  }
}

plotBaselineDistributionParameters <- function(baselineKey,directory,...){
  require(plot3D)

  means <- getBaselineMeans(baselineKey)
  sds <- getBaselineSDs(baselineKey)

  if (is.null(means)){
    print(paste("Baseline key",baselineKey,"not found"))
    return() 
  }
  M <- mesh(means$Fit$x, means$Fit$y)
  zMask <- lower.tri(means$Fit$z)
  zFit <- means$Fit$z
  zFit[zMask] <- NA

  png(paste(directory,"/BaselineDistMeans-Surface-",gsub(" ","_",baselineKey),"-Fit.png",sep=""),height=500,width=500)
  par(mar = c(0.5, 0.5, 0.5, 0.5))
  par(mfrow = c(1, 1))
  surf3D(M$x,M$y,zFit,
         colkey=FALSE,
         bty="b2",...)
  dev.off()

  M <- mesh(log(means$Raw$Lengths), log(means$Raw$Lengths))
  zRaw <- means$Raw$Stats
  zMask <- lower.tri(means$Raw$Stats)
  zRaw[zMask] <- NA
  
  png(paste(directory,"/BaselineDistMeans-Surface-",gsub(" ","_",baselineKey),"-Raw.png",sep=""),height=500,width=500)
  par(mar = c(0.5, 0.5, 0.5, 0.5))
  par(mfrow = c(1, 1))
  surf3D(M$x,M$y,zRaw,
         colkey=FALSE,
         bty="b2",...)
  dev.off()

  M <- mesh(sds$Fit$x, sds$Fit$y)
  zMask <- lower.tri(sds$Fit$z)
  zFit <- sds$Fit$z
  zFit[zMask] <- NA
  
  png(paste(directory,"/BaselineDistSDs-Surface-",gsub(" ","_",baselineKey),"-Fit.png",sep=""),height=500,width=500)
  par(mar = c(0.5, 0.5, 0.5, 0.5))
  par(mfrow = c(1, 1))
  surf3D(M$x,M$y,zFit,
         colkey=FALSE,
         bty="b2",theta=160,...)
  dev.off()

  M <- mesh(log(sds$Raw$Lengths), log(sds$Raw$Lengths))
  zRaw <- sds$Raw$Stats
  zMask <- lower.tri(sds$Raw$Stats)
  zRaw[zMask] <- NA
  
  png(paste(directory,"/BaselineDistSDs-Surface-",gsub(" ","_",baselineKey),"-Raw.png",sep=""),height=500,width=500)
  par(mar = c(0.5, 0.5, 0.5, 0.5))
  par(mfrow = c(1, 1))
  surf3D(M$x,M$y,zRaw,
         colkey=FALSE,
         bty="b2",theta=160,...)
  dev.off()
}

plotBaselineDistribution <- function(baselineKey,inputDirectory,travOrder="STL",len1,len2,outputDirectory,
                                     fileDescrip="",fileType=c("png","eps","ps","tex","pdf","jpeg","jpg","tiff","bmp","svg")){
  fileType <- match.arg(fileType)
  # Patterns to find proper files and retrieve the lengths of their source sequences
  filename <- paste(inputDirectory,"/BaselineScore",travOrder,"-Lengths",len1,"_",len2,".tab",sep="")

  # Read scores from file
  baselineData <- read.csv(filename,sep="\t")
  normScores <- normalizeAlignmentScore(baselineData$Score,length1,length2,gapExtendCost,gapOpenCost)
  
  distPlot <- ggplot(normScores)+geom_density(linetype=0)+
    stat_function(fun=dnorm,mean=mean(normScores),sd=sd(normScores))

  # Calculate distribution parameters
  mean(normScores)
  sd(normScores)

  if (fileDescrip == ""){
    outputFilename <- paste(outputDirectory,"/BaselineDistribution_",travOrder,"_",len1,"_",len2,".",fileType,sep="")
  }else{
    outputFilename <- paste(outputDirectory,"/BaselineDistribution_",travOrder,"_",fileDescrip,"_",len1,"_",len2,".",fileType,sep="")
  }
  ggsave(outputFilename,distPlot)
  print(paste("Write to",outputFilename))
}
