#########################################################################
#
# ClusterAnalysAllData.r
#
# 1. Reads score data from pairwise alignments performed by PASTA package
# 2. Normalizes scores (and converts to distances) by lengths using 
#    random tree-sequence baselines.
# 3. Generates alignment-space using non-metric multidimensional scaling.
# 4. Clusters alignment-space using mclust.
# 5. Determines associations between clusters and known metadata groups.
#
# These steps are done on data subsets, generally by arbor type.
# Intermediary data is stored, and output is written in tables (csv) and
# visualizations (png).
#
#########################################################################

############# *SET UP ################

systemType <- "PC"
systemType <- "KrasClust"
arborNames <- c("Axon","Apical","Dendrite")
arborTypes <- c("Axon","ApicalDendrite","Dendrite")
globalAlignSubdir <- "/ScoringV3Data/"

# Set up directories
if (systemType == "PC"){
  clusterDir <- "C:/Documents and Settings/gillette/My Documents/research/ClusterAnalysis/"
  baseCodeDir <- "G:/Dropbox/RCode/"
  clusterGlobalDir <- "C:/Documents and Settings/gillette/My Documents/research/GlobalAlignment/"
  globalAlignDataDir <- paste("G:/AlignmentData/",globalAlignSubdir,sep="")
}else{
  clusterDir <- "/home/tgillett/ClusterAnalysis/"
  baseCodeDir <- "/home/tgillett/RUtils/"
  clusterGlobalDir <- "/home/tgillett/PairwiseAlignment/"
  globalAlignDataDir <- paste(clusterGlobalDir,globalAlignSubdir,sep="")
  baselineDir <- "/home/tgillett/PairwiseAlignment/Baseline/"
}
#motifAnalysisDir <- "G:/Dropbox/Motif Analysis/Motif Analysis R Files/"

# Load Utils
source(paste(baseCodeDir,"Utils.r",sep=""))
source(paste(baseCodeDir,"VisualizationUtils.r",sep=""))
source(paste(baseCodeDir,"StringUtils.r",sep=""))
source(paste(baseCodeDir,"StatisticsUtils.r",sep=""))

# Get sequence names
#sequenceNamesAll <- read.table(paste(clusterGlobalDir,"/fastaSequenceNames.txt",sep=""),stringsAsFactors=F)[,1]

# Get metadata and load related functions
#masterMetaData <- read.csv(paste(clusterDir,"arborMetaData-NM-MapType2.csv",sep=""),na.strings="",stringsAsFactors=FALSE,row.names=1)
masterMetaData <- read.csv(paste(clusterDir,"arborMetaData-NM_SpurLen2D-MapType2.csv",sep=""),na.strings="",stringsAsFactors=FALSE,row.names=1)
if ("CellType"%in%colnames(masterMetaData)){
  colnames(masterMetaData)[colnames(masterMetaData)=="CellType"] <- "Class2"
}
source(paste(clusterDir,"ClusterStatistics.r",sep=""))
source(paste(clusterDir,"ClusterAnalysisFunctions.r",sep=""))
source(paste(clusterDir,"ClassComparisonDefs.r",sep=""))
clusterMetaData <- usedMetaData
#missingSeqInds <- which(!sequenceNamesAll%in%rownames(clusterMetaData)) # Acute culture seqs removed from meta data
#missingSeqNames <- sequenceNamesAll[missingSeqInds]
#sequenceNames <- intersect(sequenceNamesAll,rownames(clusterMetaData))
sequenceNamesLists <- list()
for (i in 1:length(arborNames)){
  if (arborNames[i] != "AllArbors"){
    sequenceNamesLists[[arborNames[i]]] <- rownames(subset(clusterMetaData,ArborType==arborTypes[i]))
  }
}
sequenceNamesLists[["AllArbors"]] <- rownames(clusterMetaData)
#apicalSequences <- rownames(subset(clusterMetaData,ArborType=="ApicalDendrite"))
#dendriteSequences <- rownames(subset(clusterMetaData,ArborType=="Dendrite"|ArborType=="BasalDendrite"))

library(parallel)
library(MASS)
library(mclust)

###### *Specific Run Settings ########

matrixExt <- ".tab"
rownamesVal <- 1 # Could be NULL
# Read correlation matrix
travOrder <- "STL"
# If this is a random dataset rather than NMO
isRandom <- F
#baseline <- 10
randStr <- ""
spurLen <- 0
#spurLen <- 5
spurStr <- "_SpurLen2D"
#scoreParamStr <- "TreeIdentityG3E1"
scoreParamStr <- "G3E1Identity"
scoreParamStr <- "G2E1Identity"
scoreParamStr <- "G1E1Identity"
scoreParamStr <- "G0E1Identity"
# scoreParamStr <- ""
# scoreParamStr <- "Blosum50G6E2"
# scoreParamStr <- "Custom1G3E3"
# scoreParamStr <- "TreeIdentityG0E1"
#matrixExt <- ".csv"

if (isRandom){
  randStr <- paste("Rand",baseline,sep="")
}
if (spurLen > 0){
  spurStr <- paste("_SpurLen",spurLen,sep="")
}
if (spurStr != ""){
  spurRemovedLengths <- read.csv(paste("SequenceLengths-NM",spurStr,"-MapType2.csv",sep=""),row.names=1)
  clusterMetaData$SeqLen <- spurRemovedLengths[rownames(clusterMetaData),"SeqLen"]
}else{
  clusterMetaData <- usedMetaData
}


########### *Data loading ###########

# if (isRandom){
#   normScoreMatFilename <- paste(globalAlignDataDir,"/Rand",baseline,travOrder,"/normScores",randStr,travOrder,".",scoreParamStr,matrixExt,sep="")
# } else {
#   normScoreMatFilename <- paste(globalAlignDataDir,"/Neurite",travOrder,"/normScores",travOrder,".",scoreParamStr,matrixExt,sep="")
#   #normScoreMatFilename <- paste(globalAlignDataDir,"/",scoreParamStr,"/normScores",travOrder,".",scoreParamStr,matrixExt,sep="")
# }

# Read list of flagged neurons and eliminate them from data
if (systemType == "PC"){
  projectConfigDir <- "C:/Documents and Settings/gillette/workspace/NeuriteSequence/resources/"
}else{
  projectConfigDir <- "/home/tgillett/NeuriteSequence/resources/"
}
flaggedNeurons <- read.table(paste(projectConfigDir,"FlaggedNeurons.txt",sep=""),stringsAsFactors = F)[,1]

## For raw scores
arborType <- "AllArbors"
normScoreMats <- list() # Use normScoreMats
scoreMatFilename <- paste(globalAlignDataDir,"/ScoreMat_",arborType,travOrder,spurStr,"_",scoreParamStr,matrixExt,sep="")
normScoreMats[[arborType]] <- read.csv(scoreMatFilename,sep="\t",check.names=F,row.names=rownamesVal,na.strings="None",stringsAsFactors=F)

## Read per-character scores
normScoreMats <- list()
for (arborType in arborNames){
  if (isRandom){
    normScoreMatFilename <- paste(globalAlignDataDir,"/Rand",baseline,travOrder,"/normScores",randStr,travOrder,".",scoreParamStr,matrixExt,sep="")
  } else {
    normScoreMatFilename <- paste(globalAlignDataDir,"/NormScore_",arborType,travOrder,spurStr,"_",scoreParamStr,matrixExt,sep="")
    #normScoreMatFilename <- paste(globalAlignDataDir,"/",scoreParamStr,"/normScores",travOrder,".",scoreParamStr,matrixExt,sep="")
  }
  normScoreMatDF <- read.csv(normScoreMatFilename,sep="\t",check.names=F,row.names=rownamesVal,na.strings="None",stringsAsFactors=F)
  # Get in the same order
  normScoreMatDF <- normScoreMatDF[rownames(normScoreMatDF),rownames(normScoreMatDF)]
#  if (ncol(normScoreMatDF) > nrow(normScoreMatDF)){
#    normScoreMatDF <- normScoreMatDF[,-ncol(normScoreMatDF)]
#  }
  # Removed flagged neurons
  if (arborType == "Dendrite"){
    flaggedNeurites <- c(paste(flaggedNeurons,"Dendrite",sep="-"),paste(flaggedNeurons,"BasalDendrite",sep="-"))
  }else if (arborType == "Apical"){
    flaggedNeurites <- paste(flaggedNeurons,"ApicalDendrite",sep="-")
  }else if (arborType == "Axon"){
    flaggedNeurites <- paste(flaggedNeurons,arborType,sep="-")
  }else if (arborType == "AllArbors"){
    flaggedNeurites <- c(paste(flaggedNeurons,"Dendrite",sep="-"),paste(flaggedNeurons,"BasalDendrite",sep="-"),paste(flaggedNeurons,"ApicalDendrite",sep="-"),paste(flaggedNeurons,"Axon",sep="-"))
  }
  normScoreMatDF <- normScoreMatDF[
    !rownames(normScoreMatDF)%in%flaggedNeurites,
    !colnames(normScoreMatDF)%in%flaggedNeurites]
  
  normScoreMats[[arborType]] <- normScoreMatDF
}
rm(normScoreMatDF)

## If taking alignments for each arbor type from AllArbors
# seqNamesInMat <- rownames(normScoreMats[["AllArbors"]])
# normScoreMats[["Apical"]] <- normScoreMats[["AllArbors"]][intersect(seqNamesInMat,sequenceNamesLists[["Apical"])],]
# normScoreMats[["Axon"]] <- normScoreMats[["AllArbors"]][intersect(seqNamesInMat,sequenceNamesLists[["Axon"])],]
# normScoreMats[["Dendrite"]] <- normScoreMats[["AllArbors"]][intersect(seqNamesInMat,sequenceNamesLists[["Dendrite"])],]
#

########## *Data Cleaning ###########

# Trim out sequences not in other data sets (e.g. metadata), and get columns and rows in same order
#sequenceNames <- rownames(normScoreMatDF)[rownames(normScoreMatDF)%in%sequenceNames]
for (arborType in arborNames){
  sequenceNamesLists[[arborType]] <- intersect(sequenceNamesLists[[arborType]],rownames(normScoreMats[[arborType]]))
  normScoreMats[[arborType]] <- normScoreMats[[arborType]][sequenceNamesLists[[arborType]],sequenceNamesLists[[arborType]]]
}

# Replace any None/NA cases with data from across the diagonal
fillSymMatrix <- function(symMat){
  for (i in 1:nrow(symMat)){
    naPos <- is.na(symMat[i,]) | is.nan(as.numeric(symMat[i,]))
    symMat[i,naPos] <- symMat[naPos,i]
  }
  return(symMat)
}

#normScoreMatDF <- fillSymMatrix(normScoreMatDF)
for (arborType in arborNames){
  if (sum(is.na(normScoreMats[[arborType]])) > 0){
    normScoreMats[[arborType]] <- fillSymMatrix(normScoreMats[[arborType]])
  }
}

# Remove neurites not in clusterMetaData and non-Control neurites
for (arborType in arborNames){
  sharedSeqs <- intersect(rownames(normScoreMats[[arborType]]),rownames(clusterMetaData))
  normScoreMats[[arborType]] <- normScoreMats[[arborType]][sharedSeqs,sharedSeqs]
  
  controlNeurites <- clusterMetaData[rownames(normScoreMats[[arborType]]),"ExperCond"]=="Control"
  normScoreMats[[arborType]] <- normScoreMats[[arborType]][controlNeurites,controlNeurites]
}

############### *Distance matrix calculation ###################

# Calculate normalized distances based on baseline distributions for given length pairs
source(paste(clusterGlobalDir,"BaselineHandler.r",sep=""))
#baselineKey <- paste("G3E1_e5",travOrder)
#loadBaselineAlignmentStats(paste("G3E1_e5",travOrder),directory=paste(baselineDir,"GapOpen3_e5/",sep=""),travOrder=travOrder,gapOpenCost=-3)
#loadBaselineAlignmentStats(paste("G3E1_e3",travOrder),directory=paste(baselineDir,"GapOpen3_1000/",sep=""),travOrder=travOrder,gapOpenCost=-3)
#loadBaselineAlignmentStats(paste("G2E1_e3",travOrder),directory=paste(baselineDir,"GapOpen2_1000/",sep=""),travOrder=travOrder,gapOpenCost=-2)
baselineKey <- paste("G1E1_e3",travOrder)
loadBaselineAlignmentStats(paste("G1E1_e3",travOrder),directory=paste(baselineDir,"GapOpen1_1000/",sep=""),travOrder=travOrder,gapOpenCost=-1)
#loadBaselineAlignmentStats(paste("G0E1_e3",travOrder),directory=paste(baselineDir,"GapOpen0_1000/",sep=""),travOrder=travOrder,gapOpenCost=0)
#distMatDF <- baselineNormalizedDistancesFromScores(baselineKey,normScoreMatDF,clusterMetaData[colnames(normScoreMatDF),"SeqLen"])

arborType <- "Dendrite"
normScoreMatFilename <- paste(globalAlignDataDir,"/NormScore_",arborType,travOrder,spurStr,"_",scoreParamStr,matrixExt,sep="")
normScoreMatDF <- read.csv(normScoreMatFilename,sep="\t",check.names=F,row.names=rownamesVal,na.strings="None",stringsAsFactors=F)
if (ncol(normScoreMatDF) > nrow(normScoreMatDF)){
  normScoreMatDF <- normScoreMatDF[,-ncol(normScoreMatDF)]
}
# Removed flagged neurons
flaggedNeurites <- c(paste(flaggedNeurons,"Dendrite",sep="-"),paste(flaggedNeurons,"BasalDendrite",sep="-"))
normScoreMatDF <- normScoreMatDF[
  !rownames(normScoreMatDF)%in%flaggedNeurites,
  !colnames(normScoreMatDF)%in%flaggedNeurites]
normScoreMats[[arborType]] <- normScoreMatDF
rm(normScoreMatDF)
distMats[[arborType]] <- baselineNormalizedDistancesFromScores(baselineKey,normScoreMats[[arborType]],clusterMetaData[colnames(normScoreMats[[arborType]]),"SeqLen"],numCores=numCores)

numCores <- 8
distMats <- list()
for (arborType in c("AllArbors","Axon","Apical","Dendrite")){
  before <- date()
  print(paste("Starting",arborType,"at",before))
  distMats[[arborType]] <- baselineNormalizedDistancesFromScores(baselineKey,normScoreMats[[arborType]],clusterMetaData[colnames(normScoreMats[[arborType]]),"SeqLen"],numCores=numCores,batchSize=500)
  after <- date()
  print(paste("Finished at",after))
  write.csv(distMats[[arborType]],paste(globalAlignDataDir,arborType,"Distances_",travOrder,spurStr,"_",scoreParamStr,".csv",sep=""))
  #write.csv(distMats[[arborType]],paste(globalAlignDataDir,"ScoringV2Data/",arborType,"Distances_",travOrder,spurStr,"_",scoreParamStr,".csv",sep=""))
  #write.csv(distMats[[arborType]],paste(globalAlignDataDir,"ScoringV3Data/",arborType,"Distances_",travOrder,spurStr,"_",scoreParamStr,".csv",sep=""))
  print("")
}

rm(distMatsTmp)
subsetSize <- 5000
distMatsTmp <- baselineNormalizedDistancesFromScores(baselineKey,normScoreMats[[arborType]][1:subsetSize,1:subsetSize],clusterMetaData[colnames(normScoreMats[[arborType]])[1:subsetSize],"SeqLen"],numCores=11)

#tmp<-baselineNormalizedDistancesFromScores(baselineKey,normScoreMats[[arborType]][1:10,1:10],clusterMetaData[colnames(normScoreMats[[arborType]])[1:10],"SeqLen"])
#distMatFilename <- paste(clusterDir,"/alignmentDist",travOrder,".",scoreParamStr,".csv",sep="")
#write.csv(distMatDF,distMatFilename)

#axonSequences <- intersect(rownames(distMatDF),rownames(subset(clusterMetaData,ArborType=="Axon"&ExperCond=="Control")))
#axonDists <- distMatDF[axonSequences,axonSequences]
write.csv(distMats[["Axon"]],paste(globalAlignDataDir,"AxonDistances_",travOrder,spurStr,"_",scoreParamStr,".csv",sep=""))

#apicalSequences <- intersect(rownames(distMatDF),rownames(subset(clusterMetaData,ArborType=="ApicalDendrite"&ExperCond=="Control")))
#apicalDists <- distMatDF[apicalSequences,apicalSequences]
write.csv(distMats[["Apical"]],paste(globalAlignDataDir,"ApicalDistances_",travOrder,spurStr,"_",scoreParamStr,".csv",sep=""))

#dendriteSequences <- intersect(rownames(distMatDF),rownames(subset(clusterMetaData,(ArborType=="Dendrite"|ArborType=="BasalDendrite")&ExperCond=="Control")))
#dendriteDists <- distMatDF[dendriteSequences,dendriteSequences]
write.csv(distMats[["Dendrite"]],paste(globalAlignDataDir,"DendriteDistances_",travOrder,spurStr,"_",scoreParamStr,".csv",sep=""))

write.csv(distMats[["Arbor"]],paste(globalAlignDataDir,"ArborDistances_",travOrder,spurStr,"_",scoreParamStr,".csv",sep=""))
write.csv(distMats[["AllArbors"]],paste(globalAlignDataDir,"AllArborDistances_",travOrder,spurStr,"_",scoreParamStr,".csv",sep=""))

## Special case, load combination of STL and LTS, only apical and axon for now
distMats <- list()
for (arborName in arborNames[1:2]){
  tmpMat <- as.matrix(read.csv(paste(globalAlignDataDir,arborName,"Distances_STL",spurStr,"_",scoreParamStr,".csv",sep=""),row.names=1,check.names=F))
  tmpMat2 <- as.matrix(read.csv(paste(globalAlignDataDir,arborName,"Distances_LTS",spurStr,"_",scoreParamStr,".csv",sep=""),row.names=1,check.names=F))
  seqs <- intersect(rownames(tmpMat),rownames(tmpMat2))
  distMats[[arborName]] <- tmpMat[seqs,seqs] + tmpMat2[seqs,seqs]
  write.csv(distMats[[arborName]],paste(globalAlignDataDir,arborName,"Distances_",travOrder,spurStr,"_",scoreParamStr,".csv",sep=""))
}
rm(tmpMat)
rm(tmpMat2)

## For loading the distance matrices from file
distMats <- list()
for (arborName in arborNames){
#  distMats[[arborName]] <- as.matrix(read.csv(paste(globalAlignDataDir,arborName,"Distances_",travOrder,spurStr,"_",scoreParamStr,".csv",sep=""),row.names=1,check.names=F))
  # Special case when using Dendrite-only spur removal
  if (arborName in c("Axon","Apical"){
    distMats[[arborName]] <- as.matrix(read.csv(paste(globalAlignDataDir,"ScoringV2Data/",arborName,"Distances_",travOrder,"_",scoreParamStr,".csv",sep=""),row.names=1,check.names=F))
  }else{
    distMats[[arborName]] <- as.matrix(read.csv(paste(globalAlignDataDir,"ScoringV2Data/",arborName,"Distances_",travOrder,spurStr,"_",scoreParamStr,".csv",sep=""),row.names=1,check.names=F))
  }
}



###### Load arbor type lists ######
#apicals <- loadSubsetSeqs(list(list(ArborType="ApicalDendrite")),clusterMetaData,rownames(distMatDF))$Sequences
#dendrites <- loadSubsetSeqs(list(list(ArborType=c("Dendrite","BasalDendrite"))),clusterMetaData,rownames(distMatDF))$Sequences
#axons <- loadSubsetSeqs(list(list(ArborType="Axon")),clusterMetaData,rownames(distMatDF))$Sequences

selectNeurites <- list()
selectNeurites[["Apical"]] <- loadSubsetSeqs(AllApicalSpecificationsByLayer,clusterMetaData,rownames(distMats[["Apical"]]))$Sequences
selectNeurites[["Axon"]] <- loadSubsetSeqs(AllAxonSpecifications,clusterMetaData,rownames(distMats[["Axon"]]))$Sequences
selectNeurites[["Dendrite"]] <- loadSubsetSeqs(AllDendriteSpecifications,clusterMetaData,rownames(distMats[["Dendrite"]]))$Sequences

#apicals <- loadSubsetSeqs(AllApicalSpecificationsByLayer,clusterMetaData,rownames(distMatDF))$Sequences
#dendrites <- loadSubsetSeqs(AllDendriteSpecifications,clusterMetaData,rownames(distMatDF))$Sequences
#axons <- loadSubsetSeqs(AllAxonSpecifications,clusterMetaData,rownames(distMatDF))$Sequences

#### Determine number of dimensions for each arbor type ####

numDimsInitList <- list()

numDimsInitList[["G3E1Identity-STL-Apical"]] <- 8
numDimsInitList[["G3E1Identity_SpurLen2D-STL-Apical"]] <- 8
numDimsInitList[["G2E1Identity-STL-Apical"]] <- 8
numDimsInitList[["G1E1Identity-STL-Apical"]] <- 7
numDimsInitList[["G3E1Identity-LTS-Apical"]] <- 8
numDimsInitList[["G3E1Identity-Both-Apical"]] <- 8

numDimsInitList[["G3E1Identity-STL-Axon"]] <- 8
numDimsInitList[["G3E1Identity_SpurLen5-STL-Axon"]] <- 7
numDimsInitList[["G2E1Identity-STL-Axon"]] <- 7
numDimsInitList[["G1E1Identity-STL-Axon"]] <- 7
numDimsInitList[["G3E1Identity-LTS-Axon"]] <- 9
numDimsInitList[["G3E1Identity-Both-Axon"]] <- 8
  
numDimsInitList[["G3E1Identity-STL-Dendrite"]] <- 7
numDimsInitList[["G3E1Identity_SpurLen5-STL-Dendrite"]] <- 7
# unconfirmed, but likely given SpurLen5 agrees with no spur removal
numDimsInitList[["G3E1Identity_SpurLen2D-STL-Dendrite"]] <- 7
numDimsInitList[["G2E1Identity-STL-Dendrite"]] <- 7
numDimsInitList[["G1E1Identity-STL-Dendrite"]] <- 7
numDimsInitList[["G3E1Identity-LTS-Dendrite"]] <- 9
numDimsInitList[["G3E1Identity-Both-Dendrite"]] <- 9

numDimsList <- list()

numDimsList[["G3E1Identity-STL-Apical"]] <- 4
numDimsList[["G3E1Identity_SpurLen5-STL-Apical"]] <- 4 #Unconfirmed
numDimsList[["G2E1Identity-STL-Apical"]] <- 4 #Unconfirmed
numDimsList[["G1E1Identity-STL-Apical"]] <- 4 #Unconfirmed
#numDimsList[["G3E1Identity-LTS-Apical"]] <- 8
#numDimsList[["G3E1Identity-Both-Apical"]] <- 8

numDimsList[["G3E1Identity-STL-Axon"]] <- 3
numDimsList[["G3E1Identity_SpurLen5-STL-Axon"]] <- 3 #Unconfirmed
numDimsList[["G2E1Identity-STL-Axon"]] <- 3 #Unconfirmed
numDimsList[["G1E1Identity-STL-Axon"]] <- 3 #Unconfirmed
#numDimsList[["G3E1Identity-LTS-Axon"]] <- 9
#numDimsList[["G3E1Identity-Both-Axon"]] <- 8

numDimsList[["G3E1Identity-STL-Dendrite"]] <- 4
numDimsList[["G3E1Identity_SpurLen5-STL-Dendrite"]] <- 4 #Unconfirmed
numDimsList[["G3E1Identity_SpurLen2D-STL-Dendrite"]] <- 4 #Unconfirmed
numDimsList[["G2E1Identity-STL-Dendrite"]] <- 4 #Unconfirmed
numDimsList[["G1E1Identity-STL-Dendrite"]] <- 4 #Unconfirmed
#numDimsList[["G3E1Identity-LTS-Dendrite"]] <- 9
#numDimsList[["G3E1Identity-Both-Dendrite"]] <- 9

# All Arbors
numDims <- 5:8
allArborsMDSStress <- unlist(mclapply(numDims,function(m,distanceData){
  return(isoMDS(distanceData,k=m)$stress)
},distMats[["AllArbors"]],mc.cores=numCores))
save.image("ClusterAll.RData")

# Apical dendrites
numDims <- 1:15
apicalMDSStress <- as.numeric(mclapply(numDims,function(m,distanceData){
  return(isoMDS(distanceData,k=m)$stress)
},distMats[["Apical"]],mc.cores=detectCores()))
apicalMDSStressBICA4000 <- BIC.MDSStress(distMats[["Apical"]],apicalMDSStress,numDimensions=numDims,alpha=4000)
apicalMDSStressBICA10000 <- BIC.MDSStress(distMats[["Apical"]],apicalMDSStress,numDimensions=numDims,alpha=10000)
#apicalNumDims <- which.min(apicalMDSStressBICA10000)

apicalMDS15 <- isoMDS(distMats[["Apical"]],k=15)
apicalMDS_SD <- apply(apicalMDS15$points,2,sd)

# G3E1 STL: d=8, stress=12.6%
# G3E1 STL SpurLen5: d=8, stress=11.9% (approx optimal)
# G2E1 STL: d=8, stress=11.9% (optimal between 3-9)
# G1E1 STL: d=7, stress=12.1%
# G3E1 LTS: d=8, stress=13.1% (d=7 - 14.6%, d=9 - 12.1%)
# G3E1 Both: d=8, stress=12.5% (d=7 - 13.8%, d=9 - 11.6%)
apicalNumDims <- 8 # STL (?%; SpurLen5: 12.5%, optimal somewhere from 3-9)
apicalNumDims <- 11 # LTS; 11 is local minimum w/ 18.7%; 28% @ 4, Something from 4-6 might be reasonable


# Axons
log(BIC.MDS(distMats[["Axon"]],alpha=400),numDimensions=1:10,10)
axonMDSBIC_A400Stress <- BIC.MDS(distMats[["Axon"]],numDimensions=1:8,alpha=400,measure="Stress")
axonMDSBIC_A400Error <- BIC.MDS(distMats[["Axon"]],numDimensions=1:8,alpha=400,measure="Error")
axonMDSBIC_A400Norm <- BIC.MDS(distMats[["Axon"]],numDimensions=1:8,alpha=400,measure="NormError")
log(axonMDSBIC_A400Norm,10)

numDims <- 1:14
axonMDSStress <- as.numeric(mclapply(numDims,function(m,distanceData){
  return(isoMDS(distanceData,k=m)$stress)},
    distMats[["Axon"]],mc.cores=detectCores()))
axonMDSStressBICA4000 <- BIC.MDSStress(distMats[["Axon"]],axonMDSStress,numDimensions=numDims,alpha=4000)
axonMDSStressBICA10000 <- BIC.MDSStress(distMats[["Axon"]],axonMDSStress,numDimensions=numDims,alpha=10000)
BIC.MDSStress(distMats[["Axon"]],axonMDSStress,numDimensions=numDims,alpha=30000)
axonNumDims <- which.min(axonMDSStressBICA10000)

axonMDS15 <- isoMDS(distMats[["Axon"]],k=15)
axonMDS_SD <- apply(axonMDS15$points,2,sd)
axonMDSStress <- append(axonMDSStress,axonMDS15$stress)

# G3E1 STL: d=8, stress=13.9% (13.0% at d=9, optimal between 5-8, 15.1% at d=7)
# G3E1 STL SpurLen5: d=7, stress=14.5% (optimal between 7-9)
# G2E1 STL: d=7, stress=14.5% (optimal between 5-8)
# G1E1 STL: d=7, stress=13.5%
# G3E1 LTS: d=9, stress=13.5% (d=8 - 15.2%, d=10 - 12.7%)
# G3E1 Both: d=8, stress=13.8% (d=7 - 15.0%, d=9 - 12.8%)
axonNumDims <- 7 # STL (?%; SpuLen5: 15.5%; optimal somewhere from 4-6)
axonNumDims <- 10 # LTS; 18%; 5 gives 28%


# Dendrites
#log(BIC.MDS(distMatDF[dendrites,dendrites],dendriteMDS,alpha=400),10)

MDSSumSquareError(distMatDF[dendrites,dendrites],percent=F)
MDSSumSquareError(distMatDF[dendrites,dendrites],percent=T)

stress2(distMatDF[dendrites,dendrites],dendriteMDS[dendrites,])

numDims <- 1:8
dendriteMDSStress <- mclapply(numDims,function(m,distanceData){
  return(isoMDS(distanceData,k=m)$stress)},
    distMats[["Dendrite"]],mc.cores=detectCores())
dendriteMDSStressBICA4000 <- BIC.MDSStress(distMats[["Dendrite"]],dendriteMDSStress,numDimensions=numDims,alpha=4000)
dendriteMDSStressBICA10000 <- BIC.MDSStress(distMats[["Dendrite"]],dendriteMDSStress,numDimensions=numDims,alpha=10000)
dendriteNumDims <- which.min(dendriteMDSStressBICA10000)

# G3E1 STL: d=7, stress=15.1% (16.8 at 6, 13.8 at 9)
# G3E1 STL SpurLen5: d=7, stress=14.3% (optimal between 6-8; 6 has 15.7%)
# G2E1 STL: d=7, stress=14.6% (16.1% at 6, 13.4% at 8, optimal between 5-9)
# G1E1 STL: ?
# G3E1 LTS: 9, stress=12.6% (13.7% at 8, 11.7% at 10)
# G3E1 Both: 9, stress=12.5% (13.6% at 8, ? at 10)
dendriteNumDims <- 7 #STL
dendriteNumDims <- 8 #LTS; 19.1%, 24.8% at 4

dendriteMDSStress10_12 <- as.numeric(mclapply(10:12,function(m,distanceData){
  return(isoMDS(distanceData,k=m)$stress)},
  distMats[["Dendrite"]],mc.cores=detectCores()))

save.image("ClusterMDSTesting.RData")

dendriteMDSStress13_14 <- as.numeric(mclapply(13:14,function(m,distanceData){
  return(isoMDS(distanceData,k=m)$stress)},
  distMats[["Dendrite"]],mc.cores=detectCores()))

save.image("ClusterMDSTesting.RData")

dendriteMDSStress <- append(append(dendriteMDSStress,dendriteMDSStress10_12),dendriteMDSStress13_14)
dendriteMDS15 <- isoMDS(distMats[["Dendrite"]],k=15)
dendriteMDS_SD <- apply(dendriteMDS15$points,2,sd)
dendriteMDSStress <- append(dendriteMDSStress,dendriteMDS15$stress)

save.image("ClusterMDSTesting.RData")

arborMDS_Stats <- data.frame(ApicalStress=apicalMDSStress,ApicalSD=apicalMDS_SD,AxonStress=axonMDSStress,AxonSD=axonMDS_SD,DendriteStress=dendriteMDSStress,DendriteSD=dendriteMDS_SD)
write.csv(arborMDS_Stats,paste("MDS_scree_data.csv",sep=""))

## See if first 2 dimensions remain approximately the same regardless of total # dims
for (arborName in arborNames){
  distMats[[arborName]] <- as.matrix(read.csv(paste(globalAlignDataDir,arborName,"Distances_",travOrder,"_",spurStr,"_",scoreParamStr,".csv",sep=""),row.names=1,check.names=F))
  for (numDims in c(2,4,8)){
    key <- paste(travOrder,arborType,spurLen,scoreParamStr,numDims)
    mdsFilename <- paste(globalAlignDataDir,arborName,"_",travOrder,spurStr,"_",scoreParamStr,"_MDS",numDims,".csv",sep="")
    if (file.exists(mdsFilename)){
      featMat <- read.csv(mdsFilename,row.names=1)
    }else{
      arborMDSAll[[arborName]] <- isoMDS(distMats[[arborName]],k=numDims)
      featMat <- arborMDSAll[[arborName]]$points
      arborMDS[[key]] <- featMat
      write.csv(featMat,mdsFilename,row.names=T)
    }
    png(paste(globalAlignDataDir,arborName,"_",travOrder,spurStr,"_",scoreParamStr,"_MDS",numDims,"_2vs1.png",sep=""))
    subsetObj <- loadSubsetSeqs(specificationsByArbor[[arborName]],clusterMetaData,candidates=rownames(featMat))
    featMat <- featMat[subsetObj$Sequences,]
    groupIds <- subsetObj$GroupIds
    plot(featMat[,1:2],col=groupIds)
    dev.off()
  }  
}

#### MDS and Normalize (transform) ####
fullDataMDS <- isoMDS(distMatDF,k=14)
#  write.csv(fullDataMDS,paste(globalAlignDataDir,"AllArbors_",travOrder,spurStr,"_",scoreParamStr,"_MDS",14,".csv",sep=""),row.names=T)

for (arborName in arborNames){
  # Set numDims for each arbor type assuming scoring, spur, and traversal parameters
  numDimsInitList[[arborName]] <- numDimsInitList[[paste(scoreParamStr,spurStr,"-",travOrder,"-",arborName,sep="")]]
  numDimsList[[arborName]] <- numDimsList[[paste(scoreParamStr,spurStr,"-",travOrder,"-",arborName,sep="")]]
  #  numDimsInitList[[arborName]] <- 8 # Set all to the same for full comparison
}

arborMDSAll <- list()
arborMDS <- list()
if (systemType == "PC"){
  for (arborName in arborNames){
    arborMDS[[arborName]] <- read.csv(paste(globalAlignDataDir,arborName,"_",travOrder,spurStr,"_",scoreParamStr,"_MDS",numDimsInitList[[arborName]],".csv",sep=""),row.names=1)
  }
#  arborMDS[["Apical"]] <- read.csv(paste(globalAlignDataDir,"Apical_",travOrder,spurStr,"_",scoreParamStr,"_MDS",apicalNumDimsInit,".csv",sep=""),row.names=1)
#  arborMDS[["Axon"]] <- read.csv(paste(globalAlignDataDir,"Axon_",travOrder,spurStr,"_",scoreParamStr,"_MDS",axonNumDimsInit,".csv",sep=""),row.names=1)
#  arborMDS[["Dendrite"]] <- read.csv(paste(globalAlignDataDir,"Dendrite_",travOrder,spurStr,"_",scoreParamStr,"_",scoreParamStr,"_MDS",dendriteNumDimsInit,".csv",sep=""),row.names=1)
}else{
  for (arborName in arborNames){
    arborMDSAll[[arborName]] <- isoMDS(distMats[[arborName]],k=numDimsInitList[[arborName]])
    arborMDS[[arborName]] <- arborMDSAll[[arborName]]$points
    write.csv(arborMDS[[arborName]],paste(globalAlignDataDir,arborName,"_",travOrder,spurStr,"_",scoreParamStr,"_MDS",numDimsInitList[[arborName]],".csv",sep=""),row.names=T)
  }

  #arborSampleMDSAll <- isoMDS(distMatsSubset,k=15)
  #write.csv(arborSampleMDSAll,paste(globalAlignDataDir,"ArborSample_",travOrder,spurStr,"_",scoreParamStr,"_MDS",15,".csv",sep=""),row.names=T)
  #arborAllMDS <- read.csv(paste(globalAlignDataDir,"ArborSample_",travOrder,spurStr,"_",scoreParamStr,"_MDS",15,".csv",sep=""),row.names=1)
  arborAllMDS <- read.csv(paste(globalAlignDataDir,"AllArbors_",travOrder,spurStr,"_",scoreParamStr,"_MDS10.csv",sep=""),row.names=1)
  
#  apicalMDSAll <- isoMDS(distMats[["Apical"]],k=apicalNumDimsInit)
#  apicalMDS <- apicalMDSAll$points
#  write.csv(apicalMDS,paste(globalAlignDataDir,"Apical_",travOrder,spurStr,"_",scoreParamStr,"_MDS",apicalNumDimsInit,".csv",sep=""),row.names=T)
  
#  dendriteMDSAll <- isoMDS(distMats[["Dendrite"]],k=dendriteNumDimsInit)
#  dendriteMDS <- dendriteMDSAll$points
#  write.csv(dendriteMDS,paste(globalAlignDataDir,"Dendrite_",travOrder,spurStr,"_",scoreParamStr,"_MDS",dendriteNumDimsInit,".csv",sep=""),row.names=T)
  
#  axonMDSAll <- isoMDS(distMats[["Axon"]],k=axonNumDimsInit)
#  axonMDS <- axonMDSAll$points
#  write.csv(axonMDS,paste(globalAlignDataDir,"Axon_",travOrder,spurStr,"_",scoreParamStr,"_MDS",axonNumDimsInit,".csv",sep=""),row.names=T)
}

### Look at all arbor types toghether

arborColors <- rep(1,nrow(arborAllMDS))
arborColors[masterMetaData[rownames(arborAllMDS),"ArborType"] == "Axon"] <- 2
arborColors[masterMetaData[rownames(arborAllMDS),"ArborType"] == "ApicalDendrite"] <- 3
plot(arborAllMDS[,1:3],col=arborColors)

sharedSeqs <- intersect(rownames(arborAllMDS),rownames(morphometricsXformed))
cor(cbind(arborAllMDS[sharedSeqs,1:3],morphometricsXformed[sharedSeqs,c("Stems","Bifurcations","BranchOrder","PartitionAsymmetry")]))^2

cor(cbind(arborAllMDS[,1:3],morphometricsXformed[rownames(arborAllMDS),c("Stems","Bifurcations","BranchOrder","PartitionAsymmetry")]))^2

x11()
dens2D <- kde2d(arborAllMDS[,1],arborAllMDS[,2])
kde2dplot(dens2D)

## Cluster all arbor types together
specifications <- list(Axon=list(ArborType="Axon",ExperCond="Control"),
                       Apical=list(ArborType="ApicalDendrite",ExperCond="Control"),
                       Dendrite=list(ArborType="Dendrite",ExperCond="Control"))
subsetObj <- loadSubsetSeqs(specifications,clusterMetaData,candidates=rownames(arborAllMDS))
featMat <- arborAllMDS[subsetObj$Sequences,1:4]

# Limit EM analysis just to chosen sequences???
#featMat <- featMat[subsetObj$Sequences,]
nonSubsetSeqs = names(featMat)[!rownames(featMat)%in%subsetObj$Sequences]
groupIds <- subsetObj$GroupIds
allGroupIds <- rep(length(specifications)+1,nrow(featMat)) # Includes non-specified neurons by adding an extra group "All Others"
names(allGroupIds) <- rownames(featMat)
allGroupIds[subsetObj$Sequences] <- subsetObj$GroupIds
subsetObj$AllGroupIds <- allGroupIds
numGroups <- length(specifications)

featEM <- Mclust(featMat,G=1:12)
numClusters <- featEM$G

groupNames <- paste(names(specifications)," (",sapply(subsetObj$SeqsByGroup,length),")",sep="")
arborClusterResults <- clusteringStatisticalAnalysis(featMat,featEM,subsetObj,groupNames)

#filenameBase <- paste(clusterDir,"/Subresults/",arborType,travOrder,"-All",spurStr,"_",scoreParamStr,"_MDS",numDims,"of",numDimsInit,sep="")
filenameBase <- paste(clusterDir,"/Subresults/",arborType,travOrder,spurStr,"_",scoreParamStr,"_MDS",ncol(featMat),"of",ncol(arborAllMDS),"-G",featEM$G,featEM$modelName,sep="")

write.csv(arborClusterResults$Observed,paste(filenameBase,"_Observed.csv",sep=""))
write.csv(round(arborClusterResults$LogP,2),paste(filenameBase,"_LogPVals",".csv",sep=""))

png(paste(clusterDir,"/Subresults/ClusterBIC_AllArbors",spurStr,scoreParamStr,"_",numDims,"of",numDimsInit,"Dims",specifiedNeuronsStr,setGStr,".png",sep=""))
plot(featEM,what="BIC")
dev.off()

png(paste(clusterDir,"/Subresults/ClusterClass_AllArbors",spurStr,scoreParamStr,"_",numDims,"of",numDimsInit,"Dims",specifiedNeuronsStr,setGStr,specifiedModel,".png",sep=""))
plot(featEM,what="classification",dimens=1:2)
dev.off()

png(paste(clusterDir,"/Subresults/ClusterClass_AllArbors",spurStr,scoreParamStr,"_",numDims,"of",numDimsInit,"Dims",specifiedNeuronsStr,setGStr,specifiedModel,"-ViewDims34.png",sep=""))
plot(featEM,what="classification",dimens=3:4)
dev.off()

###

# Exploring distribution shape of 1st and subsequent dimensions
plotDist(fullDataMDS,"All Data MDS1 Dist","MDS1Dist-AllData.png")
plotDist(apicalMDS[,1],"Apical MDS1 Dist","MDS1Dist-Apical.png")
plotDist(axonMDS$points[,1],"Axon MDS1 Dist","MDS1Dist-Axon.png")
plotDist(dendriteMDS$points[,1],"Dendrite MDS1 Dist","MDS1Dist-Dendrite.png")

plotDist <- function(mydata,title="",filename=NA){
  if (!is.na(filename)){
    png(filename)
  }
  hist(mydata,freq=F)
  lines(density(mydata))
  mdsMean <- mean(mydata)
  mdsSd <- sd(mydata)
  xs <- diff(range(mydata))*(0:200)/200 + min(mydata)
  ys <- dnorm(xs,mdsMean,mdsSd)
  lines(xs,ys,col=2,main=title)
  if (!is.na(filename)){
    dev.off()
  }
}

#### Cluster everything using mclust ####

mycolors <- c("black","red","green","blue","darkorange","purple1","cyan","brown","magenta","green4","yellow2","pink1","gray70","aquamarine","purple4","navy")
emResults <- list()

#mymodels <- c("EII","VII","EEI","VEI","EVI","VVI","EEE","EEV","VEV","VVV")
mymodels <- c("EII","VII","EEI","VEI","EEE")

specificationsByArbor <- list(Axon=AllAxonSpecifications,Apical=AllApicalSpecificationsByLayer,Dendrite=AllDendriteSpecifications)

#AllNeuriteSpecificationsByArbor
#CorticalApicalSpecificationsByLayer

# Choosing one particular dataset for now, could also combine morphometrics (deprecated)
featMat <- apicalMDS
specifications <- AllApicalSpecificationsByLayer
arborType <- "Apical"
#numDimsInit <- apicalNumDimsInit
numDimsInit <- ncol(featMat)
numDims <- numDimsInit
#numDims <- apicalNumDims
#numDims <- 3

featMat <- axonMDS
specifications <- AllAxonSpecifications
arborType <- "Axon"
#numDimsInit <- axonNumDimsInit
numDimsInit <- ncol(featMat)
numDims <- numDimsInit
#numDims <- axonNumDims
#numDims <- 2

featMat <- dendriteMDS
specifications <- AllDendriteSpecifications
arborType <- "Dendrite"
#numDimsInit <- dendriteNumDimsInit
numDimsInit <- ncol(featMat)
numDims <- numDimsInit
#numDims <- dendriteNumDims
#numDims <- 5

# Axon - Interneurons by subtype
sourceArborType <- "Axon"
featMat <- arborMDS[[sourceArborType]]
featMat <- featMat[rownames(featMat)%in%rownames(subset(usedMetaData,ArborType=="Axon"&BrainRegion=="Neocortex"&Class1=="Interneuron")),]
specifications <- CorticalInterneuronAxonSpecifications
arborName <- "InterneuronAxon"
arborType <- arborName
numDimsInit <- ncol(featMat)
numDims <- numDimsList[[sourceArborType]]


# Apical - Neocortical by Layer
#featMat <- apicalMDS
sourceArborType <- "Apical"
featMat <- arborMDS[[sourceArborType]]
featMat <- featMat[rownames(featMat)%in%rownames(subset(usedMetaData,ArborType=="ApicalDendrite"&BrainRegion=="Neocortex"&Class2=="Pyramidal cell")),]
specifications <- CorticalApicalSpecificationsByLayer
arborName <- "NCApical"
arborType <- arborName
numDimsInit <- ncol(featMat)
numDims <- numDimsList[[sourceArborType]]

setG <- NA
setGStr <- ""
if (!is.na(setG)){
  setGStr <- paste("_",setG,sep="")
}
specifiedNeurons <- F
specifiedNeuronsStr <- ""
if (specifiedNeurons){
  specifiedNeuronsStr <- "_SpecNeurons"
}
specifiedModel <- "EII"

clusterResults <- list()
for (arborName in arborNames){

  arborType <- arborName
  
  featMat <- arborMDS[[arborType]]
  specifications <- specificationsByArbor[[arborName]]
  #arborType <- "Dendrite"
  #numDimsInit <- dendriteNumDimsInit
  numDimsInit <- ncol(featMat)
  #numDims <- numDimsInit
  numDims <- numDimsList[[arborType]]
  
  # Set the relevant groups and get rid of feature vectors associated with unrepresented neurites
  subsetObj <- loadSubsetSeqs(specifications,clusterMetaData,candidates=rownames(featMat))
  if (specifiedNeurons){
    featMat <- featMat[subsetObj$Sequences,]
  }
  
  # Limit EM analysis just to chosen sequences???
  #featMat <- featMat[subsetObj$Sequences,]
  nonSubsetSeqs = names(featMat)[!rownames(featMat)%in%subsetObj$Sequences]
  groupIds <- subsetObj$GroupIds
  allGroupIds <- rep(length(specifications)+1,nrow(featMat)) # Includes non-specified neurons by adding an extra group "All Others"
  names(allGroupIds) <- rownames(featMat)
  allGroupIds[subsetObj$Sequences] <- subsetObj$GroupIds
  subsetObj$AllGroupIds <- allGroupIds
  numGroups <- length(specifications)

  # Remove outliers???
  if (numDims == numDimsInit){
    key <- paste(travOrder,arborType,spurStr,scoreParamStr,specifiedNeuronsStr,specifiedModel,numDimsInit)
  }else{
    key <- paste(travOrder,arborType,spurStr,scoreParamStr,specifiedNeuronsStr,specifiedModel,numDimsInit,numDims)
  }
  if (!is.na(setG)){
    key <- paste(key," G",setG,sep="")
  }
  
  # Run EM (picking best model)
  if (key %in% names(emResults)){
    featEM <- emResults[[key]]
  }else{
    if (is.na(setG)){
      if (specifiedModel == ""){
        featEM <- Mclust(featMat[,1:numDims],G=1:(numGroups+2))
      }else{
        featEM <- Mclust(featMat[,1:numDims],G=1:(numGroups+2),modelNames=specifiedModel)
      }
    }else{
      if (specifiedModel == ""){
        featEM <- Mclust(featMat[,1:numDims],G=setG)
      }else{
        featEM <- Mclust(featMat[,1:numDims],G=setG,modelNames=specifiedModel)
      }
    }
    emResults[[key]] <- featEM
  }
  numClusters <- featEM$G
  
  groupNames <- paste(names(specifications)," (",sapply(subsetObj$SeqsByGroup,length),")",sep="")
  arborClusterResults <- clusteringStatisticalAnalysis(featMat,featEM,subsetObj,groupNames)
  clusterResults[[arborName]] <- arborClusterResults

  #filenameBase <- paste(clusterDir,"/Subresults/",arborType,travOrder,"-All",spurStr,"_",scoreParamStr,"_MDS",numDims,"of",numDimsInit,sep="")
  filenameBase <- paste(clusterDir,"/Subresults/",arborType,travOrder,spurStr,"_",scoreParamStr,specifiedNeuronsStr,"_MDS",numDims,"of",numDimsInit,"G",featEM$G,specifiedModel,sep="")

  #write.csv(arborClusterResults$Observed,paste(filenameBase,"_Observed.csv",sep=""))
  #write.csv(round(arborClusterResults$Expected,0),paste(filenameBase,"_Expected.csv",sep=""))
  #write.csv(round(arborClusterResults$Difference,1),paste(filenameBase,"_Diff.csv",sep=""))
  #write.csv(round(arborClusterResults$Ratio,2),paste(filenameBase,"_Ratio.csv",sep=""))
  write.csv(round(arborClusterResults$LogP,2),paste(filenameBase,"_LogPVals",".csv",sep=""))
  #write.csv(round(arborClusterResults$LogP,2),paste(filenameBase,"EEI_LogPVals",".csv",sep=""))
  write.csv(round(arborClusterResults$Residuals,3),paste(filenameBase,"_Residuals",".csv",sep=""))
  
  png(paste(clusterDir,"/Subresults/ClusterBIC_",arborType,spurStr,scoreParamStr,"_",numDims,"of",numDimsInit,"Dims",specifiedNeuronsStr,setGStr,".png",sep=""))
  plot(featEM,what="BIC")
  dev.off()
  
  png(paste(clusterDir,"/Subresults/ClusterClass_",arborType,spurStr,scoreParamStr,"_",numDims,"of",numDimsInit,"Dims",specifiedNeuronsStr,setGStr,specifiedModel,".png",sep=""))
  plot(featEM,what="classification",dimens=1:2)
  dev.off()
  
  png(paste(clusterDir,"/Subresults/ClusterClass_",arborType,spurStr,scoreParamStr,"_",numDims,"of",numDimsInit,"Dims",specifiedNeuronsStr,setGStr,specifiedModel,"-ViewDims34.png",sep=""))
  plot(featEM,what="classification",dimens=3:4)
  dev.off()
  
}

#### Cluster distances from center and each other ####
#arborName <- ?
arborType <- arborName

featMat <- arborMDS[[sourceArborType]]
specifications <- specificationsByArbor[[arborName]]
numDimsInit <- ncol(featMat)
numDims <- numDimsList[[sourceArborType]]

# Set the relevant groups and get rid of feature vectors associated with unrepresented neurites
subsetObj <- loadSubsetSeqs(specifications,clusterMetaData,candidates=rownames(featMat))
# Limit EM analysis just to chosen sequences???
#featMat <- featMat[subsetObj$Sequences,]
nonSubsetSeqs = names(featMat)[!rownames(featMat)%in%subsetObj$Sequences]
groupIds <- subsetObj$GroupIds
numGroups <- length(specifications)

if (numDims == numDimsInit){
  key <- paste(travOrder,arborType,spurLen,scoreParamStr,numDimsInit)
}else{
  key <- paste(travOrder,arborType,spurLen,scoreParamStr,numDimsInit,numDims)
}

# Run EM (picking best model)
if (key %in% names(emResults)){
  featEM <- emResults[[key]]
}else{
  featEM <- Mclust(featMat[,1:numDims],G=1:(numGroups+2))
  numClusters <- featEM$G
  emResults[[key]] <- featEM
}

# Distances of clusters from center
center <- apply(featMat,2,mean)
clustCenters <- c()
clustScatters <- c()
rawDists <- c()
dists <- c()
for (i in 1:featEM$G){
  clustCenters[i] <- apply(featMat[featEM$classification==i,],2,mean)
  clustScatters[i] <- sqrt(sum(apply(featMat[featEM$classification==i,],2,var)))
  rawDists[i] <- sqrt(sum((clustCenters[i]-center)^2))
  dists[i] <- rawDists[i]/clustScatters[i]
}
clustDistMat <- matrix(0,nrow=featEM$G,ncol=featEM$G)
for (i in 1:(featEM$G-1)){
  for (j in (i+1):featEM$G){
    clustDistMat[i,j] <- sqrt(sum((clustCenters[i]-clustCenters[j])^2))/(clustScatters[i]+clustScatters[j])
  }
}
round(clustDistMat,2)

#### Evaluate cluster quality ####
#ClusterQUalityByParams
paramSets <- list(list(travOrder="STL",gapParams="G3E1",spurLen=0),
                     list(travOrder="STL",gapParams="G3E1",spurLen=5),
                     list(travOrder="LTS",gapParams="G3E1",spurLen=0),
                     list(travOrder="Both",gapParams="G3E1",spurLen=0),
                     list(travOrder="STL",gapParams="G2E1",spurLen=0),
                     list(travOrder="STL",gapParams="G1E1",spurLen=0))

paramSets <- list(list(travOrder="STL",gapParams="G2E1",spurLen=0),
                  list(travOrder="STL",gapParams="G1E1",spurLen=0))
paramSets <- list(list(travOrder="STL",gapParams="G3E1",spurLen=5))
paramSets <- list(list(travOrder="LTS",gapParams="G3E1",spurLen=0),
                  list(travOrder="Both",gapParams="G3E1",spurLen=0))

require(clv)
clusterStatsDF <- NULL
for (arborName in arborNames){
  arborType <- arborName
  specifications <- specificationsByArbor[[arborName]]
  
  results <- mclapply(paramSets,evaluateSequenceParameters,arborType,specifications,mc.cores=10)
  for (result in results){
    if (!is.null(result)){
      if (is.null(clusterStatsDF)){
        clusterStatsDF <- result
      }else{
        clusterStatsDF <- rbind(clusterStatsDF,result)
      }
    }
  }
  #for (paramSet in paramSets){
  #}  
}
evaluateSequenceParameters <- function(paramSet,arborType,specifications){
  travOrder <- paramSet$travOrder
  gapParams <- paramSet$gapParams
  gapParamStr <- paste(gapParams,"Identity",sep="")
  spurLen <- paramSet$spurLen
  if (spurLen > 0){
    spurStr <- paste("_SpurLen",spurLen,sep="")
  }else{ spurStr <- "" }
  numDimsInit <- numDimsInitList[[paste(gapParamStr,spurStr,"-",travOrder,"-",arborName,sep="")]]
  mdsFilename <- paste(globalAlignDataDir,arborType,"_",travOrder,spurStr,"_",gapParamStr,"_MDS",numDimsInit,".csv",sep="")
  if (!file.exists(mdsFilename)){
    print(paste("File does not exist:",mdsFilename))
    return(NULL)
  }else{
    featMat <- read.csv(mdsFilename,row.names=1)
    #featMat <- scale(read.csv(mdsFilename,row.names=1))
    print(paste("Working on ",gapParamStr,spurStr,"-",travOrder,"-",arborType," D",numDimsInit,sep=""))
    
    # Set the relevant groups and get rid of feature vectors associated with unrepresented neurites
    subsetObj <- loadSubsetSeqs(specifications,clusterMetaData,candidates=rownames(featMat))
    featMat <- featMat[subsetObj$Sequences,]
    groupIds <- subsetObj$GroupIds
    numGroups <- length(specifications)
    
    clusterStatsDFInner <- NULL
    for (numDims in 2:numDimsInit){
    
      if (numDims == numDimsInit){
        key <- paste(travOrder,arborType,spurLen,gapParamStr,numDimsInit)
      }else{
        key <- paste(travOrder,arborType,spurLen,gapParamStr,numDimsInit,numDims)
      }
      
      # Run EM (picking best model)
      if (key %in% names(emResults)){
        featEM <- emResults[[key]]
      }else{
        featEM <- Mclust(featMat[,1:numDims],G=1:(numGroups+2))
        numClusters <- featEM$G
        #featEM$bicAdj <- adjustedBIC(featMat[,1:numDims])
        emResults[[key]] <- featEM
      }
      
      # Analyze featEM clusters
      clsObject <- cls.scatt.data(featMat,featEM$classification)
      #db <- clv.Davies.Bouldin(clsObject,intracls=c("complete","average","centroid"),intercls=c("complete","average"))
      db <- clv.Davies.Bouldin(clsObject,intracls=c("average","centroid"),intercls=c("average","centroid"))
      
      # Analyze relationship between clusters and groups
      clusterResults <- clusteringStatisticalAnalysis(featMat,featEM,subsetObj,names(specifications))
#        clusterResults$LogP > 1.301
      
      # Assemble data frame
      newDF <- data.frame(Arbor=arborType,Traversal=travOrder,ScoreParam=gapParams,Spur=spurLen,
                          InitialDims=numDimsInit,D=numDims,Clusts=featEM$G,Model=featEM$modelName,
                          #BIC=adjustedBIC,
                          AvgUncert=mean(featEM$uncertainty),MedianUncert=median(featEM$uncertainty),Uncert95P=percentile(.95,featEM$uncertainty),
                          DBAvg=db[1,1],DBCentroid=db[2,2])
#                            DBComCom=db[1,1],DBComAvg=db[1,2],DBComCent=db[1,3],
#                            DBAvgCom=db[2,1],DBAvgAvg=db[2,2],DBAvgCent=db[2,3])
      if (is.null(clusterStatsDFInner)){
        clusterStatsDFInner <- newDF
      }else{
        clusterStatsDFInner <- rbind(clusterStatsDFInner,newDF)
      }
    }
    return(clusterStatsDFInner)
  }
}

write.csv(clusterStatsDF,paste(clusterDir,"ClusterStatsAnalysisByParams-FixedDendSpur.csv",sep=""))
write.csv(clusterStatsDF,paste(clusterDir,"ClusterStatsAnalysisByParams-DendLTS&Both.csv",sep=""))
write.csv(clusterStatsDF,paste(clusterDir,"ClusterStatsAnalysisByParams-FixedG23.csv",sep=""))
write.csv(clusterStatsDF,paste(clusterDir,"ClusterStatsAnalysisByParams_2.0.csv",sep=""))
#write.csv(clusterStatsDF,paste(clusterDir,"ClusterStatsAnalysisByParams-Scaled.csv",sep=""))

# Output EM BIC charts for select cases
bicPlotParamSets <- list(list(arbor="Dendrite",travOrder="STL",gapParams="G3E1",spurLen=0,numDims=2),
                  list(arbor="Dendrite",travOrder="STL",gapParams="G2E1",spurLen=0,numDims=2),
                  list(arbor="Dendrite",travOrder="STL",gapParams="G2E1",spurLen=0,numDims=4),
                  list(arbor="Dendrite",travOrder="STL",gapParams="G3E1",spurLen=0,numDims=2),
                  list(arbor="Apical",travOrder="STL",gapParams="G1E1",spurLen=0,numDims=2),
                  list(arbor="Apical",travOrder="STL",gapParams="G1E1",spurLen=0,numDims=7),
                  list(arbor="Apical",travOrder="Both",gapParams="G3E1",spurLen=0,numDims=5),
                  list(arbor="Axon",travOrder="STL",gapParams="G2E1",spurLen=0,numDims=3),
                  list(arbor="Axon",travOrder="Both",gapParams="G3E1",spurLen=0,numDims=2),
                  list(arbor="Axon",travOrder="STL",gapParams="G1E1",spurLen=0,numDims=4),
                  list(arbor="Axon",travOrder="STL",gapParams="G2E1",spurLen=0,numDims=5),
                  list(arbor="Axon",travOrder="STL",gapParams="G2E1",spurLen=0,numDims=5)
                         )

for (paramSet in bicPlotParamSets){
  arborType <- paramSet$arbor
  arborName <- arborType
  specifications <- specificationsByArbor[[arborName]]
  travOrder <- paramSet$travOrder
  gapParams <- paramSet$gapParams
  gapParamStr <- paste(gapParams,"Identity",sep="")
  numDims <- paramSet$numDims
  spurLen <- paramSet$spurLen
  if (spurLen > 0){
    spurStr <- paste("_SpurLen",spurLen,sep="")
  }else{ spurStr <- "" }
  numDimsInit <- numDimsInitList[[paste(gapParamStr,spurStr,"-",travOrder,"-",arborType,sep="")]]
  mdsFilename <- paste(globalAlignDataDir,arborType,"_",travOrder,spurStr,"_",gapParamStr,"_MDS",numDimsInit,".csv",sep="")
  if (!file.exists(mdsFilename)){
    print(paste("File does not exist:",mdsFilename))
  }else{
    featMat <- read.csv(mdsFilename,row.names=1)
    #featMat <- scale(read.csv(mdsFilename,row.names=1))
    print(paste("Working on ",gapParamStr,spurStr,"-",travOrder,"-",arborType," D",numDimsInit,sep=""))
    
    # Set the relevant groups and get rid of feature vectors associated with unrepresented neurites
    subsetObj <- loadSubsetSeqs(specifications,clusterMetaData,candidates=rownames(featMat))
    featMat <- featMat[subsetObj$Sequences,]
    groupIds <- subsetObj$GroupIds
    numGroups <- length(specifications)
    
    if (numDims == numDimsInit){
      key <- paste(travOrder,arborType,spurLen,gapParamStr,numDimsInit)
    }else{
      key <- paste(travOrder,arborType,spurLen,gapParamStr,numDimsInit,numDims)
    }
    
    # Run EM (picking best model)
    if (key %in% names(emResults)){
      featEM <- emResults[[key]]
    }else{
      featEM <- Mclust(featMat[,1:numDims],G=1:(numGroups+2))
      numClusters <- featEM$G
      emResults[[key]] <- featEM
    }
    
    png(paste(clusterDir,"/BIC_Plots/",arborType,travOrder,spurStr,"_",gapParams,"_D",numDims,"of",numDimsInit,".png",sep=""),height=500,width=500)
    plot(featEM,what="BIC")
    dev.off()
  }
}

##### Alternative statistical Evaluation (and mosaic plot) #####
numGroups <- nrow(clusterResults$Observed)
numClusters <- ncol(clusterResults$Observed)
Cs<-factor(rep(paste("C",1:numClusters,sep=""),each=numGroups))
Gs<-factor(rep(paste("G",1:numGroups,sep=""),numClusters))
model1 <- glm(as.vector(clusterResults$Observed)~Cs*Gs,poisson)
model2 <- glm(as.vector(clusterResults$Observed)~Cs+Gs,poisson)
anova(model1,model2,test="Chi")
summary(model1)
mosaicplot(clusterResults$Observed,shade=T)

#### Plots ####

mycolors <- c("black","red","green","blue","darkorange","purple1","cyan","brown","magenta",
               "green4","yellow2","pink1","gray70","aquamarine","purple4","navy")
mycolors <- c("black","blue","green","red","cyan","yellow2","gray")
mycolors <- c("black","","","blue","green","red","","cyan","brown","magenta","","","gray70")
mycolors <- c("black","","","blue","green","red","","cyan","","","","","magenta")
mycolors <- c("","black","blue","","","","green","","red","cyan","magenta","purple","")
mycolors <- c("black","blue","green","red","cyan","orange","gray")

# Significant
arborClusterResults <- clusteringStatisticalAnalysis(featMat,featEM,subsetObj,groupNames=names(specificiations))
hasSignificant <- which(apply(abs(arborClusterResults$LogP) > 1.301,1,sum) > 0)
#hasSignificant <- hasSignificant[-5]
inSignificantGroup <- sapply(groupIds,function(x){return(x%in%hasSignificant)})
notSignificantGroup <- sapply(groupIds,function(x){return(!x%in%hasSignificant)})

library(car)
# Plot on MDS axes
x11()
#parms <- featEM_EEE$parameters
parms <- featEM$parameters
dims <- c(1,2)
plot(featMat[,dims[1]],featMat[,dims[2]],col="white",cex=.55,pch=16,xlab=paste("Dimension",dims[1]),ylab=paste("Dimension",dims[2]),las=1)
text(featMat[inSignificantGroup,dims[1]],featMat[inSignificantGroup,dims[2]],labels=as.character(as.hexmode(featEM$classification[inSignificantGroup])),col=mycolors[groupIds[inSignificantGroup]],cex=.7)
#text(featMat[notSignificantGroup,dims[1]],featMat[notSignificantGroup,dims[2]],labels=as.character(as.hexmode(featEM$classification[notSignificantGroup])),col=mycolors[groupIds[notSignificantGroup]],cex=.7)
#text(featMat[,dims[1]],featMat[,dims[2]],labels=as.character(as.hexmode(featEM$classification)),col=mycolors[allGroupIds],cex=.7)
#text(featMat[,dims[1]],featMat[,dims[2]],labels=as.character(as.hexmode(featEM_VII$classification)),col=mycolors[subsetObj$AllGroupIds],cex=.7)
#text(featMat[,dims[1]],featMat[,dims[2]],labels=as.character(as.hexmode(featEM_EEE$classification)),col=mycolors[subsetObj$AllGroupIds],cex=.7)
text(featMat[,dims[1]],featMat[,dims[2]],labels=as.character(as.hexmode(featEM_EEE$classification)),col=mycolors[subsetObj$AllGroupIds],cex=.7)
for (i in 1:numClusters){
  #ellipse(parms$mean[dims,i],shape=parms$variance$sigma[dims,dims,i],radius=1,center.pch=as.character(as.hexmode(i)),col=1)
  ellipse(parms$mean[dims,i],shape=parms$variance$sigma[dims,dims,i],center.pch=F,radius=1,col=1)
}
legend(.2,-.18,legend=names(specifications)[hasSignificant],col=mycolors[hasSignificant],bg="white",pch=1,cex=.75)
#legend(.2,-.18,legend=names(specifications)[-hasSignificant],col=mycolors[-hasSignificant],bg="white",pch=1,cex=.75)
#legend(-.28,-.2,legend=names(specifications),col=mycolors,bg="white",pch=1,cex=.75)

plot(featMat[,1:3],col="white",cex=.55,pch=16)
inds <- featEM$classification%in%c(3,6,8)
plot(featMat[inds,1:3],col=mycolors[groupIds[inds]])
text(featMat[inds,1:3],col=mycolors[groupIds[inds]],labels=as.character(featEM$classification[inds]))
for (i in 1:numClusters){
  ellipse(parms$mean[dims,i],shape=parms$variance$sigma[dims,dims,i],radius=1,center.pch=as.character(i),col=1)
}

inds <- featEM$classification%in%c(3,6,8)
parms <- featEM$parameters
dims <- c(1,2)
dims <- c(3,2)
dims <- c(4,5)
plot(featMat[,dims[1]],featMat[,dims[2]],col="white",cex=.55,pch=16)
ellipse(parms$mean[dims,3],shape=parms$variance$sigma[dims,dims,3],radius=1,center.pch="3",col=1)
ellipse(parms$mean[dims,6],shape=parms$variance$sigma[dims,dims,6],radius=1,center.pch="6",col=1)
ellipse(parms$mean[dims,8],shape=parms$variance$sigma[dims,dims,8],radius=1,center.pch="8",col=1)
for (i in 1:numClusters){
  ellipse(parms$mean[dims,i],shape=parms$variance$sigma[dims,dims,i],radius=1,center.pch=as.character(i),col=1)
}
text(featMat[inds,4],featMat[inds,5],col=mycolors[groupIds[inds]],labels=as.character(featEM$classification[inds]))

# Plot just one group
groupInds <- groupIds==9
dims <- 1:2
plot(featMat[,dims[1]],featMat[,dims[2]],col="white",cex=.55,pch=16)
text(featMat[groupInds,3],featMat[groupInds,2],labels=as.character(featEM$classification[groupInds]),col=mycolors[groupIds[groupInds]],cex=.7)
ellipse(featEM$parameters$mean[dims,3],shape=featEM$parameters$variance$sigma[dims,dims,3],radius=1,center.pch="3",col=1)
ellipse(featEM$parameters$mean[dims,8],shape=featEM$parameters$variance$sigma[dims,dims,8],radius=1,center.pch="8",col=1)


daEM <- Mclust(featMat[neuriteNames,1:5])
plot(daEM)
daEM12 <- Mclust(featMat[neuriteNames,1:2])
plot(daEM12)


#plot(featMat[,dims[1]],featMat[,dims[2]],col="white",cex=.55,pch=16)
#text(featMat[neuriteNames,3],featMat[neuriteNames,2],labels=as.character(featEM$classification[groupInds]),col=mycolors[groupIds[groupInds]],cex=.7)

# Plot cluster ellipses
ellipse(featEM$parameters$mean[1:2,4],shape=featEM$parameters$variance$sigma[1:2,1:2,4],radius=1,center.pch="4",col=1)
ellipse(featEM$parameters$mean[1:2,1],shape=featEM$parameters$variance$sigma[1:2,1:2,1],radius=1,center.pch=F,col=1)
ellipse(featEM$parameters$mean[1:2,2],shape=featEM$parameters$variance$sigma[1:2,1:2,2],radius=1,center.pch="2",col=1)
ellipse(featEM$parameters$mean[1:2,7],shape=featEM$parameters$variance$sigma[1:2,1:2,7],radius=1,center.pch=F,col=1)
# Plot legend if desired
legend(x=.28,y=.3,legend=names(specifications),col=1:length(specifications),pch=19)

require(tourr)
animate_xy(featMat,col=mycolors[groupIds],labels=as.character(featEM$classification),display=)
animate_xy(featMat[,1:5],col=mycolors[groupIds],labels=as.character(featEM$classification),display=)

text(featMat[groupIds==3,1],featMat[groupIds==3,2],labels=as.character(featEM$classification[groupIds==3]),col=groupIds[groupIds==3],cex=.7)

plot(featEM)


#### Density plots ####
require(MASS)
oldPar <- par()
x11()
dens2D <- kde2d(featMat[,1],featMat[,2])
kde2dplot(dens2D)
par(oldPar)

#### Looking into Dendrites group 9 (Primate/Pyramidal cell/Neocortex/Layer2,3) ####
inds <- groupIds == 9
subEM <- Mclust(featMat[inds,1:5])

plot(featMat[,dims[1]],featMat[,dims[2]],col="white",cex=.55,pch=16)
metaType <- "Gender" # Archive, Species, CorticalRegion, AgeClass, Protocol, Gender
metaTypes <- c("Archive", "Species", "CorticalRegion", "AgeClass", "Protocol", "Gender")
metaVals <- clusterMetaData[rownames(featMat)[inds],metaType]
metaGroupIds <- rep(1,sum(inds))
uniqueVals <- unique(metaVals)
for (i in 2:length(uniqueVals)){
  metaGroupIds[metaVals==uniqueVals[i]] <- i
}
text(featMat[inds,4],featMat[inds,5],col=mycolors[metaGroupIds],labels=as.character(featEM$classification[inds]))

analyses <- list()
uniqueTypeVals <- list()
for (i in 1:length(metaTypes)){
  metaType <- metaTypes[i]
  metaVals <- clusterMetaData[rownames(featMat)[inds],metaType]
  metaGroupIds <- rep(1,sum(inds))
  uniqueVals <- unique(metaVals)
  uniqueTypeVals[[i]] <- uniqueVals
  for (j in 2:length(uniqueVals)){
    metaGroupIds[metaVals==uniqueVals[j]] <- j
  }
  analyses[[i]] <- clusteringStatisticalAnalysis(featMat[inds,1:5],subEM,metaGroupIds)
}
names(analyses) <- metaTypes

# Plot on PCA axes
featurePCA <- prcomp(featMat)

# Try PCA on MDS
mdsPca <- prcomp(featMat)
plot(mdsPca)

#### Check fly DA classes (1-4) #### No longer necessary since data was fixed
# '[cC]([1-4]|I|II|III|IV)' - RegExpr for finding Class number in a DA neuron's name
#daNeuronClass <- read.csv("C:/Documents and Settings/gillette/My Documents/research/DA-Neuron-Classes.csv")
#rownames(daNeuronClass) <- paste(daNeuronClass$Name,"Dendrite",sep="-")
#daNeuronClassUpdate <- rep("Class IV",nrow(daNeuronClass))
#daNeuronClassUpdate[daNeuronClass$Subclass=="Class IV"] <- "Class III"
#daNeuronClassUpdate[daNeuronClass$Subclass=="Class III"] <- "Class II"
#daNeuronClassUpdate[daNeuronClass$Subclass=="Class II"] <- "Class I"
#daNeuronClass$Subclass <- daNeuronClassUpdate
#groupInds <- featEM$classification%in%c(6,7)
neuriteNames <- subsetObj$SeqsByGroup[[6]]
daColors <- stringsToColorIndices(daNeuronClass[neuriteNames,"Subclass"])
daChars <- stringsToColorIndices(clusterMetaData[neuriteNames,"ExperCond"])

plot(featMat[neuriteNames,1:5],col=mycolors[daColors],cex=.75,pch=c(16,2)[daChars])

legOrder <- order(unique(daNeuronClass$Subclass))
plot(featMat[neuriteNames,1:2],col=mycolors[daColors],cex=.75,pch=c(16,2)[daChars])
legend(x=.45,y=-.5,legend=c(unique(daNeuronClass$Subclass)[legOrder],"Mutant tutl"),col=c(mycolors[unique(daColors)][legOrder],1),pch=c(rep(16,4),2))


#### Empirical p-value calculator ####
N <- nrow(featMat)
baselineArray <- generateBaselineMatrix(observedIntersectMat,groupIds,featEM$classification,10000)

rawPValsAbove <- matrix(1,nrow=length(specifications),ncol=numClusters)
rawPValsBelow <- matrix(1,nrow=length(specifications),ncol=numClusters)
for (i in 1:numGroups){
  for (j in 1:numClusters){
    observed <- sum(groupIds==i & featEM$classification==j)
    expected <- mean(groupIds==i)*mean(featEM$classification==j)*N
    #binom.test(observed,N,expected/N)$p.value
    if (observed > expected){
      rawPValsAbove[i,j] <- 2*sum(observed<=baselineArray[i,j,])/baselineSize
    }else{
      rawPValsBelow[i,j] <- 2*sum(observed>=baselineArray[i,j,])/baselineSize
    }
  }
}

generateBaselineMatrix <- function(numGroups, numClusters, groupIds, clusterIds, baselineSize=1000){
  placeholderData <- rep(0,baselineSize*numGroups*numClusters)
  baselineArray <- array(placeholderData,c(numGroups,numClusters,baselineSize))
  
  # Generate N cases
  if (systemType == "KrasClust"){
    require(parallel)
    numCores <- detectCores()
    result <- mclapply(1:baselineSize,function(a,b,c,d,e){return(randomIntersectionCounts(b,c,d,e))},numGroups,groupIds,numClusters,clusterIds,mc.cores=numCores)
    for (i in 1:baselineSize){
      baselineArray[,,i] <- result[[i]]
    }
  }else{
    for (i in 1:baselineSize){
      baselineArray[,,i] <- randomIntersectionCounts(numGroups,groupIds,numClusters,clusterIds)
    }
  }
  return(baselineArray)
}

randomIntersectionCounts <- function(numGroups, groupIds, numClusters, clusterIds){
  newGroupIds <- sample(groupIds,length(groupIds),replace=F)
  observedMat <- matrix(0,nrow=numGroups,ncol=numClusters)
  for (i in 1:numGroups){
    for (j in 1:numClusters){
      observedMat[i,j] <- sum(newGroupIds==i & clusterIds==j)
    }
  }
  return(observedMat)
}

empiricalCorrectedPValues <- function(numGroups, numClusters, groupIds, clusterIds, baselineSize=10000){
  placeholderData <- rep(0,baselineSize*numGroups*numClusters)
  baselineArray <- array(placeholderData,c(numGroups,numClusters,baselineSize))
  
  # Generate N cases
  generateBaselineMatrix(numGroups, numClusters, groupIds, clusterIds, baselineSize)
  
  # Generate random p-values
  testCounts <- randomIntersectionCounts(i,numGroups,groupIds,numClusters,clusterIds,mc.cores=numCores)
  
  pMatrix <- matrix(0,nrow=numGroups,ncol=numCols)
  for (i in 1:numGroups){
    for (j in 1:numClusters){
      
    }
  }
  return(list(RawPValues=))
}
