#############################################################################
#
# ClusterAnalysisFunctions.r
#
# various
#
#############################################################################

clusteringStatisticalAnalysis <- function(featMat,featEM,subsetObject,groupNames,excludeUnspecified=T){
  # Calculate overlap of group/cluster, as well as expected values
  N <- nrow(featMat)
  numGroups <- length(unique(subsetObject$GroupIds))
  numClusters <- featEM$G
  
  # Either create group for unspecified, or eliminate them
  groupIds <- subsetObject$GroupIds
  if (excludeUnspecified){
    classification <- featEM$classification[names(featEM$classification)%in%subsetObject$Sequences]
  }else{
    # Calculate number of sequences not in a group
    nonGroupCount <- nrow(featMat)-length(groupIds)
    if (nonGroupCount > 0){
      numGroups = numGroups + 1
      groupNames[numGroups] = "All Others"
      subsetObject$Sequences <- append(subsetObject$Sequences,names(groupIds)[!names(groupIds)%in%subsetObject$GroupIds])
      groupIds <- c(groupIds,rep(max(subsetObject$GroupIds)+1,nonGroupCount))
    }
    classification <- featEM$classification
  }

  # Set up result matrices
  clusterNames <- paste("C",1:numClusters," (",aggregate(featEM$classification,list(featEM$classification),length)[,2],")",sep="")
  observedIntersectMat <- matrix(0,nrow=numGroups,ncol=numClusters,dimnames=list(groupNames,clusterNames))
  aboveExpectPVals <- matrix(0,nrow=numGroups,ncol=numClusters,dimnames=list(groupNames,clusterNames))
  belowExpectPVals <- matrix(0,nrow=numGroups,ncol=numClusters,dimnames=list(groupNames,clusterNames))
  logPVals <- matrix(0,nrow=numGroups,ncol=numClusters,dimnames=list(groupNames,clusterNames))
  
  # Bonferonni correction
  df <- (numGroups-1)*(numClusters-1)
  bonfCor <- df
  
  # Calculate observed values
  for (i in 1:numGroups){
    for (j in 1:numClusters){
      observed <- sum(groupIds==i & featEM$classification[subsetObject$Sequences]==j)
      observedIntersectMat[i,j] <- observed
    }
  }
  
  # Run chi square test
  chiSqResult <- chisq.test(observedIntersectMat)
  
  # Generate data associated with expected results
  expectedIntersectMat <- chiSqResult$expected
  below <- chiSqResult$stdres < 0
  logPVals <- -log(1-(1-pnorm(-chiSqResult$stdres))^bonfCor,10)
  logPVals[below] <- log(1-(1-pnorm(chiSqResult$stdres[below]))^bonfCor,10)
  
  diffMat <- observedIntersectMat-expectedIntersectMat
  ratioMat <- observedIntersectMat/expectedIntersectMat
  chiSqResiduals <- chiSqResult$stdres
  return(list(ChiSq=chiSqResult,Observed=observedIntersectMat,Expected=expectedIntersectMat,Difference=diffMat,Ratio=ratioMat,LogP=logPVals,Residuals=chiSqResiduals))
}

clusterAndScoreFeatures <- function(sampleData,classPartition,clusterMethod=c("hclust","kmeans","em","kmediods"),submethod="ward",clusterMetric=AdjustedRandIndex,normalizeData=F,...){
  clusterMethod <- match.arg(clusterMethod)
  numClasses=length(unique(classPartition))
  otherParams <- list(...)
  otherParamNames <- names(otherParams)
  
  # If data normalization is required
  if (normalizeData){
    sampleData <- logTransformData(sampleData,...)
#     xformParams <- names(formals(logTransformData))
#     xformPresentParams <- list(myData=sampleData)
#     for (param in otherParamNames){
#       if (param%in%xformParams){
#         xformPresentParams <- append(xformPresentParams,list(param=otherParams[[param]]))
#       }
#     }
#     sampleData <- do.call(logTransformData,xformPresentParams)
  }
  
  if (clusterMethod=="hclust"){
    sampleDist <- dist(sampleData)
    return(hclustAndScore(sampleDist,classPartition,numClasses,submethod))
  }else if (clusterMethod=="kmeans"){
    return(kmeansAndScore(sampleData,classPartition,numClasses,clusterMetric))
  }else if (clusterMethod=="kmediods"){
    return(kmediodsAndScore(sampleData,classPartition,numClasses,clusterMetric))
  }else if (clusterMethod=="em"){
    return(emAndScore(sampleData,classPartition,numClasses,clusterMetric,...))
  }
}

clusterAndScoreDistances <- function(sampleData,classPartition,clusterMethod=c("hclust","kmeans","em","kmediods"),submethod="ward",clusterMetric=AdjustedRandIndex,numDimensions=3,...){
  clusterMethod <- match.arg(clusterMethod)
  numClasses=length(unique(classPartition))
  if (clusterMethod=="hclust"){
    return(hclustAndScore(sampleDist,classPartition,numClasses,submethod))
  }else if (clusterMethod=="kmeans"){
    subsetMDSFull <- cmdscale(subsetSampleDist,k=numDimensions,eig=T)
    subsetMDS <- subsetMDSFull$points
    return(kmeansAndScore(subsetMDS,classPartition,numClasses,clusterMetric))
  }else if (clusterMethod=="kmediods"){
    subsetMDSFull <- cmdscale(subsetSampleDist,k=numDimensions,eig=T)
    subsetMDS <- subsetMDSFull$points
    return(kmediodsAndScore(subsetMDS,classPartition,numClasses,clusterMetric))
  }else if (clusterMethod=="em"){
    subsetMDSFull <- cmdscale(subsetSampleDist,k=numDimensions,eig=T)
    subsetMDS <- subsetMDSFull$points
    return(emAndScore(subsetMDS,classPartition,numClasses,clusterMetric,...))
  }
}

hclustAndScore <- function(sampleDist,classPartition,numClasses=length(unique(classPartition)),method="ward",clusterMetric=AdjustedRandIndex,permute=F){
  sampleClust <- hclust(sampleDist,method=method)
  clusterPartition <- cutree(sampleClust,numClasses)
  if (permute){
    return(scorePermutations(clusterPartition,classPartition,clusterMetric))
  }else{
    return(clusterMetric(clusterPartition,classPartition))
  }
}

kmeansAndScore <- function(sampleData,classPartition,numClasses=length(unique(classPartition)),clusterMetric=AdjustedRandIndex,permute=F){
  kmeansClust <- kmeans(sampleData,centers=numClasses)
  clusterPartition <- kmeansClust$cluster
  if (permute){
    return(scorePermutations(clusterPartition,classPartition,clusterMetric))
  }else{
    return(clusterMetric(clusterPartition,classPartition))
  }
}

kmediodsAndScore <- function(sampleData,classPartition,numClasses=length(unique(classPartition)),clusterMetric=AdjustedRandIndex,permute=F){
  clusterPartition <- pam(sampleData,k=numClasses,cluster.only=T)
  if (permute){
    return(scorePermutations(clusterPartition,classPartition,clusterMetric))
  }else{
    return(clusterMetric(clusterPartition,classPartition))
  }
}

emAndScore <- function(sampleData,classPartition,numClasses=length(unique(classPartition)),clusterMetric=AdjustedRandIndex,permute=F,...){
  MclustRes <- Mclust(sampleData,G=numClasses,...)
  clusterPartition <- MclustRes$classification
  if (MclustRes$G == 1){
    return(0)
  }
  if (permute){
    return(scorePermutations(clusterPartition,classPartition,clusterMetric))
  }else{
    return(clusterMetric(clusterPartition,classPartition))
  }
}

BIC.kmeans <- function(kmeansObject,originalData,classificationLikelihood=F){
  K <- nrow(kmeansObject$centers) # Number of clusters (groups)
  d <- ncol(originalData) # number of features (dimensions)
  N <- nrow(originalData) # number of observations
  modelLogLik <- logLik.kmeans(kmeansObject,originalData,classificationLikelihood)
#  return(modelLogLik-(K*d+1)*log(N)/2)
  return(2*modelLogLik-(K*d+1)*log(N))
}

logLik.kmeans <- function(kmeansObject,originalData,classificationLikelihood=F){
  if (is.data.frame(originalData)){
    originalData <- as.matrix(originalData)
  }
  K <- nrow(kmeansObject$centers) # numClusters
  N <- nrow(originalData)
  d <- ncol(originalData)
  # sigma is for a given axis, so sqrt(avg withinss) would need to be devided by sqrt(d)
  sigmaSq <- kmeansObject$tot.withinss/sum(kmeansObject$size)/d
  centers <- kmeansObject$centers
  if (classificationLikelihood){
    sqDistances <- mapply(distanceSquared,originalData,centers[kmeansObject$cluster,])
    jointLogLikelihood <- N*log(1/sqrt(2*pi*sigmaSq)) + sum(-(1/2)*sqDistances/sigmaSq)
  }else{
    pointLikelihoodPartials <- apply(originalData,1,pointLikelihoodPartial,centers,sigmaSq)
    jointLogLikelihood <- N*log(1/(K*sqrt(2*pi*sigmaSq))) + sum(log(pointLikelihoodPartials))
  }
  #jointClassLik <- (1/numClusters/sqrt(2*pi*sigmaSq))*sum(exp(-(1/2)*sqDistances/sigmaSq))
  return(jointLogLikelihood)
}

pointLikelihoodPartial <- function(x,centers,sigmaSq){
  return(sum(exp(-(1/2)*apply(centers,1,distanceSquared,x)/sigmaSq)))
}

distanceSquared <- function(point1, point2){
  sum((point1-point2)^2)
}

scorePermutations <- function(clusterPartition,classPartition,clusterMetric){
  numClusters <- length(unique(clusterPartition))
  numClasses <- length(unique(classPartition))
  if (numClusters==numClasses){
    return(clusterMetric(clusterPartition,classPartition))
  }else if (numClusters < numClasses){
    print("Error: There are more classes than clusters")
    return(NaN)
  }else{
    availClusterNums <- unique(clusterPartition)
    groupLists <- getPermutationGroupings(numClasses,availClusterNums)
    scores <- c()
    for (i in 0:(getNumGroupings(numGroups,numClusters)-1)){
      groupsList <- list(getPermutationGrouping(i,numGroups,availClusterNums))
      newClusterPartition <- rep(0,length(clusterPartition))
      for (j in 1:numClasses){
        newClusterPartition[clusterPartition%in%groupLists[[j]]] <- j
      }
      scores[i+1] <- clusterMetric(newClusterPartition,classPartition)
    }
    return(max(scores))
  }
}

getPermutationGroupings <- function(numGroups,availClusterNums){
  groupLists <- list()
  if (numGroups == length(availClusterNums)){
    groupLists <- append(groupList,as.list(availClusterNums))
  }else{
    for (i in 0:(getNumGroupings(numGroups,length(availClusterNums))-1)){
      groupLists <- append(groupLists,list(getPermutationGrouping(i,numGroups,availClusterNums)))
    }
  }
  return(groupLists)
}

getPermutationGrouping <- function(groupingNum,numGroups,availClusterNums){
  numClusters <- length(availClusterNums)
  if (numGroups == 1){
    return(list(availClusterNums))
  }else if (numGroups == numClusters){
    return(as.list(availClusterNums))
  }else if (numGroups > numClusters){
    print("Number of clusters must be at least as many as number of groups")
    return(NaN)
  }
  
  numGroupings <- getNumGroupings(numGroups,numClusters)
  if (groupingNum >= numGroupings){
    print(paste("Invalid groupingNum (",groupingNum,") for ",numGroups," groups and ",numClusters," clusters. Max groupingNum is ",numGroupings-1,sep=""))
    return(NaN)
  }

  # Separate first item from the rest
  firstItem <- availClusterNums[1]
  availClusterNums <- availClusterNums[-1]
  numClusters <- numClusters - 1
  
  # Determine how many items go in the first group
  partialSum <- 0
  prevPartial <- 0
  itemsInFirstGroup <- -1
  repeat{
    prevPartial <- partialSum
    itemsInFirstGroup <- itemsInFirstGroup + 1
    partialSum <- partialSum + choose(numClusters,itemsInFirstGroup)*getNumGroupings(numGroups-1,numClusters-itemsInFirstGroup)
    if (partialSum > groupingNum){
      break
    }
  }
  itemsRemaining <- numClusters - itemsInFirstGroup
  remainingNum <- groupingNum - prevPartial
  
  if (itemsInFirstGroup == 0){
    groupingList <- append(list(firstItem),getPermutationGrouping(groupingNum,numGroups-1,availClusterNums))
  }else{
    # Determine the groupingNum for the remaining, and the choose num for the first group
    numGroupingsNext = getNumGroupings(numGroups-1,itemsRemaining)
    firstGroupingNum = remainingNum/numGroupingsNext
    nextGroupingNum = remainingNum - numGroupingsNext * firstGroupingNum
  #  print(paste(groupingNum,remainingNum,numGroupingsNext,nextGroupingNum,firstGroupingNum))
  
    # Determine which items go in the first group (in addition to the first item)
    firstGroup <- firstItem
    numItemsToPick <- itemsInFirstGroup

    i <- 0
    partialSum <- 0
    while(numItemsToPick > 0){
      i <- i + 1
      prevPartial <- partialSum
      partialSum <- partialSum + choose(numClusters-i,numItemsToPick-1)
      if (partialSum > firstGroupingNum){
        firstGroup <- append(firstGroup,availClusterNums[i])
        numItemsToPick <- numItemsToPick-1
        partialSum <- prevPartial
      }
    }
    
    remainingItems <- availClusterNums[!availClusterNums%in%firstGroup]
    groupingList <- append(list(firstGroup),getPermutationGrouping(nextGroupingNum,numGroups-1,remainingItems))
  }  
  return(groupingList)
}

numGroupingsMap <- list()
putNumGroupings <- function(key, value){
  numGroupingsMap <- get("numGroupingsMap", envir=as.environment(1), mode = "any", inherits = FALSE)
  numGroupingsMap[[key]] <- value
  assign("numGroupingsMap", numGroupingsMap)
}
getNumGroupingsFromMap <- function(key){
  numGroupingsMap <- get("numGroupingsMap", envir=as.environment(1), mode = "any", inherits = FALSE)
  return(numGroupingsMap[[key]])
}
numGroupingsMapContains <- function(key){
  numGroupingsMap <- get("numGroupingsMap", envir=as.environment(1), mode = "any", inherits = FALSE)
  return(key %in% names(numGroupingsMap))
}

getNumGroupings <- function(numGroups,numVals){
  key <- paste(numGroups,numVals,sep="_")
  if (numGroups > numVals){
    return(0)
  }
  if (numGroups == 1 || numVals == numGroups){
    return(1)
  }else if (numVals == numGroups+1){
    return(choose(numVals,2))
  }else if (numGroupingsMapContains(key)){
    return(getNumGroupingsFromMap(key))
  }else{
    numGroupings <- 0
    for (i in 0:(numVals-numGroups)){ # First val will be in first group
      numGroupings <- numGroupings + choose(numVals-1,i)*getNumGroupings(numGroups-1,numVals-1-i)
    }
    putNumGroupings(key,numGroupings)
    return(numGroupings)
  }
}

getSetString <- function(set){
  setStr <- ""
  for (elm in set){
    fullVal <- paste(elm,collapse="&")
    setStr <- paste(setStr,"/",fullVal,sep="")
  }
  return(setStr)
}

# Drop poorly performing principal components based on explained variance
determineNumSignificantPCs <- function(pcObj,threshold=1,originalData=NULL){
  if (is.null(originalData)){
    return(pcObj$x[,pcObj$sdev>=threshold])
  }
  else{
    # Best methods for determining when to stop
    # Avg-PA, Avg-Rnd, PA, Rnd-Lambda, Rnd-F and Part
    
    # Randomizing original data, running PCA 999 times
    if(is.numeric(pcObj$scale)){
      originalData <- scale(originalData)
    }
    rows <- nrow(originalData)
    sdMat <- matrix(0,nrow=999,ncol=ncol(originalData))
    for (i in 1:999){
      randData <- originalData
      for (j in 1:ncol(originalData)){
        randData[,j] <- originalData[floor(runif(rows,0,rows)+1),j]
      }
      randPCA <- prcomp(randData,scale=F)
      sdMat[i,] <- randPCA$sdev
    }
    
    # Reorder random PCA results
    for (j in 1:ncol(originalData)){
      sdMat[,j] <- sort(sdMat[,j])
    }
    # One less than the first PC with sd less than the random PCA 95th percentile
    return(which(pcObj$sdev<sdMat[950,])[1]-1)
  }
}

bestPCSeparationOrder <- function(pcObj,neuriteClasses,pcStop=determineNumSignificantPCs(pcObj),sampleRuns=10,sampleSize=floor(min(35*numClasses,0.8*numObjects)),separationMethod=c("cluster","scatter")){
  separationMethod <- match.arg(separationMethod)
  numClasses <- length(unique(neuriteClasses))
  pcMatrix <- pcObj$x[,1:pcStop]
  numObjects <- nrow(pcMatrix)
  pcOrder <- 1:pcStop
  if (separationMethod == "cluster"){
    pairedClustStats <- array(0,dim=c(sampleRuns,pcStop,pcStop))
    # Take random samples of the data to produce more accurate measure of cluster capability
    for (k in 1:sampleRuns){
      sampleSeqs <- sample(1:numObjects,sampleSize,replace=FALSE)
      pcSample <- pcMatrix[sampleSeqs,]
      sampleClasses <- neuriteClasses[sampleSeqs]
      
      # Run pairs of PCs to evaluate robustness of each
      for (i in 1:(pcStop-1)){
        for (j in (i+1):pcStop){
          pcSampleDist <- dist(pcSample[,c(i,j)])
          pcSubClust <- hclust(pcSampleDist,method=method)
          pcPartition <- cutree(pcSubClust,numClasses)
          pairedClustStats[k,i,j] <- AdjustedRandIndex(sampleClasses,pcPartition)
        }
      }
    }
    
    # Get mean of stats for each PC pair, load into dist-type array
    statMat <- rep(0,pcStop*(pcStop-1)/2)
    for (i in 1:(pcStop-1)){
      for (j in (i+1):pcStop){
        statMat[lowerTriangleIndex(i,j)] <- mean(pairedClustStats[,i,j])
      }
    }
    # Get mean of stats for each individual PC
    indPCStats <- rep(0,pcStop)
    for (i in 1:pcStop){
      sapply((1:10)[-4],FUN=lowerTriangleIndex,4)
      indPCStats[i] <- mean(statMat[sapply((1:pcStop)[-i],FUN=lowerTriangleIndex,i)])
    }
  }
  else if (separationMethod=="Scatter"){
    
  }
  
  return(order(indPCStats,decreasing=T))
}

bestPCClusterOrder <- function(pcObj,neuriteClasses,pcStop=determineNumSignificantPCs(pcObj),sampleRuns=10,sampleSize=floor(min(35*numClasses,0.8*numObjects)),pairs=T){
  numClasses <- length(unique(neuriteClasses))
  pcMatrix <- pcObj$x[,1:pcStop]
  numObjects <- nrow(pcMatrix)
  pcOrder <- 1:pcStop
  # Take random samples of the data to produce more accurate measure of cluster capability
  if (pairs){
    pairedClustStats <- array(0,dim=c(sampleRuns,pcStop,pcStop))
    for (k in 1:sampleRuns){
      sampleSeqs <- sample(1:numObjects,sampleSize,replace=FALSE)
      pcSample <- pcMatrix[sampleSeqs,]
      sampleClasses <- neuriteClasses[sampleSeqs]
      
      # Run pairs of PCs to evaluate robustness of each
      for (i in 1:(pcStop-1)){
        for (j in (i+1):pcStop){
          pcSampleDist <- dist(pcSample[,c(i,j)])
          pcSubClust <- hclust(pcSampleDist,method=method)
          pcPartition <- cutree(pcSubClust,numClasses)
          pairedClustStats[k,i,j] <- AdjustedRandIndex(sampleClasses,pcPartition)
        }
      }
    }
    
    # Get mean of stats for each PC pair, load into dist-type array
    statMat <- rep(0,pcStop*(pcStop-1)/2)
    for (i in 1:(pcStop-1)){
      for (j in (i+1):pcStop){
        statMat[lowerTriangleIndex(i,j)] <- mean(pairedClustStats[,i,j])
      }
    }
    # Get mean of stats for each individual PC
    indPCStats <- rep(0,pcStop)
    for (i in 1:pcStop){
      sapply((1:10)[-4],FUN=lowerTriangleIndex,4)
      indPCStats[i] <- mean(statMat[sapply((1:pcStop)[-i],FUN=lowerTriangleIndex,i)])
    }
    
    return(order(indPCStats,decreasing=T))
  }else{
    # Singles
    clustStats <- array(0,dim=c(sampleRuns,pcStop))
    for (k in 1:sampleRuns){
      sampleSeqs <- sample(1:numObjects,sampleSize,replace=FALSE)
      pcSample <- pcMatrix[sampleSeqs,]
      sampleClasses <- neuriteClasses[sampleSeqs]
      
      # Run pairs of PCs to evaluate robustness of each
      for (i in 1:pcStop){
        pcSampleDist <- dist(pcSample[,i])
        pcSubClust <- hclust(pcSampleDist,method=method)
        pcPartition <- cutree(pcSubClust,numClasses)
        clustStats[k,i] <- AdjustedRandIndex(sampleClasses,pcPartition)
      }
    }
    
    # Get mean of stats for each individual PC
    indPCStats <- apply(clustStats,2,mean)
    
    return(order(indPCStats,decreasing=T))
  }
}

##################################################
# Cosine similarity clustering functions
#

# Scaled cosine similarity based on the relative distribution of distance of elements
# from the population center.
scaledCosineSimilarity <- function(x){
  cosineSim <- cosine(x)
  magnitudes <- sort(apply(x,1,magnitude))
  similarityScalingFactors <- apply(x,1,similarityScalingFactor,magnitudes)
  numRows <- nrow(x)
  for (i in 1:(numRows-1)){
    for (j in (i+1):numRows){
      simInd <- (i-1)*(i-2)/2+j
      cosineSim[simInd] <- cosineSim[simInd] * similarityScalingFactors[i] * similarityScalingFactors[j]
    }
  }
  return(cosineSim)
}

## Similarity scaling factor (0-1) scales vector similarity such that vectors close to the origin are
## more similar (smaller factor) to other vectors, all else being equal.
# Uses the percentage of a vector's magnitude relative to the average(median is default) of the set
similarityScalingFactor <- function(vect,magnitudeDistribution,average=median){
  return(max(magnitude(vect)/average(magnitudeDistribution),1))
}

# Calculates a cosine similarity matrix based on a matrix or dataframe (columns are features)
cosine <- function(x){
  numRows <- nrow(x)
  xHat <- normalize(x)
  cosDistMat <- matrix(0,nrow=numRows,ncol=numRows)
  for (i in 1:(numRows-1)){
    for (j in (i+1):numRows){
      cosDistMat[j,i] <- sum(xHat[i,]*xHat[j,])
    }
  }
  return(as.dist(cosDistMat))
}

centerFeatures <- function(mydata,FUN=mean){
  centers <- apply(mydata,2,FUN=FUN)
  for (i in 1:length(centers)){
    mydata[i] <- mydata[i] - centers[i]
  }
  return(mydata)
}

normalize <- function(x){
  xMag <- apply(x,1,magnitude)
  return(x/xMag)
}

magnitude <- function(x){
  return(sqrt(sum(x^2)))
}

squareSum <- function(x){
  return(sum(x^2))
}

#######################################################
# Basic clustering and scoring on multiple subsamples
#

sampleClusterRIsEvenly <- function(featureVect,groupIds,clusterMethod=c("hclust","kmeans","em","kmediods"),submethod="ward",clusterMetric=AdjustedRandIndex,proportionOfSmallest=0.75,sampleSize=proportionOfSmallest*minGroupSize,numSamples,randomizeGroups=F){
  minGroupSize <- min(aggregate(groups,by=list(groups),length)[,2])
  uniqueIds <- unique(groupIds)
  return(sampleClusterRIs(featureVect,groupIds,rep(sampleSize,length(uniqueIds)),numSamples,clusterMethod,submethod,clusterMetric,randomizeGroups))
}
sampleClusterRIsProportionally <- function(featureVect,groupIds,clusterMethod=c("hclust","kmeans","em","kmediods"),submethod="ward",clusterMetric=AdjustedRandIndex,proportion=0.75,numSamples,randomizeGroups=F){
  uniqueIds <- unique(groupIds)
  sampleSizes <- c()
  for (i in 1:length(uniqueIds)){
    sampleSizes[i] <- round(proportion*sum(groupIds==uniqueIds[i]))
  }
  return(sampleClusterRIs(featureVect,groupIds,sampleSizes,numSamples,clusterMethod,submethod,clusterMetric,randomizeGroups))
}

getSampleSizesEvenly <- function(groupSizes,proportionOfSmallest=0.75,maxSize=NA){
  size <- round(min(groupSizes)*proportionOfSmallest)
  if (!is.na(maxSize)){
    size <- min(maxSize,size)
  }
  return(rep(size,length(groupSizes)))
}
getSampleSizesProportionally <- function(groupSizes,proportion=0.75){
  sampleSizes <- c()
  for (i in 1:length(groupSizes)){
    sampleSizes[i] <- round(proportion*groupSizes[i])
  }
  return(sampleSizes)
}

sampleClusterRIs <- function(featureVect,groupIds,sampleSizes,numSamples=length(subsamples),subsamples=generateSubsamples(groupIds,sampleSizes,numSamples),clusterMethod=c("hclust","kmeans","em","kmediods"),submethod="ward",clusterMetric=AdjustedRandIndex,randomizeGroups=F){
  if (length(subsamples) != numSamples){
    stop("numSamples does not match length of subsamples. Only one of those paraemters should be provided.")
  }
  if (randomizeGroups){
    groupIds <- sample(groupIds,length(groupIds),replace=F)
  }
  
  require(parallel)
  results <- mclapply(subsamples,
          function(subsample,featureVect,groupIds,clusterMethod,submethod,clusterMetric){
            return(clusterAndScoreFeatures(featureVect[subsample,],groupIds[subsample],methodClass=clusterMethod,submethod=submethod,clusterMetric=clusterMetric))
          },
          featureVect,groupIds,clusterMethod,submethod,clusterMetric,mc.cores=detectCores())
  return(results)
}

generateSubsamples <- function(groupIds,sampleSizes,numSamples){
  return(lapply(1:numSamples,function(sampleNum,groupIds,sampleSizes){generateSubsample(groupIds,sampleSizes)},groupIds,sampleSizes))
}
generateSubsample <- function(groupIds,sampleSizes){
  uniqueIds <- unique(groupIds)
  subsample <- c()
  for (i in 1:length(sampleSizes)){
    subsample <- append(subsample,sample(which(uniqueIds[i]==groupIds),sampleSizes[i],replace=F))
  }
  return(subsample)
}
##################################################
# Cluster specification sets by:
#    AlignmentScoreLTS (currently commented)
#    AlignmentScoreSTL (currently commented)
#    Feature Vectors
#    Feature Vectors + better AlignmentScore
#    PCA of nmer profiles (relative)
#    PCA of 4-grams
#    ngram cosign similarity
#    Cumulative PCs on feature vectors
#    Cumulative PCs on feature vectors + better AlignmentScore
clusterTypesComplex <- function(specifications,typeMetaData,alignmentScoresLTS,alignmentScoresSTL,featureVectors,nmers,ngrams,method="ward",clusterMetric=AdjustedRandIndex,minGroupSize=30,numClusterRuns=20,sampleProportionally=T,sampleSize=if (sampleProportionally) numGroups*minGroupSize else minGroupSize,alignmentScoreWeight=2,maxCumulativePCs=6,debugFlag=1){
  numGroups <- length(specifications)
  
  typeMetaData <- typeMetaData[rownames(featureVectors),]
  
  # Get sequences of each group
  groupSequences <- list()
  allSequences <- c()
  specStrings <- c()
  groupsTooSmall <- c()
  neuriteClasses <- c()
  for (i in 1:numGroups){
    groupSequences[[i]] <- rownames(getSubset2(typeMetaData,specifications[[i]],typeMetaData))
    neuriteClassesSub <- rep(i,length(groupSequences[[i]]))
    names(neuriteClassesSub) <- groupSequences[[i]]
    neuriteClasses <- append(neuriteClasses,neuriteClassesSub)
    specStrings[i] <- getSetString(specifications[[i]])
    # Make sure all groups are large enough
    if (length(groupSequences[[i]]) < minGroupSize){
      warning(paste("Specification ",specStrings[i]," too small: ",length(groupSequences[[i]])," (min of ",minGroupSize,")\n",sep=""))
      groupsTooSmall <- append(groupsTooSmall,i)
    }
    else{
      allSequences <- append(allSequences,groupSequences[[i]])
    }
  }

  if (length(groupsTooSmall > 0)){
    # If one or no groups remain, must stop
    if (numGroups - length(groupsTooSmall) < 2){
      warning(paste(numGroups - length(groupsTooSmall)," groups remaing large enough to run test",sep=""))
      return(NA)
    }
    # Remove groups that are too small
    else {
      specifications <- specifications[-groupsTooSmall]
      groupSequences <- groupSequences[-groupsTooSmall]
      specStrings <- specStrings[-groupsTooSmall]
      numGroups <- length(specifications)
    }
  }
  
  # Check to make sure no overlap between groups
  for (i in 1:(numGroups-1)){
    for (j in (i+1):numGroups){ 
      if (sum(groupSequences[[i]]%in%groupSequences[[j]]) > 0){
        stop("Overlap between specifications")
      }
    }
  }

  ### Preclustering data set up ###

  # Set all feature vector type data frames to contain only the relevant data
  featureVectors <- featureVectors[allSequences,]
  nmers <- nmers[allSequences,]
  ngrams <- ngrams[allSequences,]
  
  # Apply weight to alignmentScores (won't affect clustering on alignmentScores alone)
  #alignmentScoresLTS <- alignmentScoreWeight * alignmentScoresLTS[allSequences,allSequences]
  #alignmentScoresSTL <- alignmentScoreWeight * alignmentScoresSTL[allSequences,allSequences]
  
  # Get morphometric feature principal components
  # First need to ensure all features have variance
  noVar <- which(apply(featureVectors,2,var)==0)
  if (length(noVar) > 0){
    warning(paste("Zero variance found in features:",paste(colnames(featureVectors)[noVar],collapse=", "),sep=" "))
    featureVectors <- featureVectors[-noVar]
  }
  pcObj <- prcomp(featureVectors,scale.=T)
  PCs <- pcObj$x
  # Determine number of useful PCs
  print("Determining num PCs for feature vector")
  numPCs <- determineNumSignificantPCs(pcObj,originalData=featureVectors)
  # Get order of PCs best for clustering
  pcOrder <- bestPCClusterOrder(pcObj,neuriteClasses,pcStop=numPCs)
  numPCs <- min(numPCs,maxCumulativePCs)
  pcOrder <- pcOrder[1:numPCs]
  # Reorder PCs based on ability to cluster
  PCs <- PCs[,pcOrder]

  # Get NMer feature principal components
  noVar <- which(apply(nmers,2,var)==0)
  if (length(noVar) > 0){
    warning(paste("Zero variance found in nmers:",paste(colnames(nmers)[noVar],collapse=", "),sep=" "))
    nmers <- nmers[-noVar]
  }
  pcObj <- prcomp(nmers,scale.=F) 
  # Determine number of useful PCs
  print("Determining num PCs for nmers")
  numNMerPCs <- determineNumSignificantPCs(pcObj,originalData=nmers)
  nmerPCs <- pcObj$x[,1:numNMerPCs]
  
  # Get ngram feature principal components
  noVar <- which(apply(nmers,2,var)==0)
  if (length(noVar) > 0){
    warning(paste("Zero variance found in nmers:",paste(colnames(ngrams)[noVar],collapse=", "),sep=" "))
    ngrams <- ngrams[-noVar]
  }
  pcObj <- prcomp(ngrams,scale.=F)
  # Determine number of useful PCs
  print("Determining num PCs for ngrams")
  numNGramPCs <- determineNumSignificantPCs(pcObj,originalData=ngrams)
  ngramPCs <- pcObj$x[,1:numNGramPCs]

  # Helper values for the number of columns in the various datasets
  numFeatures <- ncol(featureVectors)
  numNMers <- ncol(nmers)
  numNGrams <- ncol(ngrams)
  
  # Get Z-scores of feature matrix
  standardizedFeatureVects <- as.data.frame(apply(featureVectors,2,scale))
  rownames(standardizedFeatureVects) <- rownames(featureVectors)
  
  # Get ngram center (use general sample instead of subsample)
  centeredNGrams <- centerFeatures(ngrams)
  
  ### Cluster Analysis ###
  
  # AlignmentScoreLTS, AlignmentScoreSTL, Feature Vectors (w/ & w/oAlignmentScore), 
  # Each feature (w/ & w/oAlignmentScore),
  # Cumulative PCs on feature vectors (w/ & w/o alignmentScore),
  # Significant PCs combined on nmer profiles (relative), Significant PCs combined PCs on 4-grams
  numColumns <- 7+2*numFeatures+2*maxCumulativePCs
  comparisonMeans <- rep(NA,numColumns)
  comparisonSEMs <- rep(NA,numColumns)
  columnNamesAbsolute <- c("AlignmentScoreLTS","AlignmentScoreSTL","NMerPCs","NGramPCs","NGramCosineSim","AllFeatures","AllFeatures+AS",colnames(featureVectors),paste(colnames(featureVectors),"+AS",sep=""),
                           paste("CumulFeaturePCs",1:maxCumulativePCs,sep=""),paste("CumulFeaturePCs+AS",1:maxCumulativePCs,sep=""))
  names(comparisonMeans) <- columnNamesAbsolute
  names(comparisonSEMs) <- columnNamesAbsolute

  # Set up matrix for storing Rand Index across subsample runs
  clusterRunsRIMat <- matrix(0,nrow=numClusterRuns,ncol=numColumns)

  subsamples <- list()
  classPartitions <- list()
  
  # create samples and run alignments scores (to determine whether LTS or STL is better)
  for (k in 1:numClusterRuns){
    # Generate sample set
    if (sampleProportionally){
      classPartition <- rep(0,sampleSize)
      subsample <- sample(allSequences,sampleSize,replace=F)
      # Determine which group each sample element is a part
      for (i in 1:numGroups){
        classPartition[subsample%in%groupSequences[[i]]] <- i
      }
    }
    else{
      subsample <- c()
      classPartition <- c()
      for (i in 1:numGroups){
        subsample <- append(subsample,sample(groupSequences[[i]],sampleSize,replace=F))
        classPartition <- append(classPartition,rep(i,sampleSize))
      }
    }
    subsamples[[k]] <- subsample
    classPartitions[[k]] <- classPartition
        
    numClasses <- length(unique(classPartition))
    #alignmentScoresLTSSubsample <- alignmentScoresLTS[subsample,subsample]
    #alignmentScoresSTLSubsample <- alignmentScoresSTL[subsample,subsample]
    
    # Alignment score clustering
#     alignmentDistLTS <- dist(alignmentScoresLTSSubsample)
#     clusterRunsRIMat[k,1] <- clusterAndScoreDistances(alignmentDistLTS,classPartition,method=method,clusterMetric=clusterMetric)

#     alignmentDistSTL <- dist(alignmentScoresSTLSubsample)
#     clusterRunsRIMat[k,2] <- clusterAndScoreDistances(alignmentDistSTL,classPartition,method=method,clusterMetric=clusterMetric)
  }
  
  # Determine which alignment traversal produces better clustering
#   ltsScore <- mean(clusterRunsRIMat[,1])
#   stlScore <- mean(clusterRunsRIMat[,2])
#   if (ltsScore > stlScore){
#     betterAlignmentScores <- alignmentScoresLTS
#   }
#   else{
#     betterAlignmentScores <- alignmentScoresLTS
#   }
#  
  # Run the rest of the cluster analyses
  for (k in 1:numClusterRuns){
    print(paste("Running subsample",k))
    subsample <- subsamples[[k]]
    classPartition <- classPartitions[[k]]
    
#     betterAlignmentScoresSubsample <- betterAlignmentScores[subsample,subsample]
#     alignmentScoreDist <- dist(betterAlignmentScoresSubsample)
    
    # NMer Principal Components
    sampleDist <- dist(nmerPCs[subsample,1:numNMerPCs])
    clusterRunsRIMat[k,3] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)
    
    # NGram Principal Components
    sampleDist <- dist(ngramPCs[subsample,1:numNGramPCs])
    clusterRunsRIMat[k,4] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)
    
    # NGram cosine similarity
    sampleDist <- scaledCosineSimilarity(centeredNGrams[subsample,])
    clusterRunsRIMat[k,5] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)
    
    # All features clustering
    fvDist <- dist(standardizedFeatureVects[subsample,])
    clusterRunsRIMat[k,6] <- clusterAndScoreDistances(fvDist,classPartition,method=method,clusterMetric=clusterMetric)
    
    # All features + alignment score clustering
#     sampleDist <- combineDists(fvDist,alignmentScoreDist)
#     clusterRunsRIMat[k,7] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)

    # Features and cumulative feature PCs
    for (f in 1:numFeatures){
      # Individual feature clustering
      sampleDist <- dist(featureVectors[subsample,f])
      clusterRunsRIMat[k,7+f] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)
      
      # Individual feature + alignment score clustering
#       sampleDist <- combineDists(sampleDist,alignmentScoreDist)
#       clusterRunsRIMat[k,7+numFeatures+f] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)
      
      # Cumulative PC clustering
      if (f <= numPCs){
        sampleDist <- dist(PCs[subsample,1:f])
        clusterRunsRIMat[k,7+2*numFeatures+f] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)

        # Cumulative PC + alignment score clustering
#         sampleDist <- combineDists(sampleDist,alignmentScoreDist)
#         clusterRunsRIMat[k,7+2*numFeatures+maxCumulativePCs+f] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)
      }
    }
  }
  
  # Store means and SEMs for completed processes
  comparisonMeans <- apply(clusterRunsRIMat,2,mean,na.rm=T)
  names(comparisonMeans) <- columnNamesAbsolute
  comparisonSEMs <- apply(clusterRunsRIMat,2,sem,na.rm=T)
  names(comparisonSEMs) <- columnNamesAbsolute
  
  comparisonData <- list(ComparisonMeans=comparisonMeans,ComparisonSEMs=comparisonSEMs)
  return(comparisonData)
}

####
# Just clusters by whatever features are in the featureVectors data frame
clusterTypesSimple <- function(specifications,typeMetaData,featureVectors,methodClass=c("hclust","kmeans","em"),method="ward",clusterMetric=AdjustedRandIndex,minGroupSize=30,numClusterRuns=20,sampleProportionally=T,sampleSize=if (sampleProportionally) numGroups*minGroupSize else minGroupSize,debugFlag=1,random=F,minBifs=0,usePCs=F,removeOutliers=F){
  methodClass <- match.arg(methodClass)
  numGroups <- length(specifications)

  if (minBifs > 0){
    typeMetaData <- subset(typeMetaData,SeqLen>=minBifs)
    featureVectors <- as.data.frame(featureVectors[rownames(featureVectors)%in%rownames(typeMetaData),])
  }

  typeMetaData <- typeMetaData[rownames(featureVectors),]
  
  if (random){
    for (i in 1:ncol(featureVectors)){
      featureVectors[,i] <- featureVectors[sample(1:nrow(featureVectors),nrow(featureVectors),replace=F),i]
    }
  }
  
  if (usePCs){
    pcObj <- prcomp(featureVectors,scale.=T)
    featureVectors <- pcObj$x
  }
  
  # Remove outliers
  if (removeOutliers){
    for (i in 1:ncol(featureVectors)){
      Q1 <- quantile(featureVectors[,i],.25)
      Q3 <- quantile(featureVectors[,i],.75)
      iqr <- Q3-Q1
      featureVectors <- featureVectors[featureVectors[,i]>(Q1-3*iqr)&featureVectors[,i]<(Q3+3*iqr),]
    }
  }
  
  
  # Get sequences of each group
  groupSequences <- list()
  allSequences <- c()
  specStrings <- c()
  groupsTooSmall <- c()
  neuriteClasses <- c()
  for (i in 1:numGroups){
    groupSequences[[i]] <- rownames(getSubset2(typeMetaData,specifications[[i]],typeMetaData))
    neuriteClassesSub <- rep(i,length(groupSequences[[i]]))
    names(neuriteClassesSub) <- groupSequences[[i]]
    neuriteClasses <- append(neuriteClasses,neuriteClassesSub)
    specStrings[i] <- getSetString(specifications[[i]])
    # Make sure all groups are large enough
    if (length(groupSequences[[i]]) < minGroupSize){
      warning(paste("Specification ",specStrings[i]," too small: ",length(groupSequences[[i]])," (min of ",minGroupSize,")\n",sep=""))
      groupsTooSmall <- append(groupsTooSmall,i)
    }
    else{
      allSequences <- append(allSequences,groupSequences[[i]])
    }
  }
  
  if (is.na(sampleSize)){
    sampleSize <- min(80,floor(0.75*min(sapply(groupSequences,length))))
  }
  
  if (length(groupsTooSmall > 0)){
    # If one or no groups remain, must stop
    if (numGroups - length(groupsTooSmall) < 2){
      warning(paste(numGroups - length(groupsTooSmall)," groups remaing large enough to run test",sep=""))
      return(NA)
    }
    # Remove groups that are too small
    else {
      specifications <- specifications[-groupsTooSmall]
      groupSequences <- groupSequences[-groupsTooSmall]
      specStrings <- specStrings[-groupsTooSmall]
      numGroups <- length(specifications)
    }
  }
  
  # Check to make sure no overlap between groups
  for (i in 1:(numGroups-1)){
    for (j in (i+1):numGroups){ 
      if (sum(groupSequences[[i]]%in%groupSequences[[j]]) > 0){
        stop("Overlap between specifications")
      }
    }
  }
  
  ### Preclustering data set up ###
  
  # Set feature vector type data frame to contain only the relevant data
  featureVectors <- featureVectors[allSequences,]
  
  # Helper values for the number of columns
  numFeatures <- ncol(featureVectors)
  
  # Get Z-scores of feature matrix
  standardizedFeatureVects <- as.data.frame(scale(featureVectors))
  rownames(standardizedFeatureVects) <- rownames(featureVectors)
  
  ### Cluster Analysis ###
  
  # Each feature
  numColumns <- numFeatures
  comparisonMeans <- rep(NA,numColumns)
  comparisonSEMs <- rep(NA,numColumns)
  columnNamesAbsolute <- colnames(featureVectors)
  names(comparisonMeans) <- columnNamesAbsolute
  names(comparisonSEMs) <- columnNamesAbsolute
  
  # Set up matrix for storing Rand Index across subsample runs
  clusterRunsRIMat <- matrix(0,nrow=numClusterRuns,ncol=numColumns)
  
  subsamples <- list()
  classPartitions <- list()
  
  numCores <- detectCores()

  for (k in 1:numClusterRuns){
    # Generate sample set
    if (sampleProportionally){
      classPartition <- rep(0,sampleSize)
      subsample <- sample(allSequences,sampleSize,replace=F)
      # Determine which group each sample element is a part
      for (i in 1:numGroups){
        classPartition[subsample%in%groupSequences[[i]]] <- i
      }
    }
    else{
      subsample <- c()
      classPartition <- c()
      for (i in 1:numGroups){
        subsample <- append(subsample,sample(groupSequences[[i]],sampleSize,replace=F))
        classPartition <- append(classPartition,rep(i,sampleSize))
      }
    }
    subsamples[[k]] <- subsample
    classPartitions[[k]] <- classPartition
  }
  
  # create samples and run alignments scores (to determine whether LTS or STL is better)
#  if (numCores > 1){
#    results <- mclapply(subsamples,runSimpleSub)
#  }
  
  # Run the rest of the cluster analyses
  
  for (k in 1:numClusterRuns){
    if (debugFlag > 0) print(paste("Running subsample",k))
    subsample <- subsamples[[k]]
    classPartition <- classPartitions[[k]]
    
    # Features
    for (f in 1:numFeatures){
      # Individual feature clustering
      #sampleDist <- dist(featureVectors[subsample,f])
      clusterRunsRIMat[k,f] <- clusterAndScoreFeatures(featureVectors[subsample,f],classPartition,methodClass=methodClass,method=method,clusterMetric=clusterMetric)
      #print(paste(colnames(featureVectors)[f],clusterRunsRIMat[k,f]))
    }
  }
  
  # Store means and SEMs for completed processes
  comparisonMeans <- apply(clusterRunsRIMat,2,mean,na.rm=T)
  names(comparisonMeans) <- columnNamesAbsolute
  comparisonSEMs <- apply(clusterRunsRIMat,2,sem,na.rm=T)
  names(comparisonSEMs) <- columnNamesAbsolute
  
  if (usePCs){
    comparisonData <- list(ComparisonMeans=comparisonMeans,ComparisonSEMs=comparisonSEMs,PCData=pcObj)
  }else{
    comparisonData <- list(ComparisonMeans=comparisonMeans,ComparisonSEMs=comparisonSEMs)
  }
  return(comparisonData)
}

clusterTypesMerged <- function(specifications,typeMetaData,featureVectors,alignmentDistMat,methodClass=c("hclust","kmeans","em"),method="ward",clusterMetric=AdjustedRandIndex,minGroupSize=30,numClusterRuns=20,sampleProportionally=T,sampleSize=if (sampleProportionally) numGroups*minGroupSize else minGroupSize,debugFlag=1,random=F,minBifs=0,usePCs=F,removeOutliers=F){
  methodClass <- match.arg(methodClass)
  numGroups <- length(specifications)
  
  # Make sure all data points are shared
  knownSequences <- intersect(rownames(featureVectors),rownames(alignmentDistMat))
  
  # Get sequences of each group
  groupSequences <- list()
  allSequences <- c()
  specStrings <- c()
  groupsTooSmall <- c()
  neuriteClasses <- c()
  for (i in 1:numGroups){
    groupSequences[[i]] <- intersect(knownSequences,rownames(getSubset2(typeMetaData,specifications[[i]],typeMetaData,minLength=minBifs)))
    neuriteClassesSub <- rep(i,length(groupSequences[[i]]))
    names(neuriteClassesSub) <- groupSequences[[i]]
    neuriteClasses <- append(neuriteClasses,neuriteClassesSub)
    specStrings[i] <- getSetString(specifications[[i]])
    # Make sure all groups are large enough
    if (length(groupSequences[[i]]) < minGroupSize){
      warning(paste("Specification ",specStrings[i]," too small: ",length(groupSequences[[i]])," (min of ",minGroupSize,")\n",sep=""))
      groupsTooSmall <- append(groupsTooSmall,i)
    }
    else{
      allSequences <- append(allSequences,groupSequences[[i]])
    }
  }

  featureVectors <- featureVectors[allSequences,]
  alignmentDistMat <- alignmentDistMat[allSequences,allSequences]
  #typeMetaData <- typeMetaData[rownames(featureVectors),]

  # Get alignment MDS
  alignmentMDS <- cmdscale(alignmentDistMat,k=3)
  # Scale alignment components so the first component has SD=1
  alignmentMDS <- alignmentMDS/sd(alignmentMDS[,1])
  
  if (random){
    for (i in 1:ncol(featureVectors)){
      featureVectors[,i] <- featureVectors[sample(1:nrow(featureVectors),nrow(featureVectors),replace=F),i]
    }
    for (i in 1:ncol(alignmentMDS)){
      alignmentMDS[,i] <- alignmentMDS[sample(1:nrow(alignmentMDS),nrow(alignmentMDS),replace=F),i]
    }
  }
  
  if (usePCs){
    pcObj <- prcomp(featureVectors,scale.=T)
    featureVectors <- pcObj$x
  }
  
  # Remove outliers, using 4 times inter-quartile range since being done on many samples
  if (removeOutliers){
    outlierList <- c()
    for (i in 1:ncol(featureVectors)){
      Q1 <- quantile(featureVectors[,i],.25)
      Q3 <- quantile(featureVectors[,i],.75)
      iqr <- Q3-Q1
      notOutliers <- featureVectors[,i]>(Q1-4*iqr)&featureVectors[,i]<(Q3+4*iqr)
      if (sum(!notOutliers) > 0){
        print(paste("Outliers in feature",i))
      }
      outlierList <- append(outlierList,rownames(featureVectors)[!notOutliers])
      featureVectors <- featureVectors[notOutliers,]
      alignmentMDS <- alignmentMDS[notOutliers,]
    }
    for (i in 1:ncol(alignmentMDS)){
      Q1 <- quantile(alignmentMDS[,i],.25)
      Q3 <- quantile(alignmentMDS[,i],.75)
      iqr <- Q3-Q1
      notOutliers <- alignmentMDS[,i]>(Q1-4*iqr)&alignmentMDS[,i]<(Q3+4*iqr)
      if (sum(!notOutliers) > 0){
        print(paste("Outliers in component",i))
      }
      outlierList <- append(outlierList,rownames(featureVectors)[!notOutliers])
      featureVectors <- featureVectors[notOutliers,]
      alignmentMDS <- alignmentMDS[notOutliers,]
    }
    if (length(outlierList) > 0){
      print("Outliers: ",paste(outlierList,collapse=","),sep="")
    }
  }
  
  if (is.na(sampleSize)){
    sampleSize <- min(80,floor(0.75*min(sapply(groupSequences,length))))
  }
  
  if (length(groupsTooSmall > 0)){
    # If one or no groups remain, must stop
    if (numGroups - length(groupsTooSmall) < 2){
      warning(paste(numGroups - length(groupsTooSmall)," groups remaing large enough to run test",sep=""))
      return(NA)
    }
    # Remove groups that are too small
    else {
      specifications <- specifications[-groupsTooSmall]
      groupSequences <- groupSequences[-groupsTooSmall]
      specStrings <- specStrings[-groupsTooSmall]
      numGroups <- length(specifications)
    }
  }
  
  # Check to make sure no overlap between groups
  for (i in 1:(numGroups-1)){
    for (j in (i+1):numGroups){ 
      if (sum(groupSequences[[i]]%in%groupSequences[[j]]) > 0){
        stop("Overlap between specifications")
      }
    }
  }
  
  ### Preclustering data set up ###
  
  # Helper values for the number of columns
  numFeatures <- ncol(featureVectors)
  
  # Get Z-scores of feature matrix
  standardizedFeatureVects <- as.data.frame(scale(featureVectors))
  rownames(standardizedFeatureVects) <- rownames(featureVectors)
  
  ### Cluster Analysis ###
  
  # Each feature
  numColumns <- numFeatures
  comparisonMeans <- rep(NA,numColumns)
  comparisonSEMs <- rep(NA,numColumns)
  columnNamesAbsolute <- colnames(featureVectors)
  names(comparisonMeans) <- columnNamesAbsolute
  names(comparisonSEMs) <- columnNamesAbsolute
  
  # Set up matrix for storing Rand Index across subsample runs
  clusterRunsRIMat <- matrix(0,nrow=numClusterRuns,ncol=numColumns)
  
  subsamples <- list()
  classPartitions <- list()
  
  numCores <- detectCores()
  
  for (k in 1:numClusterRuns){
    # Generate sample set
    if (sampleProportionally){
      classPartition <- rep(0,sampleSize)
      subsample <- sample(allSequences,sampleSize,replace=F)
      # Determine which group each sample element is a part
      for (i in 1:numGroups){
        classPartition[subsample%in%groupSequences[[i]]] <- i
      }
    }
    else{
      subsample <- c()
      classPartition <- c()
      for (i in 1:numGroups){
        subsample <- append(subsample,sample(groupSequences[[i]],sampleSize,replace=F))
        classPartition <- append(classPartition,rep(i,sampleSize))
      }
    }
    subsamples[[k]] <- subsample
    classPartitions[[k]] <- classPartition
  }
  
  # create samples and run alignments scores (to determine whether LTS or STL is better)
  #  if (numCores > 1){
  #    results <- mclapply(subsamples,runSimpleSub)
  #  }
  
  # Run the rest of the cluster analyses
  
  for (k in 1:numClusterRuns){
    if (debugFlag > 0) print(paste("Running subsample",k))
    subsample <- subsamples[[k]]
    classPartition <- classPartitions[[k]]
    
    # Features
    for (f in 1:numFeatures){
      # Individual feature clustering
      #sampleDist <- dist(featureVectors[subsample,f])
      mergedData <- cbind(alignmentMDS,featureVectors[,f])
      clusterRunsRIMat[k,f] <- clusterAndScoreFeatures(mergedData[subsample,],classPartition,methodClass=methodClass,method=method,clusterMetric=clusterMetric)
      #print(paste(colnames(featureVectors)[f],clusterRunsRIMat[k,f]))
    }
  }
  
  # Store means and SEMs for completed processes
  comparisonMeans <- apply(clusterRunsRIMat,2,mean,na.rm=T)
  names(comparisonMeans) <- columnNamesAbsolute
  comparisonSEMs <- apply(clusterRunsRIMat,2,sem,na.rm=T)
  names(comparisonSEMs) <- columnNamesAbsolute
  
  if (usePCs){
    comparisonData <- list(ComparisonMeans=comparisonMeans,ComparisonSEMs=comparisonSEMs,PCData=pcObj)
  }else{
    comparisonData <- list(ComparisonMeans=comparisonMeans,ComparisonSEMs=comparisonSEMs)
  }
  return(comparisonData)
}

####
# Just clusters by whatever features are in the featureVectors data frame
clusterTypesBestNmerAndMetricsSimple <- function(specifications,typeMetaData,nmers,metrics,bestNmer,method="ward",clusterMetric=AdjustedRandIndex,minGroupSize=30,numClusterRuns=20,sampleProportionally=T,sampleSize=if (sampleProportionally) numGroups*minGroupSize else minGroupSize,debugFlag=1,random=F,minBifs=0,usePCs=F){
  numGroups <- length(specifications)
  
  if (minBifs > 0){
    typeMetaData <- subset(typeMetaData,SeqLen>=minBifs)
    featureVectors <- as.data.frame(featureVectors[rownames(featureVectors)%in%rownames(typeMetaData),])
  }
  
  typeMetaData <- typeMetaData[rownames(featureVectors),]
  
#   if (random){
#     for (i in 1:ncol(featureVectors)){
#       featureVectors[,i] <- featureVectors[sample(1:nrow(featureVectors),nrow(featureVectors),replace=F),i]
#     }
#   }
#   
#   if (usePCs){
#     pcObj <- prcomp(featureVectors,scale.=T)
#     featureVectors <- pcObj$x
#   }

  # Get sequences of each group
  groupSequences <- list()
  allSequences <- c()
  specStrings <- c()
  groupsTooSmall <- c()
  neuriteClasses <- c()
  for (i in 1:numGroups){
    groupSequences[[i]] <- rownames(getSubset2(typeMetaData,specifications[[i]],typeMetaData))
    neuriteClassesSub <- rep(i,length(groupSequences[[i]]))
    names(neuriteClassesSub) <- groupSequences[[i]]
    neuriteClasses <- append(neuriteClasses,neuriteClassesSub)
    specStrings[i] <- getSetString(specifications[[i]])
    # Make sure all groups are large enough
    if (length(groupSequences[[i]]) < minGroupSize){
      warning(paste("Specification ",specStrings[i]," too small: ",length(groupSequences[[i]])," (min of ",minGroupSize,")\n",sep=""))
      groupsTooSmall <- append(groupsTooSmall,i)
    }
    else{
      allSequences <- append(allSequences,groupSequences[[i]])
    }
  }
  
  if (length(groupsTooSmall > 0)){
    # If one or no groups remain, must stop
    if (numGroups - length(groupsTooSmall) < 2){
      warning(paste(numGroups - length(groupsTooSmall)," groups remaing large enough to run test",sep=""))
      return(NA)
    }
    # Remove groups that are too small
    else {
      specifications <- specifications[-groupsTooSmall]
      groupSequences <- groupSequences[-groupsTooSmall]
      specStrings <- specStrings[-groupsTooSmall]
      numGroups <- length(specifications)
    }
  }
  
  # Check to make sure no overlap between groups
  for (i in 1:(numGroups-1)){
    for (j in (i+1):numGroups){ 
      if (sum(groupSequences[[i]]%in%groupSequences[[j]]) > 0){
        stop("Overlap between specifications")
      }
    }
  }
  
  ### Preclustering data set up ###
  
  # Set feature vector type data frame to contain only the relevant data
  featureVectors <- metrics
  featureVectors <- featureVectors[allSequences,]
  
  # Helper values for the number of columns
  numFeatures <- ncol(featureVectors)
  
  # Get Z-scores of feature matrix
  standardizedFeatureVects <- as.data.frame(scale(featureVectors))
  rownames(standardizedFeatureVects) <- rownames(featureVectors)
  
  ### Cluster Analysis ###
  
  # Each feature
  numColumns <- numFeatures
  comparisonMeans <- rep(NA,numColumns)
  comparisonSEMs <- rep(NA,numColumns)
  columnNamesAbsolute <- colnames(featureVectors)
  names(comparisonMeans) <- columnNamesAbsolute
  names(comparisonSEMs) <- columnNamesAbsolute
  
  # Set up matrix for storing Rand Index across subsample runs
  clusterRunsRIMat <- matrix(0,nrow=numClusterRuns,ncol=numColumns)
  
  subsamples <- list()
  classPartitions <- list()
  
  # create samples and run alignments scores (to determine whether LTS or STL is better)
  for (k in 1:numClusterRuns){
    # Generate sample set
    if (sampleProportionally){
      classPartition <- rep(0,sampleSize)
      subsample <- sample(allSequences,sampleSize,replace=F)
      # Determine which group each sample element is a part
      for (i in 1:numGroups){
        classPartition[subsample%in%groupSequences[[i]]] <- i
      }
    }
    else{
      subsample <- c()
      classPartition <- c()
      for (i in 1:numGroups){
        subsample <- append(subsample,sample(groupSequences[[i]],sampleSize,replace=F))
        classPartition <- append(classPartition,rep(i,sampleSize))
      }
    }
    subsamples[[k]] <- subsample
    classPartitions[[k]] <- classPartition
  }
  
  # Run the rest of the cluster analyses
  for (k in 1:numClusterRuns){
    if (debugFlag > 0) print(paste("Running subsample",k))
    subsample <- subsamples[[k]]
    classPartition <- classPartitions[[k]]
    
    # Features
    for (f in 1:numFeatures){
      # Individual feature clustering
      sampleDist <- dist(data.frame(featureVectors[subsample,f],nmers[subsample,bestNmer]))
      clusterRunsRIMat[k,f] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)
      #print(paste(colnames(featureVectors)[f],clusterRunsRIMat[k,f]))
    }
  }
  
  # Store means and SEMs for completed processes
  comparisonMeans <- apply(clusterRunsRIMat,2,mean,na.rm=T)
  names(comparisonMeans) <- columnNamesAbsolute
  comparisonSEMs <- apply(clusterRunsRIMat,2,sem,na.rm=T)
  names(comparisonSEMs) <- columnNamesAbsolute
  
  if (usePCs){
    comparisonData <- list(ComparisonMeans=comparisonMeans,ComparisonSEMs=comparisonSEMs,PCData=pcObj)
  }else{
    comparisonData <- list(ComparisonMeans=comparisonMeans,ComparisonSEMs=comparisonSEMs)
  }
  return(comparisonData)
}

getSpecificationGroups <- function(specifications,typeMetaData,minBifs=0){
  groupSequences <- list()
  numGroups <- length(specifications)
  for (i in 1:numGroups){
    groupSequences[[i]] <- rownames(getSubset2(typeMetaData,specifications[[i]],typeMetaData,minLength=minBifs))
  }
  return(groupSequences)
}

##################################################
# Cluster specification sets by:
#    All NMer PRs
#    All NMer PRs + All Morphometrics
#    Individual NMers
#    Cumulative best NMers
#    NMer cumulative PCs, ordered by ability to cluster
#    Individual features + best NMer PRs
#    Individual features + best NMer PCs
clusterTypesNMers <- function(specifications,typeMetaData,morphometrics,nmers,method="ward",clusterMetric=AdjustedRandIndex,minGroupSize=30,numClusterRuns=20,sampleProportionally=T,sampleSize=if (sampleProportionally) numGroups*minGroupSize else minGroupSize,maxCumulativePCs=6,debugFlag=1){
  numGroups <- length(specifications)
  
  typeMetaData <- typeMetaData[rownames(morphometrics),]
  
  # Get sequences of each group
  groupSequences <- list()
  allSequences <- c()
  specStrings <- c()
  groupsTooSmall <- c()
  neuriteClasses <- c()
  for (i in 1:numGroups){
    groupSequences[[i]] <- rownames(getSubset2(typeMetaData,specifications[[i]],typeMetaData))
    neuriteClassesSub <- rep(i,length(groupSequences[[i]]))
    names(neuriteClassesSub) <- groupSequences[[i]]
    neuriteClasses <- append(neuriteClasses,neuriteClassesSub)
    specStrings[i] <- getSetString(specifications[[i]])
    # Make sure all groups are large enough
    if (length(groupSequences[[i]]) < minGroupSize){
      warning(paste("Specification ",specStrings[i]," too small: ",length(groupSequences[[i]])," (min of ",minGroupSize,")\n",sep=""))
      groupsTooSmall <- append(groupsTooSmall,i)
    }else{
      allSequences <- append(allSequences,groupSequences[[i]])
    }
  }
  
  if (length(groupsTooSmall > 0)){
    # If one or no groups remain, must stop
    if (numGroups - length(groupsTooSmall) < 2){
      warning(paste(numGroups - length(groupsTooSmall)," groups remaing large enough to run test",sep=""))
      return(NA)
    }
    # Remove groups that are too small
    else {
      specifications <- specifications[-groupsTooSmall]
      groupSequences <- groupSequences[-groupsTooSmall]
      specStrings <- specStrings[-groupsTooSmall]
      numGroups <- length(specifications)
    }
  }
  
  # Check to make sure no overlap between groups
  for (i in 1:(numGroups-1)){
    for (j in (i+1):numGroups){ 
      if (sum(groupSequences[[i]]%in%groupSequences[[j]]) > 0){
        stop("Overlap between specifications")
      }
    }
  }
  
  ### Preclustering data set up ###
  
  # Set all feature vector type data frames to contain only the relevant data
  morphometrics <- morphometrics[allSequences,]
  nmers <- nmers[allSequences,]

  # Get Z-scores of morphometrics
  scaledMorphometrics <- scale(morphometrics)
  rownames(scaledMorphometrics) <- rownames(morphometrics)

  scaledNmers <- scale(nmers)
  rownames(scaledNmers) <- rownames(nmers)
  
  nmersAndMorpho <- cbind(scaledMorphometrics,scaledNmers)
  
  # Get morphometric feature principal components - avoids giving excessive weight to overrepresented features
  # First need to ensure all features have variance
  noVar <- which(apply(morphometrics,2,var)==0)
  if (length(noVar) > 0){
    warning(paste("Zero variance found in features:",paste(colnames(featureVectors)[noVar],collapse=", "),sep=" "))
    morphometrics <- morphometrics[-noVar]
  }
  pcObj <- prcomp(featureVectors,scale.=T)
  PCs <- pcObj$x
  
  numMetrics <- ncol(morphometrics)
  numNmers <- ncol(nmers)
  
  # Determine number of useful PCs
#  print("Determining num PCs for feature vector")
#  numPCs <- determineNumSignificantPCs(pcObj,originalData=featureVectors)
#  PCs <- PCs[,1:numPCs]
  
  # Get order of PCs best for clustering - using all of them
#  pcOrder <- bestPCClusterOrder(pcObj,neuriteClasses,pcStop=numPCs)
#  numPCs <- min(numPCs,maxCumulativePCs)
#  pcOrder <- pcOrder[1:numPCs]
  # Reorder PCs based on ability to cluster
#  PCs <- PCs[,pcOrder]
  
  # Get NMer feature principal components
  noVar <- which(apply(nmers,2,var)==0)
  if (length(noVar) > 0){
    warning(paste("Zero variance found in nmers:",paste(colnames(nmers)[noVar],collapse=", "),sep=" "))
    nmers <- nmers[-noVar]
  }
  pcObj <- prcomp(nmers,scale.=T) 
  # Determine number of useful PCs
  print("Determining num PCs for nmers")
#  numNMerPCs <- determineNumSignificantPCs(pcObj,originalData=nmers)
#  nmerPCs <- pcObj$x[,1:numNMerPCs]
  nmerPCs <- pcObj$x

  # Get order of nmer PCs best for clustering - using all of them
  nmerPCOrder <- bestPCClusterOrder(pcObj,neuriteClasses,numNmers,pairs=F)
  numNmerPCs <- maxCumulativePCs
  pcOrder <- pcOrder[1:numNmerPCs]
  # Reorder PCs based on ability to cluster
  nmerPCs <- nmerPCs[,pcOrder]
  
  ### Cluster Analysis ###
  
  # All Nmer PRs,
  # All Nmer PRs + all morphometrics
  # Individual NMers,
  # NMer cumulative PCs, 
  # Individual morphometrics + best NMer PRs
  # Individual morphometrics + best NMer PCs
  numColumns <- 2 + numNmers + maxCumulativePCs + 2*numMetrics
  comparisonMeans <- rep(NA,numColumns)
  comparisonPVals <- rep(NA,numColumns)
  comparisonSEMs <- rep(NA,numColumns)
  columnNamesAbsolute <- c("Nmers","Nmers&Morphos",colnames(nmers),
                           paste("CumulNmerPCs",1:maxCumulativePCs,sep=""),paste(colnames(morphometrics),"+BestNmers",sep=""),paste(colnames(morphometrics),"+BestNmerPCs",sep=""))
  names(comparisonMeans) <- columnNamesAbsolute
  names(comparisonSEMs) <- columnNamesAbsolute
  
  # Set up matrix for storing Rand Index across subsample runs
  clusterRunsRIMat <- matrix(0,nrow=numClusterRuns,ncol=numColumns)
  #clusterRunsBaselineMat <- matrix(0,nrow=numClusterRuns,ncol=numColumns)
  
  subsamples <- list()
  classPartitions <- list()
  
  # create samples
  for (k in 1:numClusterRuns){
    # Generate sample set
    if (sampleProportionally){
      classPartition <- rep(0,sampleSize)
      subsample <- sample(allSequences,sampleSize,replace=F)
      # Determine which group each sample element is a part
      for (i in 1:numGroups){
        classPartition[subsample%in%groupSequences[[i]]] <- i
      }
    }
    else{
      subsample <- c()
      classPartition <- c()
      for (i in 1:numGroups){
        subsample <- append(subsample,sample(groupSequences[[i]],sampleSize,replace=F))
        classPartition <- append(classPartition,rep(i,sampleSize))
      }
    }
    subsamples[[k]] <- subsample
    classPartitions[[k]] <- classPartition
    
    numClasses <- length(unique(classPartition))
  }
  
  # Run the rest of the cluster analyses
  for (k in 1:numClusterRuns){
    print(paste("Running subsample",k))
    subsample <- subsamples[[k]]
    classPartition <- classPartitions[[k]]
    
    # NMers
    sampleDist <- dist(scaledNmers[subsample,])
    clusterRunsRIMat[k,1] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)
    
    # NMers + Morphometrics
    sampleDist <- dist(nmersAndMorpho[subsample,])
    clusterRunsRIMat[k,2] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)
    
    # Features and cumulative feature PCs
    for (f in 1:numNmers){
      # Individual Nmer clustering
      sampleDist <- dist(nmers[subsample,f])
      clusterRunsRIMat[k,2+f] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)

      # Cumulative PC clustering
      if (f <= maxCumulativePCs){
        sampleDist <- dist(nmerPCs[subsample,1:f])
        clusterRunsRIMat[k,2+numNmers+f] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)
      }
    }
  }

  # Individual features + best Nmers; Individual features + best Nmer PCs
  nmerScores <- apply(clusterRunsRIMat[,2+1:numNmers],2,mean)
  bestNmerOrder <- order(nmerScores,decreasing=T)[1:maxCumulativePCs]
  # Get best individual nmers performing better than .1, up to maxCumulativePCs of them
  bestNmerOrder <- bestNmerOrder[nmerScores[bestNmerOrder] > .1]
  bestNmers <- colnames(nmers)[bestNmerOrder]
  bestNmerPCs <- which.max(apply(clusterRunsRIMat[,2+numNmers+1:maxCumulativePCs],2,mean))
  for (k in 1:numClusterRuns){
    subsample <- subsamples[[k]]
    classPartition <- classPartitions[[k]]
    
    for (f in 1:numMetrics){
      sampleDist <- dist(cbind(scaledMorphometrics[subsample,f],nmers[subsample,bestNmerOrder]))
      clusterRunsRIMat[k,2+numNmers+maxCumulativePCs+f] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)

      sampleDist <- dist(cbind(scaledMorphometrics[subsample,f],nmerPCs[subsample,1:bestNMerPCs]))
      clusterRunsRIMat[k,2+numNmers+maxCumulativePCs+numMetrics+f] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)
    }
  }
  
  # Store means and SEMs for completed processes
  comparisonMeans <- apply(clusterRunsRIMat,2,mean,na.rm=T)
  names(comparisonMeans) <- columnNamesAbsolute
  comparisonSEMs <- apply(clusterRunsRIMat,2,sem,na.rm=T)
  names(comparisonSEMs) <- columnNamesAbsolute
  
  comparisonData <- list(ComparisonMeans=comparisonMeans,ComparisonSEMs=comparisonSEMs,BestNmers=bestNmers)
  return(comparisonData)
}

# Pick out the best features at clustering
bestFeaturesClustering <- function(specifications,typeMetaData,morphometrics,nmers,method="ward",clusterMetric=AdjustedRandIndex,minGroupSize=30,numClusterRuns=20,sampleProportionally=T,sampleSize=if (sampleProportionally) numGroups*minGroupSize else minGroupSize,maxCumulativePCs=6,debugFlag=1){
  numGroups <- length(specifications)
  
  #typeMetaData <- typeMetaData[rownames(morphometrics),]
  
  # Get sequences of each group
  groupSequences <- list()
  allSequences <- c()
  specStrings <- c()
  groupsTooSmall <- c()
  neuriteClasses <- c()
  for (i in 1:numGroups){
    groupSequences[[i]] <- rownames(getSubset2(typeMetaData,specifications[[i]],typeMetaData))
    neuriteClassesSub <- rep(i,length(groupSequences[[i]]))
    names(neuriteClassesSub) <- groupSequences[[i]]
    neuriteClasses <- append(neuriteClasses,neuriteClassesSub)
    specStrings[i] <- getSetString(specifications[[i]])
    # Make sure all groups are large enough
    if (length(groupSequences[[i]]) < minGroupSize){
      warning(paste("Specification ",specStrings[i]," too small: ",length(groupSequences[[i]])," (min of ",minGroupSize,")\n",sep=""))
      groupsTooSmall <- append(groupsTooSmall,i)
    }else{
      allSequences <- append(allSequences,groupSequences[[i]])
    }
  }
  
  if (length(groupsTooSmall > 0)){
    # If one or no groups remain, must stop
    if (numGroups - length(groupsTooSmall) < 2){
      warning(paste(numGroups - length(groupsTooSmall)," groups remaing large enough to run test",sep=""))
      return(NA)
    }
    # Remove groups that are too small
    else {
      specifications <- specifications[-groupsTooSmall]
      groupSequences <- groupSequences[-groupsTooSmall]
      specStrings <- specStrings[-groupsTooSmall]
      numGroups <- length(specifications)
    }
  }
  
  # Check to make sure no overlap between groups
  for (i in 1:(numGroups-1)){
    for (j in (i+1):numGroups){ 
      if (sum(groupSequences[[i]]%in%groupSequences[[j]]) > 0){
        stop("Overlap between specifications")
      }
    }
  }
  
  ### Preclustering data set up ###
  
  # Set all feature vector type data frames to contain only the relevant data
  nmers <- nmers[allSequences,]
  morphometrics <- morphometrics[allSequences,]
  allFeatures <- cbind(nmers,morphometrics)

  # Find the best combinations of nmer features and their DBIs
  bestNmerFeatureSets <- bestFeatureSets(nmers,neuriteClasses)
  # Find the best combinations of morphometric features and their DBIs
  bestMorphometricFeatureSets <- bestFeatureSets(morphometrics,neuriteClasses)
  
  # Determine whether adding nmer feature sets improves any morphometric feature sets
  
  
  # Once the final features list is complete (additional features fail to improve any feature combinations), test actual clustering performance, reporting results of the best feature combinations, down to 90% of the top feature combination.
  # Find the best combinations of nmer features and their DBIs
  bestNmerFeatureSets <- bestFeatureSets(nmers,neuriteClasses)
  # Find the best combinations of morphometric features and their DBIs
  bestMorphometricFeatureSets <- bestFeatureSets(morphometrics,neuriteClasses)
  
  # Go through morphometric feature sets and get rid of below threshold sets
  metricSets <- bestMorphometricFeatureSets$FeatureSets
  metricSetValues <- bestMorphometricFeatureSets$FeatureSetDbis
  metricSets <- metricSets[metricSetValues>minDbiThreshold]
  
  # Determine whether adding nmer feature sets improves any morphometric feature sets
  
}

############################
# Representatively samples from given groups and compares result to randomized group assignment clustering
########################
clusterTypesCompareToRandom <- function(specifications,typeMetaData,featureVectors=NA,distanceMatrix=NA,numDimensions=5,methodClass=c("hclust","kmeans","em"),method="ward",clusterMetric=AdjustedRandIndex,minGroupSize=30,numClusterRuns=20,numRandomRuns=20,sampleProportion=NA,debugFlag=1,minBifs=0,removeOutliers=F){
  methodClass <- match.arg(methodClass)
  numGroups <- length(specifications)
  
  if (is.na(featureVectors) && is.na(distanceMatrix)){
    stop("One of featureVectors or distanceMatrix must contain data")
  }
  
  availableSequences <- rownames(subset(typeMetaData,SeqLen>=minBifs))
  if (!is.na(featureVectors)){
    availableSequences <- intersect(availableSequences,rownames(featureVectors))
  }
  if (!is.na(distanceMatrix)){
    availableSequences <- intersect(availableSequences,rownames(distanceMatrix))
  }
  
  # Get sequences of each group
  groupSequences <- list()
  allSequences <- c()
  specStrings <- c()
  groupsTooSmall <- c()
  neuriteClasses <- c()
  for (i in 1:numGroups){
    groupSequences[[i]] <- intersect(availableSequences,
                                     rownames(getSubset2(typeMetaData,specifications[[i]],typeMetaData)))
    neuriteClassesSub <- rep(i,length(groupSequences[[i]]))
    names(neuriteClassesSub) <- groupSequences[[i]]
    neuriteClasses <- append(neuriteClasses,neuriteClassesSub)
    specStrings[i] <- getSetString(specifications[[i]])
    # Make sure all groups are large enough
    if (length(groupSequences[[i]]) < minGroupSize){
      warning(paste("Specification ",specStrings[i]," too small: ",length(groupSequences[[i]])," (min of ",minGroupSize,")\n",sep=""))
      groupsTooSmall <- append(groupsTooSmall,i)
    }
    else{
      allSequences <- append(allSequences,groupSequences[[i]])
    }
  }

  # Set feature vector type data frame to contain only the relevant data
  if (!is.na(distanceMatrix)){
    require(MASS)
    distMDS <- isoMDS(distanceMatrix[allSequences,allSequences],k=numDimensions)$points
    if (!is.na(featureVectors)){
      featureVectors <- cbind(featureVectors,distMDS)
    }else{
      featureVectors <- distMDS
    }
  }
  
  # Remove outliers
  if (removeOutliers){
    for (i in 1:ncol(featureVectors)){
      Q1 <- quantile(featureVectors[,i],.25)
      Q3 <- quantile(featureVectors[,i],.75)
      iqr <- Q3-Q1
      notOutliers <- rownames(featureVectors)[featureVectors[,i]>(Q1-3*iqr)&featureVectors[,i]<(Q3+3*iqr)]
      allSequences <- intersect(allSequences,notOutliers)
      for (j in 1:length(groupSequences)){
        groupSequences[[j]] <- intersect(groupSequences[[j]])
      }
      featureVectors <- featureVectors[notOutliers,]
    }
  }
  
  # Remove any groups that are too small
  if (length(groupsTooSmall > 0)){
    # If one or no groups remain, must stop
    if (numGroups - length(groupsTooSmall) < 2){
      warning(paste(numGroups - length(groupsTooSmall)," groups remaing large enough to run test",sep=""))
      return(NA)
    }
    # Remove groups that are too small
    else {
      specifications <- specifications[-groupsTooSmall]
      groupSequences <- groupSequences[-groupsTooSmall]
      specStrings <- specStrings[-groupsTooSmall]
      numGroups <- length(specifications)
    }
  }

  # Determine the sample sizes
  if (is.na(sampleProportion)){
    sampleSizes <- c(lapply(groupSequences,length))
  }else{
    sampleSizes <- c(lapply(groupSequences,function(x,sampleProportion){return(round(sampleProportion&length(x)))},sampleProportion))
  }
  
  # Check to make sure no overlap between groups
  for (i in 1:(numGroups-1)){
    for (j in (i+1):numGroups){ 
      if (sum(groupSequences[[i]]%in%groupSequences[[j]]) > 0){
        stop("Overlap between specifications")
      }
    }
  }
  
  ### Preclustering data set up ###
  
  # Helper values for the number of columns
  numFeatures <- ncol(featureVectors)
  
  # Get Z-scores of feature matrix
  standardizedFeatureVects <- as.data.frame(scale(featureVectors))
  rownames(standardizedFeatureVects) <- rownames(featureVectors)
  
  ### Cluster Analysis ###
  
  subsamples <- list()
  classPartitions <- list()
  
  require(parallel)
  numCores <- detectCores()
  
  # Run subsamples of nonrandomly grouped
  realResults <- mclapply(1:numClusterRuns,getSampleKmeansCluster,
                         subsetMDS,groupSeqs,sampleSizes,randomizeGroups=F)
  realRIs <- c()
  for (i in 1:numClusterRuns){
    realRIs <- append(realRIs,realResults[[i]])
  }
  
  # Run subsamples of randomly grouped
  randomResults <- mclapply(1:numRandomRuns,getSampleKmeansCluster,
                        subsetMDS,groupSeqs,sampleSizes,randomizeGroups=T)
  randomRIs <- c()
  for (i in 1:numRandomRuns){
    randomRIs <- append(randomRIs,randomResults[[i]])
  }

  if (numClusterRuns == 1){
    return(list(Percentile=percentile(realRIs,randomRIs),RealRIs=realRIs,RandomRIs=randomRIs))
  }else{
    testResult <- wilcox.test(realRIs,randomRIs,alternative="greater")
    return(list(TestResult=testResult,MeanPercentileRank=percentileRank(mean(realRIs),randomRIs),RealRIs=realRIs,RandomRIs=randomRIs))
  }
}

getSampleKmeansCluster <- function(runNumber,dataMDS,groupSeqs,sampleSizes,randomizeGroups){
  subsampleSeqs <- c()
  groupIds <- c()
  
  # Randomly sample data
  for (i in 1:length(groupSeqs)){
    subsampleSeqs <- append(subsampleSeqs,groupSeqs[[i]][sample(1:length(groupSeqs[[i]]),sampleSizes[i],replace=F)])
    groupIds <- append(groupIds,rep(i,sampleSizes[i]))
  }
  
  # Randomize group identities
  if (randomizeGroups){
    groupIds <- sample(groupIds,length(groupIds),replace=F)
  }
  
  # Cluster
  kmeansObj <- kmeans(dataMDS[subsampleSeqs,],k=2)
  
  # Return aRI
  return(AdjustedRandIndex(kmeansObj$cluster,groupIds))
}

basicClusterResults <- function(features,partitions,minGroupSize=30,numClusterRuns=20,sampleProportionally=T,sampleSize=if (sampleProportionally) numGroups*minGroupSize else minGroupSize,method=c("ward","single","complete","average","mcquitty","median","centroid","kmeans"),clusterMetric=AdjustedRandIndex,plotClusters=0){
#  if (is.na(features)&&is.na(distMat)){
#    stop("At least one of features or distMat must be specified")
#  }
#distMat=if(!is.na(features)) dist(features) else NA
  method <- match.arg(method)
  
  riVals <- c()
  length(riVals) <- numClusterRuns
  
  numGroups <- length(unique(partitions))
  
  # create samples
  for (k in 1:numClusterRuns){
    # Generate sample set
    if (sampleProportionally){
      subsample <- sample(1:nrow(features),sampleSize,replace=F)
      # Determine which group each sample element is a part
    }else{
      subsample <- c()
      classPartition <- c()
      for (i in 1:numGroups){
        subsample <- append(subsample,sample(which(partitions==i),sampleSize,replace=F))
      }
    }
    classPartition <- partitions[subsample]
    
    if (method=="kmeans"){
      kmeansClust <- kmeans(features[subsample,],centers=numGroups)
      clusterPartition <- kmeansClust$cluster
      #riVals[k] <- kmeansAndScore(features[subsample,,drop=F],classPartition,method=method,clusterMetric=clusterMetric)
    }
    else{
      sampleDist <- dist(features[subsample,])
      sampleClust <- hclust(sampleDist,method=method)
      clusterPartition <- cutree(sampleClust,2)
      #riVals[k] <- clusterAndScoreDistances(sampleDist,classPartition,method=method,clusterMetric=clusterMetric)
    }
    riVals[k] <- clusterMetric(classPartition,clusterPartition)
    if (plotClusters && k%%floor(numClusterRuns/plotClusters)==0){
      allPch <- rep(21,length(partitions))
      allPch[subsample] <- 21+clusterPartition
      sampleCex <- rep(.8,length(partitions))
      sampleCex[subsample] <- 1.2
      allBG <- rep(0,length(partitions))
      allBG[subsample] <- partitions[subsample]
      if (ncol(features) == 1){
        plot(1:nrow(features),features[,1],col=partitions,bg=allBG,ylab=colnames(features),pch=allPch,cex=sampleCex)
      }
      else{
        plot(features[,1],features[,2],col=partitions,bg=allBG,pch=allPch,cex=sampleCex,xlab=colnames(features)[1],ylab=colnames(features)[2])
      }
    }
    
    #allPch <- rep(21,length(partitions))
    #allPch[subsample] <- 21+clusterPartition
    #plot(1:nrow(features),features[,1],col=partitions,pch=allPch,bg=partitions*((1:nrow(features))%in%subsample))
    
  }
  return(riVals)
}

bestFeatureSets <- function(features,partitions){
  featureSetList <- list()
  soloDbiList <- c()
  length(soloDbiList) <- ncol(features)
  for (i in 1:ncol(features)){
    if (i!=ncol(features)){ # Don't add the final solo feature to the initial featureSetList
      featureSetList[[i]] <- i
    }
    # Calculate Davies-Bouldin Index (DBI) for each individual feature
    soloDbiList[i] <- DaviesBouldinIndex(partition=partitions,featureVector=features[,i])
  }
  # Copy dbi scores to dbiList for current feature combinations (initially individuals)
  dbiList <- soloDbiList
  
  finalFeatureSets <- list()
  finalFeatureSetDbis <- c()
  # Put together to find best feature combinations
  repeat{ 
    newFeatureSetList <- list()
    newFeatureSetDbiList <- c()
    changes <- F
    # Go through the current set of features
    for (i in 1:length(featureSetList)){
      # Flag determining whether feature set is improved by adding any other feature
      featureSetNotImproved <- T 
      featureSet <- featureSetList[[i]]
      # Try adding any feature not already in the feature set, and that wouldn't just create a set that's already been tried
      for (j in (max(featureSet)+1):ncol(features)){
        tmpCombination <- append(featureSetList[[i]],j)
        dbi <- DaviesBouldinIndex(partition=partitions,featureVector=features[,tmpCombination])
        # If both the solo feature and the feature group are improved sufficiently, merge them
        if (dbi < 0.99*dbiList[i] && dbi < 0.99*soloDbiList[j]){
          featureSetNotImproved <- F
          changes <- T
          newFeatureSetList <- append(newFeatureSetList,tmpCombination)
          newFeatureSetDbiList <- append(newFeatureSetDbiList,dbi)
        }
      }
      if (featureSetNotImproved){
        finalFeatureSets <- append(finalFeatureSets,featureSetList[[i]])
        finalFeatureSetDbis <- append(finalFeatureSetDbis,dbiList[i])
      }
    }
    
    if (!changes){
      break;
    }
    # Move "new" lists into current lists
    featureSetList <- newFeatureSetList
    dbiList <- newFeatureSetDbiList
  }
  # Convert feature indices to feature names
  for (i in 1:length(featureSetList)){
    indexSet <- featureSetList[[i]]
    featureSetList[[i]] <- colnames(features)[indexSet]
  }
  return(list(FeatureSets=finalFeatureSets,FeatureSetDbis=finalFeatureSetDbis))
}

fixNmerTransform <- function(nmerData){
  sortedData <- sort(nmerData[!nmerData%in%c(-Inf,Inf)])
  if (-Inf%in%nmerData){
    lowest <- sortedData[1]
    lowest2 <- sortedData[2]
    nmerData[nmerData==-Inf] <- lowest - (lowest2-lowest)
  }
  if (Inf%in%nmerData){
    highest <- rev(sortedData)[1]
    highest2 <- rev(sortedData)[2]
    nmerData[nmerData==Inf] <- highest - (highest2-highest)
  }
  return(nmerData)
}

logTransformData <- function(myData,threshold=1,maxIterations=2){
  dataSkewness <- skewness(myData)
  if (abs(dataSkewness) > threshold){
    if (dataSkewness < 0){
      myData <- -myData + 1.5 + max(myData)
    }
    iterations <- 0
    repeat {
      if (min(myData,na.rm=T) <= 1){
        myData <- myData + (1-min(myData,na.rm=T))
      }
      myData <- log(myData)
      iterations <- iterations + 1
      dataSkewness <- skewness(myData)
      if (dataSkewness <= threshold || (maxIterations != -1 && iterations >= maxIterations)){
        break;
      }
    } 
  }
  return(myData)
}

################################################################################
# Functions related to a complete combinatorial exploration of possible classes
# 

# Gets the sets that overlap with the provided set (via set or setIndex).
# Whether overlap is directed or mutual is determined by the overlap input,
# which should be given as a numeric array indexed like a "dist" object.
getOverlappingSetsOfSet <- function(overlap,allSets,overlapThreshold=1,set=allSets[[setIndex]],setIndex=typeIndexMap[[getSetKey(set)]],typeIndexMap=getTypeIndexMap(allSets)){
  # Determine number of sets
  #numSets <- (1+sqrt(1+8*length(mutualOverlap)))/2
  numSets <- length(allSets)
  setsWLowerInds <- c()
  if (setIndex > 1){
    setsWLowerInds <- 1:(setIndex-1)+(setIndex-1)*(setIndex-2)/2
  }
  highInds <- (setIndex+1):numSets
  setsWHigherInds <- setIndex+(highInds-1)*(highInds-2)/2
  overlapInds <- c(setsWLowerInds,NA,setsWHigherInds)
  # Overlap values
  overlapVals <- overlap[overlapInds]
  highOverlapInds <- which(overlapVals > overlapThreshold)
  return(allSets[highOverlapInds])
}

#######################################################
