# Some stats can be derived by package "clValid" - dunn, silhouette, etc

getClusterComponents <- function(clusterRoot,clusterObj){
  clusterQueue <- c(clusterRoot)
  components <- c()
  while (length(clusterQueue) > 0){
    curPos <- clusterQueue[1]
    clusterQueue <- clusterQueue[-1]
    for (joinInd in 1:2){
      if (clusterObj$merge[curPos,joinInd] < 0){
        components <- c(components,-clusterObj$merge[curPos,joinInd])
      }
      else{
        clusterQueue <- c(clusterQueue,clusterObj$merge[curPos,joinInd])
      }
    }
  }
  return(components)
}

# Generates a,b,c,d values used by several cluster statistics
# a = number of observation pairs that are in the same class and in the same cluster
# b = number of observation pairs that are in the same class but in a different cluster
# c = number of observation pairs that are in a different class but in the same cluster
# d = number of observation pairs that are in different classes and clusters
calcClusterABCD <- function(classPartition,clusterPartition){
  aVal <- 0
  bVal <- 0
  cVal <- 0
  dVal <- 0
  numObs <- length(classPartition)
  for (i in 1:(numObs-1)){
    for (j in (i+1):numObs){
      if (classPartition[i]==classPartition[j]){
        if (clusterPartition[i]==clusterPartition[j]){
          aVal <- aVal+1
        }
        else{
          bVal <- bVal+1
        }
      }
      else{
        if (clusterPartition[i]==clusterPartition[j]){
          cVal <- cVal+1
        }
        else{
          dVal <- dVal+1
        }
      }
    }
  }
  return(c(a=aVal,b=bVal,c=cVal,d=dVal))
}

# Assumes all partitions are of the same size
SimpleAdjustedRandIndex <- function(classPartition,clusterPartition){
  if (length(classPartition) != length(clusterPartition)){
    stop("classPartition and clusterPartition are not the same length")
  }
  numPartitions <- length(unique(classPartition))
  return(
    (RandStat(classPartition,clusterPartition)-1/numPartitions)/(1-1/numPartitions)
    )
}
AdjustedRandIndex <- function(classPartition,clusterPartition){
  if (length(classPartition) != length(clusterPartition)){
    stop("classPartition and clusterPartition are not the same length")
  }
  numPartitions1 <- length(unique(classPartition))
  numPartitions2 <- length(unique(clusterPartition))
  numObs <- length(classPartition)
  # Sum(Choose(elements in class,2))
  classSum <- sum(choose(aggregate(classPartition,by=list(classPartition),FUN=length)$x,2))
  clusterSum <- sum(choose(aggregate(clusterPartition,by=list(clusterPartition),FUN=length)$x,2))
  totalSum <- choose(length(classPartition),2)
  maxIndex <- .5*(classSum+clusterSum)
  expectedIndex <- classSum*clusterSum/totalSum
  clusterABCD <- calcClusterABCD(classPartition,clusterPartition)
  return(
    (clusterABCD[[1]]-expectedIndex)/(maxIndex-expectedIndex)
    )
}
RandIndex <- function(classPartition,clusterPartition){
  if (length(classPartition) != length(clusterPartition)){
    stop("classPartition and clusterPartition are not the same length")
  }
  numObs <- length(classPartition)
  M <- numObs*(numObs-1)/2
  clusterABCD <- calcClusterABCD(classPartition,clusterPartition)
  aVal <- clusterABCD[[1]]
  dVal <- clusterABCD[[4]]
  return((aVal+dVal)/M)
}

JaccardCoefficient <- function(classPartition,clusterPartition){
  clusterABCD <- calcClusterABCD(classPartition,clusterPartition)
  aVal <- clusterABCD[[1]]
  bVal <- clusterABCD[[2]]
  cVal <- clusterABCD[[3]]
  return(aVal/(aVal+bVal+cVal))
}

FMIndex <- function(classPartition,clusterPartition){
  clusterABCD <- calcClusterABCD(classPartition,clusterPartition)
  aVal <- clusterABCD[[1]]
  bVal <- clusterABCD[[2]]
  cVal <- clusterABCD[[3]]
  return(sqrt((aVal/(aVal+bVal))*(aVal/(aVal+cVal))))
}
  
HubertStatistic <- function(classPartition,clusterPartition){
  numObs <- length(classPartition)
  M <- numObs*(numObs-1)/2
  statSum <- 0
  for (i in 1:(numObs-1)){
    for (j in (i+1):numObs){
      statSum <- statSum + (classPartition[i]==classPartition[j])*(clusterPartition[i]==clusterPartition[j])
    }
  }
  return(statSum/M)
}

NormalizedHubertStatistic <- function(classPartition,clusterPartition){
  numObs <- length(classPartition)
  M <- numObs*(numObs-1)/2
  statSum <- 0
  classMatrix <- matrix(0,numObs,numObs)
  clusterMatrix <- matrix(0,numObs,numObs)
  for (i in 1:numObs){
    for (j in i:numObs){
      classMatrix[i,j] <- classPartition[i]==classPartition[j]
      classMatrix[j,i] <- classMatrix[i,j]
      clusterMatrix[i,j] <- clusterMatrix[i]==clusterMatrix[j]
      clusterMatrix[j,i] <- clusterMatrix[i,j]
    }
  }
  classMatMean <- mean(classMatrix)
  clusterMatMean <- mean(clusterMatrix)
  classMatSD <- sd(classMatrix)[1]
  clusterMatSD <- sd(clusterMatrix)[1]
  for (i in 1:(numObs-1)){
    for (j in (i+1):numObs){
      statSum <- statSum + ((classPartition[i]==classPartition[j]) - classMatMean)*((clusterPartition[i]==clusterPartition[j]) - clusterMatMean)
    }
  }
  return(statSum/(M*classMatSD*clusterMatSD))
}
  
# Expects matched arrays of observations assigned to class and cluster
ClusterValidationError <- function(classPartition,clusterPartition){
  minVE <- 1
  numGroups <- length(unique(classPartition))
  numPartitions <- factorial(numGroups)
  numObs <- length(classPartition)
  availGroups <- 1:numGroups
  # Go through the number of possible partition alignments
  for (i in 0:(numPartitions-1)){
    groupOrder <- generateGroupOrder(i,availGroups)
    tmpClusters <- rep(0,numObs)
    for (j in 1:numGroups){
      tmpClusters[clusterPartition==j] <- groupOrder[j]
    }
    validationError <- mean(classPartition != tmpClusters)
    if (validationError < minVE){
      minVE <- validationError
    }
  }
  return(minVE)
}

# Generates a unique ordering of numbers 1:numGroups based on factVal
generateGroupOrder <- function(factVal,availGroups){
  if (length(availGroups) == 1){
    return(availGroups)
  }
  numGroups <- length(availGroups)
  curPos <- factVal%%numGroups+1
  curVal <- availGroups[curPos]
  availGroups <- availGroups[-curPos]
  nextFactVal <- floor(factVal/numGroups)+1
  return(c(curVal,generateGroupOrder(nextFactVal,availGroups)))
}

getValidationErrorCDF <- function(trueGroupCounts,numTrials=10000,precision=1/numTrials){
  numTrials <- 1/precision
  veDist <- rep(0,numTrials)
  numGroups <- length(trueGroupCounts)
  actualGroups <- rep(1,trueGroupCounts[1])
  for (j in 2:numGroups){
    actualGroups <- c(actualGroups,rep(j,trueGroupCounts[j]))
  }
  numObs <- sum(trueGroupCounts)

  for (k in 1:numTrials){
    # Assign observations to clusters
    randomClusters <- sample.int(numGroups,numObs,replace=T)
      
    # Calculate validation error
    veDist[k] <- ClusterValidationError(actualGroups,randomClusters)
  }
  return(ecdf(veDist))
}

ValidationErrorPWithGroups <- function(errorVal,groups,clusters=NULL,plotDist=F){
  trueGroupCounts <- aggregate(groups,by=list(groups),FUN=length)$x
  if (!is.null(clusters)){
    clusterCounts <- aggregate(clusters,by=list(clusters),FUN=length)$x
  }
  else{
    clusterCounts <- NULL
  }
  return(ValidationErrorP(errorVal,trueGroupCounts,clusterCounts,plotDist))
}
ValidationErrorP <- function(errorVal,trueGroupCounts,clusterCounts=NULL,numTrials=1000,plotDist=F){
  veDist <- rep(0,numTrials)
  numGroups <- length(trueGroupCounts)
  actualGroups <- rep(1,trueGroupCounts[1])
  for (j in 2:numGroups){
    actualGroups <- c(actualGroups,rep(j,trueGroupCounts[j]))
  }
  numObs <- sum(trueGroupCounts)

  for (k in 1:numTrials){
    # Assign observations to clusters
    if (is.null(clusterCounts)){
      clusters <- sample.int(numGroups,numObs,replace=T)
    }
    else{
      clusters <- rep(1,clusterCounts[1])
      for (j in 2:numGroups){
        clusters <- c(clusters,rep(j,clusterCounts[j]))
      }
      clusters <- sample(clusters)
    }
      
    # Calculate validation error
    veDist[k] <- ClusterValidationError(actualGroups,clusters)
  }
  if (plotDist){
    veEcdf <- ecdf(veDist)
    plot(veEcdf)
    points(errorVal,veEcdf(errorVal),col=2,cex=1.5)
  }
  return(veEcdf(errorVal))
}

# Calculates the average cluster error between two disjoint subsets
stability <- function(distanceMatrix,numClusters=2,iterations=20){
  errVals <- c(NA)
  length(errVals) <- iterations
  n <- ncol(distanceMatrix)
  sampleSize <- floor(n/2)
  for (i in 1:iterations){
    subsetIndices <- sample(1:ncol(distanceMatrix),n)
    subsetIndices1 <- subsetIndices[1:sampleSize]
    subsetIndices2 <- subsetIndices[(sampleSize+1):(sampleSize*2)]
    set1 <- distanceMatrix[subsetIndices1,subsetIndices1]
    set2 <- distanceMatrix[subsetIndices2,subsetIndices2]
    partition1 <- cutree(hclust(set1),numClusters)
    partition2 <- cutree(hclust(set2),numClusters)
    errVals[i] <- ClusterValidationError(partition1,partition2)
  }
  return(mean(errVals))
}

# Takes _iterations_ subsets of the distance matrix, each of a given size, clusters, and calculates the
# Stability. Subset size is the max of N/2 and 25 (or 80% of N if N < 25).
# Returns a distribution of scores given your chosen cluster statistic (set in FUN)
clusterStatOfSubset <- function(distanceMatrix,groupPartition,iterations=20,subsetSize=max(ncol(distanceMatrix)/2,min(25,.8*ncol(distanceMatrix))),FUN=c(ClusterValidationError,RandStat,JaccardCoefficient,FMIndex,HubertStatistic,NormalizedHubertStatistic)){
  FUN <- match.arg(FUN)
  numClusters <- length(unique(groupPartition))
  statDist <- c(NA)
  length(statDist) <- iterations
  for (i in 1:iterations){
    subsetIndices <- sample(1:ncol(distanceMatrix),subsetSize)
    DMSubset <- distanceMatrix[subsetIndices,subsetIndices]
    subsetClust <- hclust(DMSubset)
    subsetPartition <- cutree(subsetClust,numClusters)
    statDist[i] <- FUN(subsetPartition,groupPartition)
  }
  return(statDist)
}

#################################################
# Internal evaluation
#################################################

# Elementary functions

# Assumes "partition" contains only two unique partition labels by taking the first two
interclusterCentroidDistance <- function(partition,featureVector,partitionIndices=c(1,2),p=2){
  partitionLabels <- unique(partition)
  if (length(partitionLabels) < 2){
    stop("Fewer than 2 partitions in parameter 'partition'")
  }
  center1 <- apply(as.data.frame(featureVector[partition==partitionLabels[partitionIndices[1]],]),2,mean)
  center2 <- apply(as.data.frame(featureVector[partition==partitionLabels[partitionIndices[2]],]),2,mean)
  return(sum((center2-center1)^2)^(1/p))
}

interclusterDistance <- function(x,partitions,partitionIndices=c(1,2),interDistanceType=c("centroidDistance","meanDistance","minDistance","maxDistance")){
  interDistanceType <- match.arg(interDistanceType)
  if (is.data.frame(x)){
    distanceMatrix <- dist(x)
    if (interDistanceType == "centroidDistance"){
      return(interclusterCentroidDistance(partitions,x,partitionIndices))
    }
  }
  else if (is.numeric(x)){
    distanceMatrix <- x
    if (interDistanceType == "centroidDistance"){
      stop("Cannot use centroidDistance on distance matrix input. Choose another interDistanceType or provide a feature vector for x.")
    }
  }
  else{
    stop("Input value for parameter 'x' is neither a feature vector (data frame) nor a distance matrix")
    return(NA)
  }
  partitionLabels <- unique(partitions)
  group1Inds <- which(partitions==partitionLabels[partitionIndices[1]])
  group2Inds <- which(partitions==partitionLabels[partitionIndices[2]])
  if (interDistanceType == "minDistance"){
    minDist <- max(distanceMatrix)
    for (i in 1:length(group1Inds)){
      distInds <- sapply(group2Inds,lowerTriangleIndex,group1Inds[i])
      currMinDist <- min(distanceMatrix[distInds])
      if (currMinDist < minDist){
        minDist <- currMinDist
      }
    }
    return(minDist)
  }
  else if (interDistanceType == "maxDistance"){
    maxDist <- 0
    for (i in 1:length(group1Inds)){
      distInds <- sapply(group2Inds,lowerTriangleIndex,group1Inds[i])
      currMaxDist <- max(distanceMatrix[distInds])
      if (currMaxDist > maxDist){
        maxDist <- currMaxDist
      }
    }
    return(maxDist)
  }
  else if (interDistanceType == "meanDistance"){
    dists <- NA
    length(dists) <- length(group1Inds)*length(group2Inds)
    for (i in 1:length(group1Inds)){
      distInds <- sapply(group2Inds,lowerTriangleIndex,group1Inds[i])
      dists <- append(dists,distanceMatrix[distInds])
    }
    return(mean(dists))
  }
  return(NA)
}

intraclusterScatter <- function(featureVector,q=2){
  sumScatter <- 0
  # Calculate given feature vector input
  center <- apply(featureVector,2,mean)
  for (i in 1:nrow(featureVector)){
    sumScatter <- sumScatter + sum((featureVector[i,]-center)^2)^(q/2)
  }
  return((sumScatter/nrow(featureVector))^(1/q))
}

intraclusterDistance <- function(x,intraDistanceType=c("meanDistance","maxDistance","scatter")){
  intraDistanceType <- match.arg(intraDistanceType)
  if (is.data.frame(x)){
    distanceMatrix <- dist(x)
    if (intraDistanceType == "scatter"){
      return(intraclusterScatter(x))
    }
  }
  else if (is.numeric(x)){
    distanceMatrix <- x
    if (intraDistanceType == "scatter"){
      stop("Cannot use scatter on distance matrix input. Choose another intraDistanceType or provide a feature vector for x.")
    }
  }
  else{
    stop("Input value for parameter 'x' is neither a feature vector (data frame) nor a distance matrix")
    return(NA)
  }
  if (intraDistanceType == "maxDistance"){
    return(max(distanceMatrix))
  }
  else if (intraDistanceType == "meanDistance"){
    return(mean(distanceMatrix))
  }
  return(NA)
}

DaviesBouldinIndex <- function(partition,featureVector){
  clusters <- unique(partition)
  numClusters <- length(clusters)
  scatters <- NA
  length(scatters) <- numClusters
  if (is.numeric(featureVector)){
    featureVector <- as.data.frame(featureVector)
  }
  for (i in 1:numClusters){
    scatters[i] <- intraclusterScatter(as.data.frame(featureVector[partition==clusters[i],]))
  }
  RValues <- NA
  length(RValues) <- numClusters*(numClusters-1)/2
  for (i in 1:(numClusters-1)){
    for (j in (i+1):numClusters){
      RValues[lowerTriangleIndex(i,j)] <- (scatters[i]+scatters[j])/interclusterDistance(featureVector,partition,partitionIndices=c(i,j))
    }
  }
  DValues <- NA
  length(DValues) <- numClusters
  for (i in 1:numClusters){
    DValues[i] <- max(RValues[sapply((1:numClusters)[-i],lowerTriangleIndex,i)])
  }
  DBInd <- sum(DValues)/numClusters
  return(DBInd)
}

DunnIndex <- function(partitions,x,
                      intraDistanceType=c("meanDistance","maxDistance","scatter"),
                      interDistanceType=c("minDistance","maxDistance","meanDistance","centroidDistance")){
  partitionLabels <- unique(partitions)
  numPartitions <- length(partitionLabels)
  
  intraclusterDistances <- NA
  length(intraclusterDistances) <- numPartitions
  partitionVectors <- list()
  for (i in 1:numPartitions){
    if (is.data.frame(x)){
      partitionValues[[i]] <- x[partitions==partitionLabels[i],]
    }
    else if (is.numeric(x)){
      groupIndices <- which(partitions==partitionLabels[i])
      distanceMatrix <- matrix(nrow=length(groupIndices),ncol=length(groupIndices))
      for (j in 1:(length(groupIndices)-1)){
        distanceMatrix[(j+1):length(groupIndices),j] <- x[sapply(groupIndices[-(1:j)],lowerTriangleIndex,groupIndices[j])]
      }
      partitionValues[[i]] <- as.dist(distanceMatrix)
    }
    else{
      stop("Input value for parameter 'x' is neither a feature vector (data frame) nor a distance matrix")
    }
    intraclusterDistances[i] <- intraclusterDistance(partitionValues[[i]],intraDistanceType)
  }
  
  interclusterDistances <- c()
  for (i in 1:(numPartitions-1)){
    for (j in (i+1):numPartitions){
      interclusterDistances <- append(interclusterDistances,interclusterDistance(x,partitions,partitionIndices=c(i,j),interDistanceType))
    }
  }
  return(min(interclusterDistances)/max(intraclusterDistances))
}
